//$Id: IngresDialect.java,v 1.8 2005/04/13 04:34:44 oneovthafew Exp $
package org.hibernate.dialect;


import org.hibernate.Hibernate;
import org.hibernate.MappingException;
import org.hibernate.QueryException;
import org.hibernate.dialect.function.*;
import org.hibernate.engine.Mapping;
import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.type.Type;

import java.sql.CallableStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.List;


/**
 * An Ingres SQL dialect
 *
 * @author Ian Booth
 */
public class IngresDialect extends Dialect {

  public IngresDialect() {
    super();
    registerColumnType(Types.BIT, "tinyint");
    registerColumnType(Types.TINYINT, "tinyint");
    registerColumnType(Types.SMALLINT, "smallint");
    registerColumnType(Types.INTEGER, "integer");
    registerColumnType(Types.BIGINT, "bigint");
    registerColumnType(Types.REAL, "real");
    registerColumnType(Types.FLOAT, "float");
    registerColumnType(Types.DOUBLE, "float");
    registerColumnType(Types.NUMERIC, "numeric($p, $s)");
    registerColumnType(Types.DECIMAL, "decimal($p, $s)");
    registerColumnType(Types.BINARY, 32000, "byte($l)");
    registerColumnType(Types.BINARY, "long byte");
    registerColumnType(Types.VARBINARY, 32000, "varbyte($l)");
    registerColumnType(Types.VARBINARY, "long byte");
    registerColumnType(Types.LONGVARBINARY, "long byte");
    registerColumnType(Types.CHAR, "char(1)");
    registerColumnType(Types.VARCHAR, 32000, "varchar($l)");
    registerColumnType(Types.VARCHAR, "long varchar");
    registerColumnType(Types.LONGVARCHAR, "long varchar");
    registerColumnType(Types.DATE, "date");
    registerColumnType(Types.TIME, "date");
    registerColumnType(Types.TIMESTAMP, "date");
    registerColumnType(Types.BLOB, "long byte");
    registerColumnType(Types.CLOB, "long varchar");

    /*
    ** Override Dialect.java mappings.
    */
    registerFunction("substring", new SQLFunctionTemplate(Hibernate.STRING, "substring(?1 from ?2 for ?3)"));
    registerFunction("length", new StandardSQLFunction("length", Hibernate.LONG));
    registerFunction("bit_length", new SQLFunctionTemplate(Hibernate.INTEGER, "bit_length(byte(?1))"));
    registerFunction("abs", new StandardSQLFunction("abs"));
    registerFunction("mod", new StandardSQLFunction("mod", Hibernate.INTEGER));
    registerFunction("sqrt", new StandardSQLFunction("sqrt", Hibernate.DOUBLE));
    registerFunction("upper", new StandardSQLFunction("upper"));
    registerFunction("lower", new StandardSQLFunction("lower"));
    registerFunction("extract", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('?1', ?3)"));
    registerFunction("second", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('second', ?1)"));
    registerFunction("minute", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('minute', ?1)"));
    registerFunction("hour", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('hour', ?1)"));
    registerFunction("day", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('day', ?1)"));
    registerFunction("month", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('month', ?1)"));
    registerFunction("year", new SQLFunctionTemplate(Hibernate.INTEGER, "date_part('year', ?1)"));

    /*
    ** locate(pat,str) => locate(str,pat)
    ** locate(pat,str,beg) =>
    **	(locate(substring(str from beg)+beg-1,pat)
    */
    registerFunction("locate",
            new SQLFunction() {
              public Type
                      getReturnType(Type columnType, Mapping mapping)
                      throws QueryException {
                return Hibernate.INTEGER;
              }

              public boolean hasArguments() {
                return true;
              }

              public boolean hasParenthesesIfNoArguments() {
                return true;
              }

              public String
                      render(List args,
                             SessionFactoryImplementor factory)
                      throws QueryException {
                boolean threeArgs = (args.size() > 2);
                Object pattern = args.get(0);
                Object string = args.get(1);
                Object start = threeArgs ? args.get(2) : null;
                StringBuffer buf = new StringBuffer();

                if (threeArgs) buf.append('(');

                buf.append("locate(");

                if (threeArgs) buf.append("substring(");

                buf.append(string);

                if (threeArgs) {
                  buf.append(" from ");
                  buf.append(start);
                  buf.append(')');
                }

                buf.append(',');
                buf.append(pattern);
                buf.append(')');

                if (threeArgs) {
                  buf.append('+');
                  buf.append(start);
                  buf.append("-1)");
                }

                return (buf.toString());
              }
            }
    );

    /*
    ** trim(str) => trim(both ' ' from str)
    ** trim(where what from str) => trim(where what from str)
    */
    registerFunction("trim",
            new SQLFunction() {
              public Type
                      getReturnType(Type columnType, Mapping mapping)
                      throws QueryException {
                return Hibernate.STRING;
              }

              public boolean hasArguments() {
                return true;
              }

              public boolean hasParenthesesIfNoArguments() {
                return true;
              }

              public String
                      render(List args,
                             SessionFactoryImplementor factory)
                      throws QueryException {
                int count = args.size();

                if (count == 1)
                  return ("trim( both ' ' from " +
                          args.get(0) + " )");

                StringBuffer buf = new StringBuffer("trim(");

                for (int arg = 0; arg < count; arg++) {
                  buf.append(' ');
                  buf.append(args.get(arg));
                }

                buf.append(" )");
                return (buf.toString());
              }
            }
    );

    /*
    ** coalesce(s1,s1) => ifnull(s1,s2)
    ** coalesce(s1,s2,...sn) => ifnull(s1,ifnull(s2,...sn))
    */
    registerFunction("coalesce",
            new VarArgsSQLFunction(Hibernate.STRING,
                    "coalesce(", ",", ")") {
              public String
                      render(List args, SessionFactoryImplementor factory)
                      throws QueryException {
                StringBuffer prefix = new StringBuffer();
                StringBuffer suffix = new StringBuffer();
                int count = args.size();

                for (int arg = 0; arg < count - 1; arg++) {
                  prefix.append("coalesce(");
                  prefix.append(args.get(arg));
                  prefix.append(',');
                  suffix.append(')');
                }

                prefix.append(args.get(count - 1));
                prefix.append(suffix);
                return (prefix.toString());
              }
            }
    );

    /*
    ** Additional Hibernate Standard Functions.
    **
    ** Transform to nested concat() calls:
    ** concat(s1,s2,s3) => concat(concat(s1,s2),s3)
    */
    registerFunction("concat",
            new VarArgsSQLFunction(Hibernate.STRING,
                    "concat(", ",", ")") {
              public String
                      render(List args, SessionFactoryImplementor factory)
                      throws QueryException {
                StringBuffer prefix = new StringBuffer();
                StringBuffer suffix = new StringBuffer();
                int count = args.size();

                for (int arg = 1; arg < count; arg++) {
                  prefix.append("concat(");
                  suffix.append(',');
                  suffix.append(args.get(arg));
                  suffix.append(')');
                }

                prefix.append(args.get(0));
                prefix.append(suffix);
                return (prefix.toString());
              }
            }
    );

    registerFunction("current_timestamp", new NoArgSQLFunction("date('now')", Hibernate.TIMESTAMP, false));
    registerFunction("rtrim", new SQLFunctionTemplate(Hibernate.STRING, "trim(trailing ' ' from ?1)"));
    registerFunction("ltrim", new SQLFunctionTemplate(Hibernate.STRING, "trim(leading ' ' from ?1)"));

    /*
    ** Ingres scalar functions
    */
    registerFunction("ascii", new StandardSQLFunction("ascii", Hibernate.STRING));
    registerFunction("atan", new StandardSQLFunction("atan", Hibernate.DOUBLE));
    registerFunction("bit_add", new StandardSQLFunction("bit_add"));
    registerFunction("bit_and", new StandardSQLFunction("bit_and"));
    registerFunction("bit_not", new StandardSQLFunction("bit_not"));
    registerFunction("bit_or", new StandardSQLFunction("bit_or"));
    registerFunction("bit_xor", new StandardSQLFunction("bit_xor"));
    registerFunction("character_length", new StandardSQLFunction("character_length", Hibernate.LONG));
    registerFunction("charextract", new StandardSQLFunction("charextract", Hibernate.STRING));
    registerFunction("cos", new StandardSQLFunction("cos", Hibernate.DOUBLE));
    registerFunction("current_user", new NoArgSQLFunction("current_user", Hibernate.STRING, false));
    registerFunction("date_gmt", new StandardSQLFunction("date_gmt", Hibernate.STRING));
    registerFunction("date_part", new StandardSQLFunction("date_part", Hibernate.INTEGER));
    registerFunction("date_trunc", new StandardSQLFunction("date_trunc", Hibernate.TIMESTAMP));
    registerFunction("dba", new NoArgSQLFunction("dba", Hibernate.STRING, true));
    registerFunction("dbmsinfo", new NoArgSQLFunction("dbmsinfo", Hibernate.STRING, true));
    registerFunction("dow", new StandardSQLFunction("dow", Hibernate.STRING));
    registerFunction("exp", new StandardSQLFunction("exp", Hibernate.DOUBLE));
    registerFunction("gmt_timestamp", new StandardSQLFunction("gmt_timestamp", Hibernate.STRING));
    registerFunction("hash", new StandardSQLFunction("hash", Hibernate.INTEGER));
    registerFunction("hex", new StandardSQLFunction("hex", Hibernate.STRING));
    registerFunction("ifnull", new StandardSQLFunction("ifnull"));
    registerFunction("initial_user", new NoArgSQLFunction("initial_user", Hibernate.STRING, false));
    registerFunction("intextract", new StandardSQLFunction("intextract", Hibernate.INTEGER));
    registerFunction("left", new StandardSQLFunction("left", Hibernate.STRING));
    registerFunction("ln", new StandardSQLFunction("ln", Hibernate.DOUBLE));
    registerFunction("log", new StandardSQLFunction("log", Hibernate.DOUBLE));
    registerFunction("lowercase", new StandardSQLFunction("lowercase"));
    registerFunction("octet_length", new StandardSQLFunction("octet_length", Hibernate.LONG));
    registerFunction("pad", new StandardSQLFunction("pad", Hibernate.STRING));
    registerFunction("position", new StandardSQLFunction("position", Hibernate.LONG));
    registerFunction("power", new StandardSQLFunction("power", Hibernate.DOUBLE));
    registerFunction("random", new NoArgSQLFunction("random", Hibernate.LONG, true));
    registerFunction("randomf", new NoArgSQLFunction("randomf", Hibernate.DOUBLE, true));
    registerFunction("right", new StandardSQLFunction("right", Hibernate.STRING));
    registerFunction("session_user", new NoArgSQLFunction("session_user", Hibernate.STRING, false));
    registerFunction("shift", new StandardSQLFunction("shift"));
    registerFunction("sin", new StandardSQLFunction("sin", Hibernate.DOUBLE));
    registerFunction("size", new NoArgSQLFunction("size", Hibernate.LONG, true));
    registerFunction("soundex", new StandardSQLFunction("soundex", Hibernate.STRING));
    registerFunction("squeeze", new StandardSQLFunction("squeeze"));
    registerFunction("system_user", new NoArgSQLFunction("system_user", Hibernate.STRING, false));
    registerFunction("unhex", new StandardSQLFunction("unhex", Hibernate.STRING));
    registerFunction("uppercase", new StandardSQLFunction("uppercase"));
    registerFunction("user", new NoArgSQLFunction("user", Hibernate.STRING, false));
    registerFunction("usercode", new NoArgSQLFunction("usercode", Hibernate.STRING, true));
    registerFunction("username", new NoArgSQLFunction("username", Hibernate.STRING, true));
    registerFunction("uuid_create", new StandardSQLFunction("uuid_create", Hibernate.BYTE));
    registerFunction("uuid_compare", new StandardSQLFunction("uuid_compare", Hibernate.INTEGER));
    registerFunction("uuid_from_char", new StandardSQLFunction("uuid_from_char", Hibernate.BYTE));
    registerFunction("uuid_to_char", new StandardSQLFunction("uuid_to_char", Hibernate.STRING));

  }

  /**
   * Do we need to drop constraints before dropping tables in this dialect?
   *
   * @return boolean
   */
  public boolean dropConstraints() {
    return false;
  }

  /**
   * Does the <tt>FOR UPDATE OF</tt> syntax specify particular
   * columns?
   */
  public boolean forUpdateOfColumns() {
    return true;
  }

  /**
   * The syntax used to add a column to a table (optional).
   */
  public String getAddColumnString() {
    return " add column ";
  }

  /**
   * The keyword used to specify a nullable column.
   *
   * @return String
   */
  public String getNullColumnString() {
    return " with null ";
  }

  /**
   * Does this dialect support sequences?
   *
   * @return boolean
   */
  public boolean supportsSequences() {
    return true;
  }

  /**
   * The syntax that fetches the next value of a sequence, if sequences are supported.
   *
   * @param sequenceName the name of the sequence
   * @return String
   * @throws MappingException if no sequences
   */
  public String getSequenceNextValString(String sequenceName) throws MappingException {
    return "select nextval for " + sequenceName;
  }

  /**
   * The syntax used to create a sequence, if sequences are supported.
   *
   * @param sequenceName the name of the sequence
   * @return String
   * @throws MappingException if no sequences
   */
  public String getCreateSequenceString(String sequenceName) throws MappingException {
    return "create sequence " + sequenceName;
  }

  /**
   * The syntax used to drop a sequence, if sequences are supported.
   *
   * @param sequenceName the name of the sequence
   * @return String
   * @throws MappingException if no sequences
   */
  public String getDropSequenceString(String sequenceName) throws MappingException {
    return "drop sequence " + sequenceName + " restrict";
  }

  /**
   * A query used to find all sequences
   */
  public String getQuerySequencesString() {
    return "select seq_name from iisequence";
  }

  /**
   * Does this <tt>Dialect</tt> have some kind of <tt>LIMIT</tt> syntax?
   */
  public boolean supportsLimit() {
    return true;
  }

  /**
   * Does this dialect support an offset?
   */
  public boolean supportsLimitOffset() {
    return false;
  }

  /**
   * Add a <tt>LIMIT</tt> clause to the given SQL <tt>SELECT</tt>
   *
   * @return the modified SQL
   */
  public String getLimitString(String querySelect, int offset, int limit) {
    if (offset > 0) return getLimitString(querySelect, true);
    return new StringBuffer(querySelect.length() + 16)
            .append(querySelect)
            .insert(6, " first " + limit)
            .toString();
  }

  public boolean supportsVariableLimit() {
    return false;
  }

  /**
   * Does the <tt>LIMIT</tt> clause take a "maximum" row number instead
   * of a total number of returned rows?
   */
  public boolean useMaxForLimit() {
    return true;
  }

  public String getSelectGUIDString() {
    return "select uuid_to_char(uuid_create())";
  }

  public boolean supportsOuterJoinForUpdate() {
    return false;
  }

  public int registerResultSetOutParameter(CallableStatement statement, int col) throws SQLException {
    return col;
  }

  public ResultSet getResultSet(CallableStatement ps) throws SQLException {
    return (ps.executeQuery());
  }

  public boolean supportsUnionAll() {
    return true;
  }
}
