package org.jblooming.company;

import org.jblooming.agenda.Period;
import org.jblooming.anagraphicalData.AnagraphicalData;
import org.jblooming.ontology.Documentable;
import org.jblooming.ontology.Node;
import org.jblooming.ontology.SecuredNodeSupport;
import org.jblooming.operator.Operator;
import org.jblooming.operator.OperatorAggregator;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.exceptions.FindException;
import org.jblooming.persistence.exceptions.QueryException;
import org.jblooming.persistence.exceptions.StoreException;
import org.jblooming.resource.Affiliate;
import org.jblooming.security.AffiliationRole;

import java.util.*;

public class Department extends SecuredNodeSupport implements OperatorAggregator {

  private String code;

  private String name;

  private List<Location> locations = new ArrayList<Location>();

  private Set anagraphicalDatas = new HashSet();

  private Set documents = new HashSet();

  private String notes;

  /**
   * determines the department type, e.g. Company, Department, Office
   */
  private DepartmentType type;

  private boolean deptHidden = false;

  /**
   * Introduced in order to introduce virtual or multiuser resources, such as "a board", "a counsel" "stearing committee"
   */
  //read only collection
  private Set<Affiliate> affiliations;

  private boolean staff = false;


  /**
   * this is a map where the key is the departmentType.code.toString() and the value is the top-most parent of this type
   */
  private HashMap<String, Department> parentsMap=null;

  public Department() {
    super();
  }

  public void setParent(Department n) {
    parent = n;
  }

  public void setParentNode(Node node) {
    setParent((Department) node);
  }

  public void addAnagraphicalData(AnagraphicalData data) {
    anagraphicalDatas.add(data);
  }

  public void removeAnagraphicalData(AnagraphicalData data) {
    anagraphicalDatas.remove(data);
  }

  public Iterator getAnagraphicalDatasIterator() {
    return anagraphicalDatas.iterator();
  }

  public boolean isInAnagraphicalData(AnagraphicalData ad) {
    return anagraphicalDatas.contains(ad) ? true : false;
  }

  public Period getAvailability() {
    return null;
  }

  public String getNotes() {
    return notes;
  }

  public void setNotes(String notes) {
    this.notes = notes;
  }


  public Department getParent() {
    return (Department) parent;
  }


  public void addDocument(Documentable document) {
    documents.add(document);
  }

  public void removeDocument(Documentable document) {
    documents.remove(document);
  }

  public Iterator getDocumentsIterator() {
    return documents.iterator();
  }

  public Node getParentNode() {
    return getParent();
  }

  public Set<AnagraphicalData> getAnagraphicalDatas() {
    return anagraphicalDatas;
  }

  public void setAnagraphicalDatas(Set<AnagraphicalData> anagraphicalDatas) {
    this.anagraphicalDatas = anagraphicalDatas;
  }

  public Iterator getAffiliationsIterator() {
    return affiliations.iterator();
  }

  public boolean affiliationsContains(Affiliate affiliation) {
    return affiliations.contains(affiliation);
  }

  public int affiliationsSize() {
    return affiliations.size();
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }


  public boolean inherits() {
    return false;  //To change body of implemented methods use File | Settings | File Templates.
  }

  public boolean propagates() {
    return false;  //To change body of implemented methods use File | Settings | File Templates.
  }

  public Collection getOperators() throws QueryException, FindException {
    Set ops = new HashSet();
    for (Iterator iterator = affiliations.iterator(); iterator.hasNext();) {
      Affiliate affiliate = (Affiliate) iterator.next();
      ops.add(affiliate.getOperator());
    }
    return ops;
  }


  public boolean isOperatorIn(Operator o) throws FindException, QueryException {
    return getOperators().contains(o);
  }

  public void addAffiliationAndPersist(Operator operator, AffiliationRole role) throws StoreException {
    Affiliate aff = new Affiliate();
    aff.setDepartment(this);
    aff.setOperator(operator);
    aff.setRole(role);
    aff.store();
  }

  public void removeAffiliateFromMemory(Affiliate affiliation) {
    affiliations.remove(affiliation);
  }

  public void removeLocationFromMemory(Location location) {
    locations.remove(location);
  }

  public void removeAllLocationFromMemory() {
    List locs = this.getLocations();
    for (int i = 0; i < locs.size(); i++) {
      Location location = (Location) locs.get(i);
      this.removeLocationFromMemory(location);
    }
//    locations.remove(location);
  }

  public DepartmentType getType() {
    return type;
  }

  public void setType(DepartmentType type) {
    this.type = type;
  }

  public String getCode() {
    return code;
  }

  public void setCode(String code) {
    this.code = code;
  }

  public boolean isStaff() {
    return staff;
  }

  public void setStaff(boolean staff) {
    this.staff = staff;
  }

  public Set<Affiliate> getAffiliations() {
    return affiliations;
  }

  private void setAffiliations(Set<Affiliate> affiliations) {
    this.affiliations = affiliations;
  }

  private Set getDocuments() {
    return documents;
  }

  private void setDocuments(Set documents) {
    this.documents = documents;
  }



  public List<Location> getLocations() {
    return locations;
  }

  public void setLocations(List<Location> locations) {
    this.locations = locations;
  }


  public static List<Department> getExistingDepartment() throws FindException {
    String hql = "from " + Department.class.getName() + " as dt ";
    OqlQuery oql = new OqlQuery(hql);
    return oql.list();
  }



  public Department getParentOfType(DepartmentType.Code departMentTypeCode){
    return getParentsMap().get(departMentTypeCode.toString());
  }

  private Map<String, Department> getParentsMap() {
    if (parentsMap==null){
      parentsMap = new HashMap<String, Department>();
      fillParentMap(parentsMap);
    }
    return (parentsMap);
  }

  private void fillParentMap(HashMap<String, Department> hm) {
    if (getType() != null) {
      hm.put(getType().getStringValue(), this);
      if (getParent() != null) {
        getParent().fillParentMap(hm);
      }
    }
  }

  public boolean isDeptHidden() {
    return deptHidden;
  }

  public void setDeptHidden(boolean deptHidden) {
    this.deptHidden = deptHidden;
  }
}



