package org.jblooming.messaging;

import org.jblooming.PlatformRuntimeException;
import org.jblooming.ontology.PerformantNode;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.FindByPrimaryKeyException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.persistence.hibernate.HibernateFactory;
import org.jblooming.persistence.hibernate.PersistenceContext;
import org.jblooming.scheduler.ExecutableSupport;
import org.jblooming.scheduler.JobLogData;
import org.jblooming.tracer.Tracer;
import org.jblooming.utilities.DateUtilities;
import org.jblooming.utilities.ReflectionUtilities;
import org.jblooming.utilities.StringUtilities;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class EventListenerMatcher extends ExecutableSupport {

  public JobLogData run(JobLogData jobLogData) throws Exception {
    boolean exception = true;
    PersistenceContext pc = null;

    try {
      pc = HibernateFactory.newFreeSession();

      // get all the events
      OqlQuery eventQH = new OqlQuery("from " + SomethingHappened.class.getName(), pc);
      List<SomethingHappened> events = eventQH.list();

      // loop through each event
      for (SomethingHappened event : events) {
        List<Listener> listeners = getListeners(event, pc);
        // these listeners already match the event
        // generate the appropriate message
        for (Listener l : listeners) {
          Message.generateAndPersistMessage(l, event, pc);
          l.setLastMatchingDate(new Date());
          if (l.isOneShot())
            l.remove(pc);
          else
            l.store(pc);
        }
      }

      for (SomethingHappened event : events)
        event.remove(pc);

      pc.commitAndClose();

      jobLogData.notes = jobLogData.notes + "EventListenerMatcher executed on " + DateUtilities.dateToTimestampString(DateUtilities.getAbsoluteNow());
      exception = false;
    } catch (Exception e) {
      Tracer.platformLogger.error("EventListenerMatcher error", e);
    } finally {
      if (exception) {
        if (pc != null) {
          pc.rollback();
        }
        jobLogData.successfull = false;
      }
    }

    return jobLogData;
  }

  private List<Listener> getListeners(SomethingHappened event, PersistenceContext pc) throws PersistenceException {

    List<Listener> listeners = new ArrayList();

    if (event.getHappeningExpiryDate() == null || event.getHappeningExpiryDate().getTime() > System.currentTimeMillis()) {

      List<String> ids = new ArrayList();

      //does event refer to node ?
      boolean refersToNode = false;
      Class main = null;
      try {
        main = Class.forName(event.getTheClass());
      } catch (ClassNotFoundException e) {
        throw new PlatformRuntimeException(e);
      }
      if (event.getTheClass() != null) {

        refersToNode = ReflectionUtilities.extendsOrImplements(main, PerformantNode.class);

      }
      //get ancestors
      if (refersToNode) {
        try {
          PerformantNode pn = (PerformantNode) PersistenceHome.findByPrimaryKey(main, event.getIdentifiableId(), pc);
          if (pn!=null && pn.getAncestorIds() != null) {
            ids = StringUtilities.splitToList(pn.getAncestorIds(), PerformantNode.SEPARATOR);
          }
        } catch (FindByPrimaryKeyException e) {
          //do nothing
        }
      }
      ids.add(event.getIdentifiableId());

      String hql = "from " + Listener.class.getName() + " as l ";
      hql = hql + " where l.theClass=:theClass ";
      hql = hql + " and (l.identifiableId in (:identifiableIds) or l.identifiableId is null) ";
      hql = hql + " and (l.eventType = :eventType or l.eventType is null) ";
      hql = hql + " and ( l.validityEnd is null or (l.validityEnd >= :nownow) )";
      hql = hql + " and ( l.validityStart is null or (l.validityStart <= :nownow) )";

      OqlQuery oql = new OqlQuery(hql, pc);
      oql.getQuery().setString("theClass", event.getTheClass());
      oql.getQuery().setParameterList("identifiableIds", ids);
      oql.getQuery().setString("eventType", event.getEventType());
      oql.getQuery().setTimestamp("nownow", event.getHappenedAt());

      listeners = oql.list();
    }
    return listeners;
  }
}
