package org.jblooming.page;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.ScrollMode;
import org.hibernate.ScrollableResults;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.tracer.Tracer;
import org.jblooming.waf.settings.ApplicationState;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 */
public class HibernatePage implements Page {

  protected Query query;

  protected List elements;
  protected int pageSize;
  protected int pageNumber;
  protected int totalElements = -1;

  public static Set jdbcClassesSupportingScrollCursors = new HashSet();
  private ScrollableResults scrollableResults;
  public static final int DEFAULT_PAGE_SIZE = 10 ;

  private HibernatePage(int pageNumber, int pageSize) {
    this.pageNumber = pageNumber;
    this.pageSize = pageSize;
  }


  public boolean isFirstPage() {
    return getPageNumber() == 0;
  }

  public boolean isLastPage() {
    return getPageNumber() >= getLastPageNumber();
  }

  public boolean hasNextPage() {
    return elements.size() > getPageSize();
  }

  public boolean hasPreviousPage() {
    return getPageNumber() > 0;
  }

  public int getLastPageNumber() {

    double totalResults = new Integer(getTotalNumberOfElements()).doubleValue();
    return new Double(Math.floor(totalResults / getPageSize())).intValue();
  }

  public List getThisPageElements() {
/*
* Since we retrieved one more than the specified pageSize when the
* class was constructed, we now trim it down to the pageSize if a next
* page exists.
*/
    return hasNextPage() ? elements.subList(0, getPageSize()) : elements;
  }

  public Logger getLogger() {
    return Tracer.hibernateLogger;
  }

  public int getTotalNumberOfElements() {
    return totalElements;
  }

  public int getThisPageFirstElementNumber() {
    return getPageNumber() * getPageSize() + 1;
  }

  public int getThisPageLastElementNumber() {
    int fullPage = getThisPageFirstElementNumber() + getPageSize() - 1;
    return getTotalNumberOfElements() < fullPage ?
            getTotalNumberOfElements() :
            fullPage;
  }

  public int getNextPageNumber() {
    return getPageNumber() + 1;
  }

  public int getPreviousPageNumber() {
    return getPageNumber() - 1;
  }

  public int getPageSize() {
    return pageSize;
  }

  public int getPageNumber() {
    return pageNumber;
  }

  public List getAllElements() {
    HibernatePage pageTmp = getHibernatePageInstance(query,1,getTotalNumberOfElements());
    return pageTmp.getThisPageElements();
  }


  public static HibernatePage getHibernatePageInstance(Query query, int pageNumber, int pageSize) {

    return getHibernatePageInstance(query, pageNumber, pageSize, ApplicationState.platformConfiguration.driver_class);
  }

  public static HibernatePage getHibernatePageInstance(Query query,
                                                       int pageNumber,
                                                       int pageSize,
                                                       String driverClass) {

    if (query.getQueryString().toLowerCase().indexOf("order by")==-1) {
      Tracer.platformLogger.warn("Using pagination without order by can lead to inconsistent results, for example on certain Oracle instances: "+query.getQueryString());
      Tracer.platformLogger.warn("Query used in: "+Tracer.getCallTrace(false));
    }

    if (jdbcClassesSupportingScrollCursors.contains(driverClass))
      return HibernatePage.getScrollPageInstanceWithTotalByScroll(query, pageNumber, pageSize);
    else
      return HibernatePage.getScrollPageInstanceWithTotalByList(query, pageNumber, pageSize);
  }

  /**
   * Construct a new HibernatePage. HibernatePage numbers are zero-based so the
   * first page is page 0.
   *
   * @param query      the Hibernate Query
   * @param pageNumber the page number (zero-based);
   *                   if Integer.MAX_VALUE will return the last page for the query
   * @param pageSize   the number of results to display on the page
   */
  protected static HibernatePage getScrollPageInstanceWithTotalByScroll(Query query, int pageNumber, int pageSize) {

    HibernatePage sp = new HibernatePage(pageNumber, pageSize);
    sp.query = query;
    try {
      sp.scrollableResults = query.scroll(ScrollMode.SCROLL_SENSITIVE);
      sp.scrollableResults.last();
      sp.totalElements = sp.scrollableResults.getRowNumber();

      sp.determineElements();
      sp.scrollableResults.close();
    } catch (HibernateException e) {
      sp.getLogger().error("Failed to create ScrollPage by getScrollPageInstanceWithTotalByScroll: " + e.getMessage());
      throw new PlatformRuntimeException(e);
    }

    return sp;
  }


  /**
   * Construct a new HibernatePage. HibernatePage numbers are zero-based so the
   * first page is page 0.
   *
   * @param query      the Hibernate Query
   * @param pageNumber the page number (zero-based);
   *                   if Integer.MAX_VALUE will return the last page for the query
   * @param pageSize   the number of results to display on the page
   */
  protected static HibernatePage getScrollPageInstanceWithTotalByList(Query query, int pageNumber, int pageSize) {

    HibernatePage sp = new HibernatePage(pageNumber, pageSize);
    sp.query = query;
    try {
      //bacame useless
      //sp.scrollableResults = query.scroll(ScrollMode.FORWARD_ONLY);
      sp.totalElements = sp.calculateTotalElementsByList();

      sp.determineElements();

    } catch (HibernateException e) {
      sp.getLogger().error("Failed to create ScrollPage by getScrollPageInstanceWithTotalByQuery: " + e.getMessage());
      throw new PlatformRuntimeException(e);
    }

    return sp;
  }

  private void determineElements() throws HibernateException {

    if (this.pageSize<=0) this.pageSize=HibernatePage.DEFAULT_PAGE_SIZE;

    if (Integer.MAX_VALUE == this.pageNumber)
      this.pageNumber = (getTotalNumberOfElements() / this.pageSize);
    else if (pageNumber>(totalElements/pageSize))
      pageNumber = totalElements/pageSize;
    /**
     * todo: this no more makes sense, as we know the total number of elements
     *
     * We set the max results to one more than the specfied pageSize to
     * determine if any more results exist (i.e. if there is a next page
     * to display). The result set is trimmed down to just the pageSize
     * before being displayed later (in getThisPageElements()).
     */
    query = query.setFirstResult(this.pageNumber * this.pageSize);
    query = query.setMaxResults(this.pageSize + 1);
    elements = query.list();
  }

  private int calculateTotalElementsByList() throws HibernateException {

    return query.list().size()-1;
    
    //this is different hacky way to do it; but actually it results slower
    /*String hql = query.getQueryString();
    // if (hql.toLowerCase().startsWith("select"))
       hql = "select count(*) " + hql.substring(hql.indexOf("from"));
    // else
    //   hql = "select count(*) " + hql;
     OqlQuery oql = null;
     try {
       oql = new OqlQuery(hql);
     } catch (PersistenceException e) {
       throw new PlatformRuntimeException(e);
     }

     try {
       Field field = ReflectionUtilities.getField("namedParameters",QueryImpl.class);
       field.setAccessible(true);
       Map parValues = (Map) field.get(((QueryImpl) query));
       for (Iterator<String> iterator = parValues.keySet().iterator(); iterator.hasNext();) {
         String  par = iterator.next();
         oql.getQuery().setParameter(par, parValues.get(par));
       }
     } catch (IllegalAccessException e) {
       throw new PlatformRuntimeException(e);
     }

     try {
       return (Integer) (oql.uniqueResult());
     } catch (FindException e) {
       throw new PlatformRuntimeException(e);
     }*/

  }


  public Object getSampleElement() {
    if (elements!=null && elements.size()>0)
      return elements.get(0);
    else
      return null;
  }

}
