package org.jblooming.persistence.hibernate;


import org.hibernate.*;
import org.hibernate.cfg.Configuration;
import org.hibernate.type.Type;
import org.jblooming.ApplicationException;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.logging.Sniffer;
import org.jblooming.ontology.Identifiable;
import org.jblooming.ontology.LoggableIdentifiable;
import org.jblooming.ontology.SecuredNodeSupport;
import org.jblooming.operator.Operator;
import org.jblooming.persistence.PersistenceSession;
import org.jblooming.persistence.exceptions.FindByPrimaryKeyException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.persistence.exceptions.RemoveException;
import org.jblooming.persistence.exceptions.StoreException;
import org.jblooming.security.Securable;
import org.jblooming.system.SystemConstants;
import org.jblooming.tracer.Tracer;
import org.jblooming.transaction.hibernate.HibernateTransaction;
import org.jblooming.utilities.DateUtilities;
import org.jblooming.utilities.ReflectionUtilities;
import org.jblooming.waf.constants.Fields;
import org.jblooming.waf.settings.ApplicationState;
import org.jblooming.waf.settings.PlatformConfiguration;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.util.Date;
import java.util.List;
import java.util.Map;

public class HibernateFactory {

  protected static String xmlFilename;
  private static SessionFactory sessionFactory;

  protected static boolean debugThreads = false;
  protected static boolean debugDatastore = false;
  private static final boolean debugLoad = false;

  protected static ThreadLocal<PersistenceSession> threadLocalSession = new ThreadLocal<PersistenceSession>() {

    protected PersistenceSession initialValue() {
      return null;
    }
  };

  protected static int sessionCount = 0;

  public static Configuration config;

  protected static int DEFAULT_PAGE_SIZE = 20;


  public static void resetSessionFactory() {
    if (debugDatastore) Tracer.addTrace("HibernateFactory::resetSessionFactory()");
    sessionFactory = null;
  }

  public static PersistenceSession getPersistenceSession() {
    return threadLocalSession.get();
  }

  public static Session getSession() {

    if (debugThreads)
      Tracer.platformLogger.debug("HibernateFactory::getSession. Thread is: " + Thread.currentThread().toString());

    if (HibernateTransaction.getCurrentTransaction() == null) {
      new HibernateTransaction();
    }

    if (threadLocalSession.get() == null) {
      newSession();
    }

    return threadLocalSession.get().session;
  }

  public static Session getSessionNoInit() {
    return threadLocalSession.get().session;
  }

  public static Session newSession() {

    Session session = null;
    if (threadLocalSession.get() != null) {

      if (!Fields.FALSE.equals(ApplicationState.getApplicationSetting("KILL_DEAD_HIBSESSIONS"))) {
        //System.out.println("GRAVE: more than a session per thread detected! Connection pool limit broken. Attempting forced session close.");
        Tracer.platformLogger.error("More than a session per thread detected. If may be that db connections have been forced to close");
        //HibernateFactory.closeSession();
        return threadLocalSession.get().session;
      } else
        throw new PlatformRuntimeException("more than a session per thread detected!");
    }

    try {
      //audit trail enaction
      if (Fields.TRUE.equals(ApplicationState.getApplicationSetting(SystemConstants.AUDIT))) {
        Sniffer interceptor = new Sniffer();
        session = getSessionFactory().openSession(interceptor);
      } else
        session = getSessionFactory().openSession();

    } catch (Exception ex) {
      throw new PlatformRuntimeException("HibernateFactory::getSession. Unable to get session.", ex);
    }

    /*if (ApplicationState.platformConfiguration.oracle10gRelease2caseInsensitiveSupport) {
      PreparedStatement ps = null;
      try {
        ps = session.connection().prepareStatement("alter session set nls_comp=ansi");
        ps.executeImportLine();
        ps.close();
        ps = session.connection().prepareStatement("alter session set nls_sort=binary_ci");
        ps.executeImportLine();
        ps.close();
      } catch (SQLException e) {
        throw new PlatformRuntimeException(e);
      }
    }*/

    PersistenceSession ps = new PersistenceSession();
    ps.session = session;
    threadLocalSession.set(ps);
    ++sessionCount;
    if (debugThreads)
      Tracer.platformLogger.debug("HibernateFactory. Session opened! Count: " + sessionCount);

    return session;
  }

  public static SessionFactory getSessionFactory() {
    return sessionFactory;
  }

  public static void closeSession() {

    if (threadLocalSession.get() == null)
      throw new PlatformRuntimeException("HibernateFactory inconsistency. Thread session is null!");

    Session session = threadLocalSession.get().session;

    --sessionCount;
    threadLocalSession.set(null);
    if (debugThreads)
      Tracer.platformLogger.debug("HibernateFactory. Session closed! Count: " + sessionCount);

    try {
      session.close();
    } catch (Exception ex) {
      throw new PlatformRuntimeException("HibernateFactory session close exception.");
    }
  }

  public static void store(Identifiable i) throws StoreException {
    store(i, getSession());
  }

  public static void store(Identifiable i, PersistenceContext pc) throws StoreException {
    store(i, pc.session);
  }

  public static void store(Identifiable i, Session s) throws StoreException {

    HibernateUtilities.incrementIdIfNew(i);
    //getSession().

    try {
//      final Date absoluteNow = DateUtilities.getAbsoluteNow();
      if (i instanceof LoggableIdentifiable)
        store((LoggableIdentifiable) i, s);
      else
        save(s, i);
    } catch (Exception ex) {
      throw new StoreException("HibernateFactory store exception saving "+(i!=null ? i.getClass().getSimpleName()+" "+i.getName()+" "+i.getId() : ""), ex);
    }
  }

  private static void store(LoggableIdentifiable i, Session s) throws StoreException {

    final Date absoluteNow = DateUtilities.getAbsoluteNow();
    Operator op = getPersistenceSession() != null ? getPersistenceSession().operator : null;

    try {
      if (op != null) {
        if (!s.contains(i)) {
          i.setCreator(op.getFullname());
          if (i instanceof SecuredNodeSupport) {
            if (((Securable) i).getOwner() == null)
              ((Securable) i).setOwner(op);
          }
          i.setCreationDate(absoluteNow);
        }
        i.setLastModifier(op.getFullname());
        i.setLastModified(absoluteNow);
      } else {
        if (!s.contains(i))
          i.setCreationDate(absoluteNow);
        else
          i.setLastModified(absoluteNow);
      }
      save(s, i);

    } catch (Exception ex) {
      throw new StoreException("HibernateFactory store exception saving LoggableIdentifiable.", ex);
    }
  }

  private static void save(Session s, Identifiable i) throws HibernateException {

    boolean b = ApplicationState.platformConfiguration != null;
    if (b)
      if (!PlatformConfiguration.StringConversionType.NONE.equals(ApplicationState.platformConfiguration.stringPrimitiveFieldsConversion) ||
              !PlatformConfiguration.StringConversionType.NONE.equals(ApplicationState.platformConfiguration.searchStringParametersConversion)) {

        Map<String, Field> dif = ReflectionUtilities.getDeclaredInheritedFields(i.getClass());
        for (Field f : dif.values()) {
          if (f.getType().equals(String.class)) {
            try {
              String value = (String) ReflectionUtilities.getFieldValue(f.getName(), i);
              if (value != null) {
                if (ApplicationState.platformConfiguration.stringPrimitiveFieldsConversion.equals(PlatformConfiguration.StringConversionType.LOWER))
                  value = value.toLowerCase();
                else
                  value = value.toUpperCase();
                ReflectionUtilities.setFieldValue(f.getName(), i, value);
              }
            } catch (Exception e) {
              throw new PlatformRuntimeException(e);
            }
          }
        }
      }

    //if (i instanceof Subscribable) {
    //  SomethingHappened.generateAndPersistSomething(s, (Subscribable) i);
    //}


    s.save(i);
  }

  public static void remove(Identifiable i) throws RemoveException {
    remove(i, getSession());
  }

  public static void remove(Identifiable i, Session s) throws RemoveException {
    try {
      s.delete(i);
    } catch (Exception ex) {
      throw new RemoveException("HibernateFactory remove exception.", ex);
    }
  }

  public static Object load(Class cls, Serializable id) throws ApplicationException {
    try {
      return getSession().load(cls, id);
    } catch (Exception ex) {
      throw new ApplicationException("HibernateFactory session close exception.", ex);
    }
  }

  public static List bulkLoad(String query, Object values[], Type types[]) throws PersistenceException {
    try {
      return getSession().createQuery(query).setParameters(values, types).list();
    } catch (Exception ex) {
      throw new PersistenceException("HibernateFactory load exception.", ex);
    }
  }

  public static Identifiable findByPrimaryKey(Class cls, Serializable id) throws FindByPrimaryKeyException {
    PersistenceContext pc = new PersistenceContext();
    pc.session = getSession();
    return findByPrimaryKey(cls, id, pc);
  }

  public static Identifiable findByPrimaryKey(Class cls, Serializable id, PersistenceContext pc) throws FindByPrimaryKeyException {

    if (id == null)
      throw new FindByPrimaryKeyException("HibernateFactory::findByPrimaryKey: passed id null.");

    //test whether id is of the rigth class
    if (id instanceof String) {
      if (HibernateUtilities.isIdInteger(cls))
        id = new Integer((String) id);
    }

    long timeToLoadObj = 0;
    java.lang.Object o = null;

    if (debugLoad) {
      timeToLoadObj = System.currentTimeMillis();
      Tracer.platformLogger.debug("HibernateScalarQueryCursorServer::findByPrimaryKey  before load" + timeToLoadObj);
    }

    o = pc.session.get(cls, id);

    if (debugLoad) {
      Tracer.platformLogger.debug("HibernateScalarQueryCursorServer::findByPrimaryKey time to load: " + (System.currentTimeMillis() - timeToLoadObj) / 1000);
    }

    return (Identifiable) o;
  }

  public static void addArgs(Object[] args, Query query) throws HibernateException {
    java.lang.Object arg;
    for (int i = 0; i < args.length; ++i) {
      arg = args[i];
      if (arg != null) {
        if (Double.class.isAssignableFrom(arg.getClass()))
          query.setDouble(i, (Double) arg);
        else if (Integer.class.isAssignableFrom(arg.getClass()))
          query.setInteger(i, (Integer) arg);
        else if (Date.class.isAssignableFrom(arg.getClass()))
          query.setTimestamp(i, (Date) arg);
        else if (Boolean.class.isAssignableFrom(arg.getClass()))
          query.setBoolean(i, (Boolean) arg);
        else if (String.class.isAssignableFrom(arg.getClass()))
          query.setString(i, (String) arg);
        else
          query.setParameter(i, arg);
      }
    }
  }

  public static void addArgs(List args, Query query) throws HibernateException {
    if (args != null) {
      Object[] objs = args.toArray();
      addArgs(objs, query);
    }
  }

  public static void initialize(Object o) throws HibernateException {
    Hibernate.initialize(o);
  }

  public static boolean isInitialized(Object o) {
    return Hibernate.isInitialized(o);
  }

  public static void update(Identifiable i) throws PersistenceException {
    try {
      getSession().update(i);
    } catch (HibernateException e) {
      throw new PersistenceException(e);
    }
  }

  public static void lock(Identifiable i, LockMode lockMode) throws PersistenceException {
    try {
      getSession().lock(i, lockMode);
    } catch (HibernateException e) {
      throw new PersistenceException(e);
    }
  }

  public static void attemptFlush() throws HibernateException {

    try {
      //HibernateTransaction t = (HibernateTransaction) HibernateTransaction.getCurrentTransaction();
      //if (t != null)
      //t.commit();
      HibernateFactory.getSession().flush();

    } catch (Throwable e) {
      HibernateTransaction t = (HibernateTransaction) HibernateTransaction.getCurrentTransaction();
      if (t != null)
        t.rollback();
      throw new HibernateException(e);
    }
  }

  /**
   * this is not managed by the filter
   */
  public static PersistenceContext newFreeSession() throws PersistenceException {

    PersistenceContext pc = new PersistenceContext();
    try {
      pc.session = getSessionFactory().openSession();
      pc.tx = pc.session.beginTransaction();

    } catch (HibernateException e) {
      throw new PersistenceException(e);
    }
    return pc;
  }

  public static PersistenceContext newFreeSession(Connection c) throws PersistenceException {

    PersistenceContext pc = new PersistenceContext();
    try {
      pc.session = getSessionFactory().openSession(c);
      pc.tx = pc.session.beginTransaction();

    } catch (HibernateException e) {
      throw new PersistenceException(e);
    }
    return pc;
  }


  public static void setSessionFactory(SessionFactory sessionFactory) {
    HibernateFactory.sessionFactory = sessionFactory;
  }

  public static PersistenceContext getDefaultPersistenceContext() {
    PersistenceContext pc = new PersistenceContext();
    pc.session = getSession();
    //pc.tx = (Transaction) HibernateTransaction.getCurrentTransaction();
    return pc;
  }

}

