package org.jblooming.persistence.hibernate;

import org.hibernate.Session;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.dialect.MySQLDialect;
import org.hibernate.mapping.Collection;
import org.hibernate.mapping.*;
import org.hibernate.tool.hbm2ddl.DatabaseMetadata;
import org.hibernate.tools.hbm2x.MyPOJOExporter;
import org.hibernate.type.IntegerType;
import org.hibernate.type.LongType;
import org.hibernate.type.Type;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.ontology.Identifiable;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.StoreException;
import org.jblooming.tracer.Tracer;
import org.jblooming.transaction.hibernate.HibernateTransaction;
import org.jblooming.uidgen.CounterHome;
import org.jblooming.utilities.*;
import org.jblooming.utilities.file.FileUtilities;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.*;
import java.util.List;
import java.util.Set;

/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 */
public class HibernateUtilities {

  public static String generateSchema(boolean analyze, boolean catchExceptions, boolean recreateSchema, HttpServletRequest request) {
    try {
      return generateSchema(analyze, catchExceptions, recreateSchema, null, null, false, request);
    } catch (Exception e) {
      throw new PlatformRuntimeException(e);
    }
  }

  public static String generateSchema(boolean analyze, boolean catchExceptions, boolean recreateSchema, String filterPrefix,
                                      HttpServletRequest request, HttpServletResponse response) {
    try {
      return generateSchema(analyze, catchExceptions, recreateSchema, null, filterPrefix, false, request);
    } catch (Exception e) {
      throw new PlatformRuntimeException(e);
    }
  }

  public static String generateSchema(boolean analyze, boolean catchExceptions, boolean recreateSchema, Method tableNameValidator, boolean feedBackOnInvalid,
                                      HttpServletRequest request
  ) throws Exception {
    return generateSchema(analyze, catchExceptions, recreateSchema, tableNameValidator, null, feedBackOnInvalid, request);
  }

  public static String generateSchema(boolean analyze, boolean catchExceptions, boolean recreateSchema, String filterPrefix, boolean feedBackOnInvalid,
                                      HttpServletRequest request) throws Exception {
    return generateSchema(analyze, catchExceptions, recreateSchema, null, filterPrefix, feedBackOnInvalid, request);
  }


  public static String generateSchema(
          boolean analyze,
          boolean catchExceptions,
          boolean recreateSchema,
          Method tableNameValidator,
          String filterPrefix,
          boolean feedBackOnInvalid,
          HttpServletRequest request

  ) throws Exception {
    return generateSchema(analyze,
            catchExceptions,
            recreateSchema,
            tableNameValidator,
            filterPrefix,
            feedBackOnInvalid,
            HttpUtilities.getFileSystemRootPathForRequest(request));
  }

  public static String generateSchema(
          boolean analyze,
          boolean catchExceptions,
          boolean recreateSchema,
          Method tableNameValidator,
          String filterPrefixes,
          boolean feedBackOnInvalid,
          String rootPath
  ) throws Exception {

    StringBuffer hist = new StringBuffer();
    hist.append("----------------------------------------------------------------------------------------------------------------\n");
    hist.append("Schema evolution: " + new Date() + "\n\n");

    String feedback = "feedback:<br>";

    Configuration store = HibernateFactory.config;
    Session s = HibernateFactory.getSession();
    Connection connection = s.connection();
    String[] createSQL;

    Dialect dialect = Dialect.getDialect(store.getProperties());
    if (recreateSchema)
      createSQL = store.generateSchemaCreationScript(dialect);
    else {
      DatabaseMetadata info = new DatabaseMetadata(connection, dialect);
      createSQL = store.generateSchemaUpdateScript(dialect, info);
    }

    Statement stmt = connection.createStatement();

    Set<String> launchedUpdates = new HashSet();

    boolean isMySql = dialect instanceof MySQLDialect;

    for (int i = 0; i < createSQL.length; i++) {

      String sql = createSQL[i];

      if (isMySql && sql.toLowerCase().indexOf("create table") > -1) {

        if (sql.indexOf("TYPE=MyISAM") > -1)
          sql = StringUtilities.replaceAllNoRegex(sql, "TYPE=MyISAM", "TYPE=InnoDB");
        else if (sql.toLowerCase().indexOf("innodb") == -1) {
          sql = (sql.endsWith(";") ? sql.substring(0, sql.length() - 1) : sql) + " TYPE=InnoDB" + (sql.endsWith(";") ? ";" : "");
        }

      }

      if (launchedUpdates.contains(sql.toLowerCase())) {
        feedback = feedback + "operation cancelled as already done:<br><font color='orange'><small >" + sql + "</small></font><br>";
        continue;
      }

      if (tableNameValidator != null && !((Boolean) tableNameValidator.invoke(tableNameValidator.getDeclaringClass().newInstance(), sql))) {
        if (feedBackOnInvalid)
          feedback = feedback + "operation cancelled:<br><font color='gray'><small >" + sql + "</small></font><br>";
        continue;
      }

      if (filterPrefixes != null) {
      List<String> filterPrefixesL = StringUtilities.splitToList(filterPrefixes, ",");
      boolean skip = false;
      for (String filterPrefix : filterPrefixesL) {
        if (filterPrefix != null && filterPrefix.trim().length() > 0 && (sql.indexOf(" " + filterPrefix) > -1 || sql.indexOf("." + filterPrefix) > -1)) {
          if (feedBackOnInvalid)
            feedback = feedback + "operation cancelled:<br><font color='gray'><small >" + sql + "</small></font><br>";
          skip=true;
          break;
        }
      }
      if (skip)
        continue;
      }

      feedback = feedback + sql + "<br>";

      if (!analyze) {
        if (catchExceptions) {
          try {
            stmt.executeUpdate(sql);
            HibernateTransaction.getCurrentTransaction().checkPoint();
          } catch (SQLException e) {
            feedback = feedback + "<font color='red'>" + JSP.encode(e.getMessage()) + "</font><br>";
            e.printStackTrace();
          }
        } else {
          //hack for indexes

          try {
            stmt.executeUpdate(sql);
          } catch (SQLException e) {
            if (sql.toLowerCase().indexOf("create index") > -1 || sql.toLowerCase().indexOf("create unique index") > -1)
              Tracer.hibernateLogger.warn("On this database index could not be created: " + sql);
            else {
              Tracer.hibernateLogger.error(sql, e);
              throw e;
            }
          }
        }
        HibernateTransaction.getCurrentTransaction().checkPoint();
      }
      launchedUpdates.add(sql.toLowerCase());
      hist.append(sql + "\n");

    }
    if (!analyze) {
      String logDirPath = rootPath + File.separator + "WEB-INF" + File.separator + "log";
      String log = logDirPath + File.separator + "schemaHistory.log";
      File logDir = new File(logDirPath);
      logDir.mkdirs();
      File file = new File(log);
      file.createNewFile();
      FileUtilities.appendToFile(log, hist.toString());
    }
    return feedback;
  }

  public static int getColumnLength(Class persClass, String property) {

    int length = 255;
    Column column = getColumn(persClass, property);
    if (column != null)
      length = column.getLength();
    return length;
  }

  public static Column getColumn(Class persClass, String property) {

    Column column = null;
    PersistentClass pc = getClassMapping(persClass);

    //int propColPos = 0;
    Iterator<Property> props = pc.getPropertyClosureIterator();
    while (props.hasNext()) {
      Property property1 = props.next();
      String name = property1.getName();
      if (name != null && name.equals(property)) {
        //todo: does NOT work if the column has a name different from property
        column = pc.getTable().getColumn(new Column(name)); // propColPos+1);
        break;
      }
      //propColPos++;
    }
    return column;
  }


  public static boolean isIdInteger(Class persClass) {

    PersistentClass pc = getClassMapping(persClass);
    //if (pc.getIdentifier() instanceof SimpleValue) {
    if (pc != null && pc.getIdentifier() instanceof SimpleValue) {
      SimpleValue sv = (SimpleValue) pc.getIdentifier();
      if (sv.getType() instanceof IntegerType)
        return true;
    }
    return false;
  }

  public static PersistentClass getClassMapping(Class persClass) {
    return getClassMapping(persClass.getName());
  }

  public static PersistentClass getClassMapping(String persClassName) {
    PersistentClass pc = HibernateFactory.config.getClassMapping(PersistenceHome.deProxy(persClassName));
    if (pc == null) {
      String name = "entityName_" + ReflectionUtilities.unqualify(persClassName);
      pc = HibernateFactory.config.getClassMapping(name);
    }
    return pc;
  }

  public static boolean isIdAssigned(Class persClass) {

    PersistentClass pc = getClassMapping(persClass);
    if (pc.getIdentifier() instanceof SimpleValue) {
      SimpleValue sv = (SimpleValue) pc.getIdentifier();
      return "assigned".equalsIgnoreCase(sv.getIdentifierGeneratorStrategy());
    } else
      return false;
  }

  public static String counterNameForIdAssigned(Class persClass) {

    String result = null;
    PersistentClass pc = getClassMapping(persClass);
    if (pc.getIdentifier() instanceof SimpleValue) {
      SimpleValue sv = (SimpleValue) pc.getIdentifier();
      if ("assigned".equalsIgnoreCase(sv.getIdentifierGeneratorStrategy())) {
        result = pc.getTable().getName();
      }
    }
    return result;
  }

  public static SimpleValue getIdentifierGenerator(Class persClass) {

    SimpleValue result = null;
    PersistentClass pc = getClassMapping(persClass);
    if (pc.getIdentifier() instanceof SimpleValue) {
      result = (SimpleValue) pc.getIdentifier();
    }
    return result;
  }

  public static void incrementIdIfNew(Identifiable i) throws StoreException {

    if ((PersistenceHome.NEW_EMPTY_ID.equals(i.getId())) || i.getId() == null) { //

      PersistentClass pc = getClassMapping(i.getClass());

      if (pc.getIdentifier() instanceof SimpleValue) {
        SimpleValue sv = (SimpleValue) pc.getIdentifier();
        if ("assigned".equalsIgnoreCase(sv.getIdentifierGeneratorStrategy())) {
          String counterName = pc.getTable().getName();

          if (sv.getType() instanceof IntegerType)
            //case int assigned
            i.setId(CounterHome.next(counterName));
          else if (sv.getType() instanceof LongType)
            //case int assigned
            i.setId((long) CounterHome.next(counterName));
          else
            //case serializable
            i.setId(CounterHome.nextSer(counterName));

        }
      }
    }
  }

  public static Type getIdType(Identifiable i) {
    return getIdType(i.getClass());
  }

  public static Type getIdType(Class clazz) {
    return getClassMapping(clazz).getIdentifier().getType();
  }

  public static void generateClasses(String outputDir, String dirWithMappings, Set<String> exportClasses) {

    File dir = new File(dirWithMappings);
    String[] mapFiles = dir.list();
    for (int i = 0; i < mapFiles.length; i++) { //mapFiles.length
      HibernateFactory.config.addFile(dirWithMappings + File.separator + mapFiles[i]);
    }
    MyPOJOExporter exporter = new MyPOJOExporter(HibernateFactory.config, new File(outputDir));
    exporter.start(false, false, exportClasses);

  }

  public static java.util.Map<String, Collection> getAllInversesOnTarget(Object target) {

    Object realDelendo = ReflectionUtilities.getUnderlyingObjectAsObject(target);
    java.util.Map<String, Collection> inverses = new HashTable<String, Collection>();
    Class realClass = realDelendo.getClass();
    String targetTable = HibernateUtilities.getClassMapping(realClass.getName()).getTable().getName();

    Iterator comMapsIt = HibernateFactory.config.getCollectionMappings();

    while (comMapsIt.hasNext()) {
      org.hibernate.mapping.Collection collection = (org.hibernate.mapping.Collection) comMapsIt.next();

      if (collection.isInverse() && collection.getOwner().getTable().getName().equals(targetTable))
        inverses.put(collection.getNodeName(), collection);
    }
    return inverses;
  }


  public static String getTableName(Class persistentClass) {
    PersistentClass classMapping = HibernateUtilities.getClassMapping(persistentClass.getName());

    if (classMapping != null)
      return classMapping.getTable().getName();
    else
      return null;
  }

  /**
   * Filters the src collection and puts the objects matching the
   * clazz into the dest collection.
   */
  public static <T> void filterHibernateCollection(Class<T> clazz,
                                                   java.util.Collection<?> src,
                                                   java.util.Collection<T> dest) {
    for (Object o : src) {
      o = ReflectionUtilities.getUnderlyingObject(o);
      if (clazz.isInstance(o)) {
        dest.add(clazz.cast(o));
      }
    }
  }

  /**
   * Filters the src collection and puts all matching objects into
   * an ArrayList, which is then returned.
   */
  public static <T> java.util.Collection<T> filterHibernateCollection(Class<T> clazz,
                                                                      java.util.Collection<?> src) {
    java.util.Collection<T> result = new ArrayList<T>();
    filterHibernateCollection(clazz, src, result);
    return result;
  }


}
