package org.jblooming.scheduler;

import org.jblooming.PlatformRuntimeException;
import org.jblooming.operator.Operator;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.FindException;
import org.jblooming.persistence.hibernate.HibernateFactory;
import org.jblooming.persistence.hibernate.PersistenceContext;
import org.jblooming.tracer.Tracer;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.Future;

/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 * @author Roberto Bicchierai rbicchierai@open-lab.com
 */

public class Scheduler extends TimerTask {

  public TreeSet<OrderedJob> toBeExecuted = new TreeSet<OrderedJob>();

  public TreeMap<Serializable, Future> inExecution = new TreeMap();

  public long tick;

  private static Scheduler scheduler;

  private Timer timer;

  public String instantiator;
  public Date instantiationTime;

  private Set<JobLogData> jobLogsToBeStored = new HashSet();

  public static final long DEFAULT_TICK = 600000;
  public static final int deltaForExcludingRelaunch = 30000;


  private Scheduler() {
  }

  public static void instantiate(long tick, Operator operator) {

    instantiate(tick, operator.getDisplayName());

  }

  public static void instantiate(long tick, String operatorName) {

    if (scheduler != null && scheduler.timer != null)
      return;
    //throw new PlatformRuntimeException("Scheduler and timer already instantiated");

    if (scheduler == null) {
      scheduler = new Scheduler();
    }

    scheduler.fillFromPersistence();

    scheduler.instantiator = operatorName;
    scheduler.instantiationTime = new Date();
    scheduler.tick = tick;
    scheduler.timer = new Timer(false);
    scheduler.timer.schedule(scheduler, new Date(), tick);

  }

  public static Scheduler getInstance() {
    return scheduler;
  }

  /**
   * tick: run due and refills toBeExecuted
   */
  public void run() {

    boolean exception = true;
    PersistenceContext pc = null;

    try {

      SortedSet<OrderedJob> toBeExecutedJobs = toBeExecuted.headSet(new OrderedJob(new Date().getTime() - tick + 1, -1));

      pc = HibernateFactory.newFreeSession();

      for (OrderedJob orderedJob : new TreeSet<OrderedJob>(toBeExecutedJobs)) {
        Serializable id = orderedJob.jobId;


        Job job = (Job) PersistenceHome.findByPrimaryKey(Job.class, id, pc);

        if (job.isEnabled()) {

          //is running ?
          if (inExecution.keySet().contains(job.getId())) {


            StringBuffer logMessage = new StringBuffer(job.getName());

            Future future = inExecution.get(job.getId());
            boolean stillRunning = !future.isDone();

            //check still running in timeout
            if (stillRunning) {
              if (job.getTimeoutTime() > 0 && System.currentTimeMillis() - job.getStartedOn() > job.getTimeoutTime()) {
                future.cancel(true);
                logMessage.append(": went in timeout, cancelled ");
              }
            }

            boolean failedAsJob = false;
            Iterator i = jobLogsToBeStored.iterator();
            while (i.hasNext()) {
              JobLogData jobLogData = (JobLogData) i.next();
              if (jobLogData.id.equals(job.getId()))
                failedAsJob = !jobLogData.successfull;
            }


            boolean failedAsThread = future.isCancelled();

            if (failedAsJob)
              logMessage.append(": job failed as job");
            if (failedAsThread)
              logMessage.append(": job failed as thread");

            //completed successfully as thread and as job
            if (!stillRunning && !failedAsJob) {

              inExecution.remove(job.getId());

              toBeExecuted.remove(orderedJob);

              //eventually put in relaunch queue

              Date dateAfter = job.getSchedule().getNextDateAfter(new Date(System.currentTimeMillis()));
              if (dateAfter != null) {
                toBeExecuted.add(new OrderedJob(dateAfter.getTime(), job.getId()));
              }
            }
            //failed
            if (!stillRunning && (failedAsJob || failedAsThread)) {

              if (!job.isOnErrorSuspendScheduling()) {
                if (!job.isOnErrorRetryNow()) {
                  inExecution.remove(job.getId());
                  toBeExecuted.remove(orderedJob);

                  Date dateAfter = job.getSchedule().getNextDateAfter(new Date(System.currentTimeMillis()));
                  if (dateAfter != null) {
                    toBeExecuted.add(new OrderedJob(dateAfter.getTime(), job.getId()));
                  logMessage.append(": job rescheduled");
                  }
                } else
                  //retry now case need not be handled, as job is not removed from toBeExecuted
                  logMessage.append(": job retried");

              } else {
                //suspended
                inExecution.remove(job.getId());
                toBeExecuted.remove(orderedJob);
              }

              if (failedAsThread) {                
                Tracer.jobLogger.error(logMessage.toString());
              }
            }

            // otherwise is still running -> do nothing

          } else {
            // this in order to avoid a second node (server) to launch exactly the same job for the same scheduled launch
            long lastExecutionTime = job.getLastExecutionTime();
            long ctm = System.currentTimeMillis();
            long delta = Math.abs(lastExecutionTime - ctm);
            //System.out.println(job.getName()+": lastExecutionTime="+ DateUtilities.dateToTimestampString(new Date(lastExecutionTime)) );
            if (delta > deltaForExcludingRelaunch) {
              job.setLastExecutionTime(System.currentTimeMillis());
              job.store(pc);
              runAJob(job);
            } else {
              Tracer.platformLogger.debug("Job "+job.getName()+" not launched as found launched by another node");
          }
        }
        }
      }

      //store the jobLogs
      while (true) {
        if (jobLogsToBeStored.size() > 0) {
          JobLogData jl = jobLogsToBeStored.iterator().next();
          Job job = (Job) PersistenceHome.findByPrimaryKey(Job.class, jl.id, pc);          
          long howLongPiuOMenTheTick = System.currentTimeMillis()-(job.getLastExecutionTime()+tick);
          howLongPiuOMenTheTick = howLongPiuOMenTheTick<0 ?  howLongPiuOMenTheTick+tick :   howLongPiuOMenTheTick;
          Tracer.jobLogger.info("job: " + job.getName() + " executed at " + jl.date + " successfull:" + jl.successfull + " duration:"+howLongPiuOMenTheTick+"ms.  notes: " + jl.notes);
          jobLogsToBeStored.remove(jl);
        } else
          break;
      }

      pc.commitAndClose();

      exception = false;

    } catch (Exception e) {
      //throw new PlatformRuntimeException(e);
      Tracer.platformLogger.error("Scheduler failed a run",e);

    } finally {
      try {
        if (pc != null && !pc.committedOrRollbacked)
          pc.rollback();
      } catch (Exception e) {
        exception = true;
        Tracer.platformLogger.error(e);
      }
      //if in any run of this there is an exception, scheduler should go down
      //10/06/2007 ma quanno mai!!!
      //if (exception)
      //  stop();
    }
  }

  public void runAJob(Job job) {
    Future future = PlatformExecutionService.executorService.submit(job);
    inExecution.put(job.getId(), future);
  }

  public void addJob(Job job) {
    fillFromPersistence();
    //scheduler.toBeExecuted.add(new OrderedJob(job.getSchedule().getNextFireTimeAfter(System.currentTimeMillis() - tick + 1), job.getIntId()));
  }

  public void removeJob(Job job) {
    scheduler.toBeExecuted.remove(new OrderedJob(0, job.getId()));
  }

  public synchronized void fillFromPersistence() {

    synchronized (this) {
      String hql = "from " + Job.class.getName() + " as job where job.schedule.end >= :now and job.enabled = :istrue";
      OqlQuery q = new OqlQuery(hql);
      q.getQuery().setTimestamp("now", new Date(System.currentTimeMillis()));
      q.getQuery().setBoolean("istrue", true);
      try {
        scheduler.toBeExecuted = new TreeSet<OrderedJob>();
        List<Job> persJobs = q.list();
        for (Job job : persJobs) {
          Date dateAfter = job.getSchedule().getNextDateAfter(new Date(System.currentTimeMillis()));
          if (dateAfter != null) {
            Serializable id = job.getId();
            scheduler.toBeExecuted.add(new OrderedJob(dateAfter.getTime(), id));
          }
        }
      } catch (FindException e) {
        throw new PlatformRuntimeException(e);
      }
    }
  }

  public static boolean isRunning() {

    boolean sched = scheduler != null;
    return sched && scheduler.timer != null;
  }

  public void stop() {
    if (scheduler != null) {
      scheduler.cancel();
      if (scheduler.timer != null)
        scheduler.timer.cancel();
      scheduler.timer = null;
      scheduler = null;
    }
  }

  public synchronized void addJobLogData(JobLogData jobLog) {
    jobLogsToBeStored.add(jobLog);
  }

  public class OrderedJob implements Comparable {

    public long exeTime;
    public Date exeTimeDate;
    public Serializable jobId;

    public OrderedJob(long exeTime, Serializable jobId) {
      this.exeTime = exeTime;
      this.exeTimeDate = new Date(exeTime);
      this.jobId = jobId;
    }

    public int compareTo(Object o) {
      int result = (new Long(exeTime).compareTo(new Long(((OrderedJob) o).exeTime)));
      if (result == 0 && !this.equals(o))
        result = -1;
      return result;
    }

    public boolean equals(Object o) {
      return jobId == ((OrderedJob) o).jobId;
    }

    public int hashCode() {
      return jobId.hashCode();
    }
  }
}
