package org.jblooming.utilities;

import org.jblooming.agenda.CompanyCalendar;
import org.jblooming.system.SystemConstants;
import org.jblooming.waf.SessionState;
import org.jblooming.waf.settings.ApplicationState;

import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * This class contains some methods to manage dates
 *
 * @author Roberto Bicchierai & Pietro Polsinelli
 * @since JDK 1.4
 *        <p/>
 *        it uses the standard date format:
 *        G  Era designator  Text  AD
 *        y  Year  Year  1996; 96
 *        M  Month in year  Month  July; Jul; 07
 *        width  Week in year  Number  27
 *        W  Week in month  Number  2
 *        D  Day in year  Number  189
 *        d  Day in month  Number  10
 *        F  Day of week in month  Number  2
 *        E  Day in week  Text  Tuesday; Tue
 *        a  Am/pm marker  Text  PM
 *        H  Hour in day (0-23)  Number  0
 *        k  Hour in day (1-24)  Number  24
 *        K  Hour in am/pm (0-11)  Number  0
 *        height  Hour in am/pm (1-12)  Number  12
 *        m  Minute in hour  Number  30
 *        s  Second in minute  Number  55
 *        S  Millisecond  Number  978
 *        z  Time zone  General time zone  Pacific Standard Time; PST; GMT-08:00
 *        Z  Time zone  RFC 822 time zone  -0800
 */
public class DateUtilities {

  public static final TimeZone DEFAULT_TIME_ZONE = DateFormat.getTimeInstance().getTimeZone();

  public static Locale DEFAULT_LOCALE = Locale.US;
  private static final String DATE_DEFAULT_FORMAT = "dd/MM/yyyy";
  public static String DAY_SHORT_CODE="d.";

  /**
   * Method dateToString
   * returns a String representing the passed date in the default format.
   *
   * @param date a  Date
   * @return a String
   */
  public static String dateToString(Date date) {
    return dateToString(date, null, null);
  }

  public static String dateToString(Date date, SessionState sm) {
    return dateToString(date, null, sm.getCurrentLocale());
  }


  public static String dateToString(Date date, Locale currentLocale) {
    return dateToString(date, null, currentLocale);
  }

  public static String dateToString(Date date, String format) {
    return dateToString(date, format, DEFAULT_LOCALE);
  }

  /**
   * Method dateToString
   * returns a String representing the passed date. The format is used to parse.
   *
   * @param date   a  Date
   * @param format a  String if null the deafault is used
   * @return a String
   */
  public static String dateToString(Date date, String format, Locale locale) {
    if (date == null)
      return "";
    if (locale == null)
      locale = DEFAULT_LOCALE;
    if (format == null)
      format = getDateFormat();
    format = setFormat(format);
    SimpleDateFormat sdf = new SimpleDateFormat(format, locale);
    return sdf.format(date);
  }

  public static String dateToTimestampString(Date date) {
    if (date == null)
      return "";
    SimpleDateFormat sdf = new SimpleDateFormat(getTimestampFormat());
    return sdf.format(date);
  }

  public static String dateToTimestampString(Date date, SessionState sm) {
    if (date == null)
      return "";
    SimpleDateFormat sdf = new SimpleDateFormat(getTimestampFormat(), sm.getCurrentLocale());
    return sdf.format(date);
  }

  public static String dateToHourMinutes(Date date) {
    return dateToString(date, getHourFormat());
  }

  /**
   * @param s
   * @return the correct date if the format is valid, else new Date();
   */
  public static Date friendlyDateFromString(String s) {
    try {
      return dateFromString(s, setFormat(null));
    } catch (ParseException e) {
      return new Date();
    }
  }


  public static Date dateFromString(String s) throws ParseException {
    return dateFromString(s, setFormat(null));
  }

  public static Date dateFromUrl(String s) throws ParseException {
    SimpleDateFormat sdf = new SimpleDateFormat(getTimestampFormat(), DEFAULT_LOCALE);
    sdf.setLenient(true);
    return sdf.parse(s);
  }

  public static String dateToUrl(Date date) {
    SimpleDateFormat sdf = new SimpleDateFormat(getTimestampFormat());
    return sdf.format(date);
  }

  /**
   * Method dateFromString
   * returns a date parsing the passed string in the default format.
   *
   * @param s a  String
   * @return a Date
   * @throws ParseException
   */
  public static Date dateFromString(String s, String format) throws ParseException {
    return dateFromString(s, format, null, null);
  }


  /**
   * Method dateFromString
   * returns a date parsing the passed string. The passed format is used to parse;
   * if null the default is used.
   *
   * @param dateString a  String
   * @param format     a  String if null the deafault is used
   * @return a Date
   * @throws ParseException
   */
  public static Date dateFromString(String dateString, String format, Locale fromLocale, TimeZone fromTimeZone) throws ParseException {
    return dateFromString(dateString, format, fromLocale, fromTimeZone, true);
  }


  public static enum shortDates {
    LASTQUARTERSTART,
    LASTMONTHSTART,
    THISMONTHSTART,
    LASTWEEKSTART,
    THISWEEKSTART,
    YESTERDAY,
    TODAY,
    NOW,
    TOMORROW,
    NEXTWEEKSTART,
    NEXTMONTHSTART,
    THISQUARTERSTART,
    NEXTQUARTERSTART
  }

  public static Map<shortDates, List<String>> shortDatesAliases = getShortDatesAliases();

  private static Map<shortDates, List<String>> getShortDatesAliases() {
    Map<shortDates, List<String>> sas = new HashTable();

    sas.put(shortDates.NOW,CollectionUtilities.toList("N",shortDates.NOW.toString()));

    sas.put(shortDates.TODAY,CollectionUtilities.toList("T",shortDates.TODAY.toString()));
    sas.put(shortDates.YESTERDAY,CollectionUtilities.toList("Y",shortDates.YESTERDAY.toString()));
    sas.put(shortDates.TOMORROW,CollectionUtilities.toList("TM",shortDates.TOMORROW.toString()));

    sas.put(shortDates.THISWEEKSTART,CollectionUtilities.toList("W","WEEK","THISWEEK","WEEKSTART",shortDates.THISWEEKSTART.toString()));
    sas.put(shortDates.LASTWEEKSTART,CollectionUtilities.toList("LW","LASTWEEK",shortDates.LASTWEEKSTART.toString()));
    sas.put(shortDates.NEXTWEEKSTART,CollectionUtilities.toList("NW","NEXTWEEK",shortDates.NEXTWEEKSTART.toString()));

    sas.put(shortDates.THISMONTHSTART,CollectionUtilities.toList("M","MONTH","THISMONTH","MONTHSTART",shortDates.THISMONTHSTART.toString()));
    sas.put(shortDates.LASTMONTHSTART,CollectionUtilities.toList("LM","LASTMONTH",shortDates.LASTMONTHSTART.toString()));
    sas.put(shortDates.NEXTMONTHSTART,CollectionUtilities.toList("NM","NEXTMONTH",shortDates.NEXTMONTHSTART.toString()));

    sas.put(shortDates.THISQUARTERSTART,CollectionUtilities.toList("Q","QUARTER","THISQUARTER","QUARTERSTART",shortDates.THISQUARTERSTART.toString()));
    sas.put(shortDates.LASTQUARTERSTART,CollectionUtilities.toList("LQ","LASTQUARTER",shortDates.LASTQUARTERSTART.toString()));
    sas.put(shortDates.NEXTQUARTERSTART,CollectionUtilities.toList("NQ","NEXTQUARTER",shortDates.NEXTQUARTERSTART.toString()));

    return sas;
  }

  public static Date dateFromString(String dateString, String format, Locale fromLocale, TimeZone fromTimeZone, boolean lenient) throws ParseException {

    if (fromLocale == null) fromLocale = DEFAULT_LOCALE;

    //if (fromTimeZone == null) fromTimeZone = DEFAULT_TIME_ZONE;

    if (dateString == null)
      return null;

    format = setFormat(format);

    //todo: i18n with default locale
    CompanyCalendar cc = new CompanyCalendar();
    String dateUpper = dateString.toUpperCase();
    if (shortDatesAliases.get(shortDates.NOW).contains(dateUpper)) {
      return new Date();

    } else if (shortDatesAliases.get(shortDates.TODAY).contains(dateUpper)) {
      return new CompanyCalendar().setAndGetTimeToDayStart();
    } else if (shortDatesAliases.get(shortDates.YESTERDAY).contains(dateUpper)) {
      return new Date(System.currentTimeMillis() - CompanyCalendar.MILLIS_IN_DAY);
    } else if (shortDatesAliases.get(shortDates.TOMORROW).contains(dateUpper)) {
      return new Date(System.currentTimeMillis() + CompanyCalendar.MILLIS_IN_DAY);

    } else if (shortDatesAliases.get(shortDates.THISWEEKSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_WEEK, cc.getFirstDayOfWeek());
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.LASTWEEKSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_WEEK, cc.getFirstDayOfWeek());
      cc.add(CompanyCalendar.WEEK_OF_YEAR, -1);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.NEXTWEEKSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_WEEK, cc.getFirstDayOfWeek());
      cc.add(CompanyCalendar.WEEK_OF_YEAR, 1);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.THISMONTHSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.LASTMONTHSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      cc.add(CompanyCalendar.MONTH, -1);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.NEXTMONTHSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      cc.add(CompanyCalendar.MONTH, 1);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.THISQUARTERSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      cc.set(CompanyCalendar.MONTH, ((int) (cc.get(CompanyCalendar.MONTH) / 3))*3 );
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.LASTQUARTERSTART).contains(dateUpper)) {
      cc.setAndGetTimeToDayStart();
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      cc.set(CompanyCalendar.MONTH, ((int) (cc.get(CompanyCalendar.MONTH) / 3))*3);
      cc.add(CompanyCalendar.MONTH, -3);
      return cc.getTime();

    } else if (shortDatesAliases.get(shortDates.NEXTQUARTERSTART).contains(dateUpper)) {
      cc.set(CompanyCalendar.DAY_OF_MONTH, 1);
      cc.set(CompanyCalendar.MONTH, ((int) (cc.get(CompanyCalendar.MONTH) / 3))*3);
      cc.add(CompanyCalendar.MONTH, 3);
      return cc.getTime();

    }


    SimpleDateFormat sdf = new SimpleDateFormat(format, fromLocale);
    sdf.setLenient(lenient);

    cc.setTime(sdf.parse(dateString));
    int y = cc.get(CompanyCalendar.YEAR);
    if (y < 49)
      cc.add(CompanyCalendar.YEAR, 2000);
    else if (y < 100)
      cc.add(CompanyCalendar.YEAR, 1900);

    return cc.getTime();
  }

  /**
   * This is a date and calendar free method, happily done by Pietro
   *
   * @param hoursMinutes assumed in the form 8:45 or 08:45
   * @deprecated use millisFromHourMinuteSmart
   */
  public static long millisFromHourMinute(String hoursMinutes) {
    long result = 0;
    try {
      result = millisFromHourMinuteSmart(hoursMinutes);
    } catch (NumberFormatException e) {
    }
    return result;

  }

  /**
   * @param value
   * @return a long value representing in millisecond the string. All this format are valid: "12:58" "13.75"  "63635676000" (this is already in milliseconds)
   * @throws NumberFormatException
   */
  public static long millisFromHourMinuteSmart(String value) throws NumberFormatException {
    long result = 0;
    if (value == null)
      throw new NumberFormatException("Null value not allowed");


    int semiColSeparator = value.indexOf(":");
    int dotSeparator = value.indexOf(".");

    if (semiColSeparator<0 && dotSeparator<0 && value.length() > 5) {
      return Long.parseLong(value);
    } else {


      if (dotSeparator > -1) {

        double d = Double.parseDouble(value);
        result = (long) (d * CompanyCalendar.MILLIS_IN_HOUR);

      } else {

        long hour = 0;
        long minute = 0;
        if (semiColSeparator == -1)
          hour = Long.parseLong(value);
        else {
          hour = Long.parseLong(value.substring(0, semiColSeparator));
          minute = Long.parseLong(value.substring(semiColSeparator + 1));
        }
        result = hour * CompanyCalendar.MILLIS_IN_HOUR + minute * CompanyCalendar.MILLIS_IN_MINUTE;
      }
      return result;
    }
  }


  public static Date getAbsoluteNow() {
    //int offSet = DEFAULT_TIME_ZONE.getOffset(System.currentTimeMillis());

    //return (new Date(System.currentTimeMillis() - offSet));
    return new Date();
  }

  public static Date getRelativeDate(Date absoluteDate) {
    int offSet = DEFAULT_TIME_ZONE.getOffset(System.currentTimeMillis());

    return (new Date(absoluteDate.getTime() + offSet));

  }

  public static Date getAbsoluteInterval(long time) {
    int offSet = DEFAULT_TIME_ZONE.getOffset(time);

    return (new Date(time - offSet));
  }


  public static String getJSFormat() {
    return getDateFormat();
  }

  public static Date dateFromStringJS(String date) throws ParseException {
    return dateFromString(date, getJSFormat());
  }

  public static String dateToStringJS(Date date) {
    return dateToString(date, getJSFormat());
  }

  public static String getReportFormattedInterval(long interval) {
    long minutes = (interval / 60000) % 60;
    long hours = (interval / 3600000);
    return hours + ":" + (minutes < 10 ? "0" : "") + minutes;
  }

  private static String setFormat(String format) {
    if (format == null || format.length() <= 0)
      format = ApplicationState.getApplicationSetting(SystemConstants.DEFAULT_DATE_FORMAT);
    if (format == null || format.length() <= 0)
      format = DATE_DEFAULT_FORMAT;
    return format;
  }

  public static String getDateFormat() {
    return ApplicationState.getApplicationSetting(SystemConstants.DEFAULT_DATE_FORMAT);
  }

  public static String getHourFormat() {
    return ApplicationState.getApplicationSetting(SystemConstants.DEFAULT_HOUR_FORMAT);
  }

  public static String getTimestampFormat() {
    return ApplicationState.getApplicationSetting(SystemConstants.DEFAULT_TIMESTAMP_FORMAT);
  }


  public static int getCurrentYear() {
    return CompanyCalendar.getInstance().get(CompanyCalendar.YEAR);
  }

  public static String timeToDisplayString(long l) {
    return new SimpleDateFormat(getHourFormat()).format(new Date(l));
  }


  public static String getMillisInHoursMinutes(Number millis) {
    int hour = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_HOUR);
    int min = (int) ((millis.longValue() % CompanyCalendar.MILLIS_IN_HOUR) / CompanyCalendar.MILLIS_IN_MINUTE);
    min = Math.abs(min);
    DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance();
    df.applyPattern("00");
    return hour + ":" + df.format(min);
  }

  public static String getMillisInHoursMinutesSeconds(Number millis) {
    int hour = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_HOUR);
    long dif = (millis.longValue() % CompanyCalendar.MILLIS_IN_HOUR);
    int min = (int) (dif / CompanyCalendar.MILLIS_IN_MINUTE);
    int sec = (int) ((dif % CompanyCalendar.MILLIS_IN_MINUTE) / 1000);
    DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance();
    df.applyPattern("00");
    return hour + ":" + df.format(min) + ":" + df.format(sec);
  }

  public static String getMillisInHours(Number millis) {
    int hour = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_HOUR);
    return hour + "";
  }

  public static String getMillisInDays(Number millis) {
    int days = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_DAY);
    return days + "";
  }

  public static String getMillisInDaysHoursMinutes(Number millis) {

    int days = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_DAY);
    int hour = (int) ((millis.longValue() % CompanyCalendar.MILLIS_IN_DAY) / CompanyCalendar.MILLIS_IN_HOUR);
    int min = (int) ((millis.longValue() % CompanyCalendar.MILLIS_IN_HOUR) / CompanyCalendar.MILLIS_IN_MINUTE);
    min = Math.abs(min);
    DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance();
    df.applyPattern("00");
    return (days > 0 ? days + "D  " : "") + hour + ":" + df.format(min);
  }


  public static String getMillisInDaysWorkHoursMinutes(Number millis) {

    int days = (int) (millis.longValue() / CompanyCalendar.MILLIS_IN_WORKING_DAY);
    int hour = (int) ((millis.longValue() % CompanyCalendar.MILLIS_IN_WORKING_DAY) / CompanyCalendar.MILLIS_IN_HOUR);
    int min = (int) ((millis.longValue() % CompanyCalendar.MILLIS_IN_HOUR) / CompanyCalendar.MILLIS_IN_MINUTE);
    min = Math.abs(min);
    DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance();
    df.applyPattern("00");

    return days + DAY_SHORT_CODE + hour + ":" + df.format(min);
  }


  public static String getDurationFromMillis(Long duration) {
    return NumberUtilities.decimal(duration.doubleValue() / CompanyCalendar.MILLIS_IN_DAY);
  }
}

