package org.jblooming.waf;

import org.jblooming.PlatformRuntimeException;
import org.jblooming.agenda.CompanyCalendar;
import org.jblooming.ontology.Identifiable;
import org.jblooming.ontology.PersistentFile;
import org.jblooming.ontology.PersistentText;
import org.jblooming.persistence.exceptions.FindByPrimaryKeyException;
import org.jblooming.persistence.hibernate.HibernateFactory;
import org.jblooming.utilities.ReflectionUtilities;
import org.jblooming.waf.exceptions.ActionException;
import org.jblooming.waf.html.input.Uploader;
import org.jblooming.waf.view.ClientEntry;
import org.jblooming.waf.view.PageState;

import java.lang.reflect.Field;
import java.text.ParseException;
import java.util.Date;


/**
 * If ce.value arrives empty and is not required, the value is reset
 * <p/>
 * if you want to check whether the update was attempted, check the return value
 * <p/>
 * in case of error, the client businessLogic error code is reset with the current value
 *
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 * @author Roberto Bicchierai rbicchierai@open-lab.com
 */
public class ActionUtilities {


  public static boolean setPersistentFile(ClientEntry ce, Identifiable mainObject, String propertyName,
                                          String persistentFileType, String fileDir, PageState pageState) {

    boolean ok = true;
    try {
      ce.stringValue();
      try {

        Field field = ReflectionUtilities.getField(propertyName, ReflectionUtilities.getUnderlyingObjectClass(mainObject));
        if (field == null)
          throw new PlatformRuntimeException("propertyName '" + propertyName + "' not found on " + mainObject.getClass());
        field.setAccessible(true);
        PersistentFile persistentFile = (PersistentFile) field.get(mainObject);
        if (persistentFile == null) {
          persistentFile = new PersistentFile(0, null, persistentFileType);
          persistentFile.fileDir = fileDir;
        }

        ReflectionUtilities.setFieldValue(propertyName, mainObject, Uploader.save(mainObject, persistentFile, ce.name, pageState));

      } catch (Exception e) {
        throw new PlatformRuntimeException(e);
      }

    } catch (ActionException e) {
      ok = false;
    }
    return ok;

  }

  public static boolean setString(ClientEntry ce, Object mainObject, String propertyName) {

    String value = null;
    boolean ok = true;

    try {
      value = ce.stringValue();
      setFieldValue(propertyName, mainObject, value);

    } catch (ActionException e) {
      ok = false;
    }
    return ok;
  }

  public static boolean setDate(ClientEntry ce, Object mainObject, String propertyName) {

    Date date = null;
    boolean ok = true;

    try {
      date = ce.dateValue();
      setFieldValue(propertyName, mainObject, date);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }

  public static boolean setBoolean(ClientEntry ce, Object mainObject, String propertyName) {
    boolean value;
    boolean ok = true;
    value = ce.checkFieldValue();
    setFieldValue(propertyName, mainObject, value);
    return ok;
  }

  public static boolean setInt(ClientEntry ce, Object mainObject, String propertyName) {
    int value;

    boolean ok = true;

    try {
      value = ce.intValue();
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }


  public static boolean setDouble(ClientEntry ce, Object mainObject, String propertyName) {

    double value;
    boolean ok = true;

    try {
      value = ce.doubleValue();
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }

  public static boolean setLong(ClientEntry ce, Object mainObject, String propertyName) {

    long value;
    boolean ok = true;

    try {
      value = ce.longValue();
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;

  }

  public static boolean setTime(ClientEntry ce, Object mainObject, String propertyName) {
    long value;
    boolean ok = true;

    try {
      value = ce.timeValueInMillis();
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }

  private static void setFieldValue(String propertyName, Object mainObject, Object value) {
    try {
      ReflectionUtilities.setFieldValue(propertyName, mainObject, value);
    } catch (Exception e) {
      throw new PlatformRuntimeException(e);
    }
  }

//   public static boolean setObject(ClientEntry ce, Object mainObject, String propertyName) throws FindByPrimaryKeyException {
//   }

  public static boolean setCurrency(ClientEntry ce, Object mainObject, String propertyName) {
    double value;
    boolean ok = true;

    try {
      value = ce.currencyValue();
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }

  public static boolean setDuration(ClientEntry ce, Object mainObject, String propertyName) {
    long value;
    boolean ok = true;

    try {
      //todo: parse formats W,D
      value = (long) (ce.doubleValue() * CompanyCalendar.MILLIS_IN_DAY);
      setFieldValue(propertyName, mainObject, value);
    } catch (ActionException e) {
      ok = false;
    } catch (ParseException e) {
      ok = false;
    }
    return ok;
  }

  public static boolean setIdentifiable(ClientEntry ce, Object mainObject, String propertyName) throws FindByPrimaryKeyException {

    Object value = null;
    boolean ok = true;

    try {
      String entityId = ce.stringValue();
      if (entityId != null && entityId.trim().length() > 0) {
        Field field = ReflectionUtilities.getField(propertyName, ReflectionUtilities.getUnderlyingObjectClass(mainObject));
        if (field == null)
          throw new PlatformRuntimeException("propertyName '" + propertyName + "' not found on " + mainObject.getClass());
        value = HibernateFactory.findByPrimaryKey(field.getType(), entityId);
      }

      try {
        ReflectionUtilities.setFieldValue(propertyName, mainObject, value);
      } catch (Exception e) {
        throw new PlatformRuntimeException(e);
      }

    } catch (ActionException e) {
      ok = false;
    }

    return ok;
  }

  public static boolean setText(ClientEntry ce, Object mainObject, String propertyName) {

    Object value = null;
    boolean ok = true;

    try {
      String text = ce.stringValue();

      Field field = ReflectionUtilities.getField(propertyName, ReflectionUtilities.getUnderlyingObjectClass(mainObject));
      if (field == null)
        throw new PlatformRuntimeException("propertyName '" + propertyName + "' not found on " + mainObject.getClass());
      field.setAccessible(true);
      PersistentText pt = (PersistentText) field.get(mainObject);
      if (pt == null)
        pt = new PersistentText();

      pt.setText(text);
      try {
        ReflectionUtilities.setFieldValue(propertyName, mainObject, pt);
      } catch (Exception e) {
        throw new PlatformRuntimeException(e);
      }
    } catch (IllegalAccessException e1) {
      throw new PlatformRuntimeException(e1);
    } catch (ActionException e1) {
      throw new PlatformRuntimeException(e1);
    }
    return ok;

  }
}


