package org.jblooming.waf.settings;

import com.Ostermiller.util.BadDelimiterException;
import com.Ostermiller.util.CSVParser;
import com.Ostermiller.util.CSVPrinter;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.tracer.Tracer;
import org.jblooming.utilities.HttpUtilities;
import org.jblooming.utilities.file.FileUtilities;
import org.jblooming.waf.constants.Commands;
import org.jblooming.waf.constants.Fields;
import org.jblooming.waf.constants.OperatorConstants;
import org.jblooming.waf.html.display.Img;
import org.jblooming.waf.view.PageSeed;
import org.jblooming.waf.view.PageState;

import javax.servlet.http.HttpServletRequest;
import java.io.*;
import java.nio.charset.Charset;
import java.util.*;

/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 */
public class I18n {

  private final static String SEPARATOR = "_-_";

  // edit status
  public final static String EDIT_STATUS_EDIT = "EDIT";
  public final static String EDIT_STATUS_READ = "READ";
  public final static String EDIT_STATUS_APPEND = "APPEND";
  public static final String EDIT_STATUS_EDIT_POPUP = "EDITPOPUP";

  // lenient level
  public static final int LENIENT_APP_LANG = 4;
  public static final int LENIENT_APP = 3;
  public static final int LENIENT_LANG = 2;
  public static final int LENIENT_NONE = 1;


  public SortedSet<String> supportedLanguages = new TreeSet<String>();

  public TreeMap<String, I18nEntry> codeEntries = new TreeMap();

  protected int lenient = LENIENT_APP_LANG;
  private String editStatus = EDIT_STATUS_READ;

  private Serializable i18nEditingOperatorId;


  // commands
  public static final String CMD_CHANGEMODALITY = "CMD_CHMDLT";
  public static final String CMD_RELOAD = "CMD_REL";
  public static final String CMD_DUMP = "CMD_DU";
  public static final String CMD_NEW_LANGUAGE = "CMD_NL";
  public static final String CMD_STORE_LABEL = "CMD_STLBL";
  public static final String CMD_NEW_ENTRY = "CMD_NEW_ENTRY";


  public SortedMap<String, I18nEntry> getEntriesForApplication(String applicationName) {
    TreeMap<String, I18nEntry> efa = new TreeMap();
    for (String code : codeEntries.keySet()) {
      final I18nEntry i18nEntry = codeEntries.get(code);
      final ApplicationEntry applicationEntry = (ApplicationEntry) i18nEntry.applicationEntries.get(applicationName);
      if (applicationEntry != null)
        efa.put(code, i18nEntry);
    }
    return efa;
  }

  public I18nEntry addEntry(String code, String application, String language, String value) {
    supportedLanguages.add(language.toUpperCase());
    I18nEntry ie = codeEntries.get(code);
    if (ie == null) {
      ie = new I18nEntry(code, application, language, value);
    } else {

      ApplicationEntry ae = ie.applicationEntries.get(application);
      if (ae == null)
        ae = new ApplicationEntry(application);
      ie.applicationEntries.put(application, ae);
      ae.addEntry(language, value);
    }
    codeEntries.put(code, ie);
    return ie;
  }

  public void removeEntry(String code, String application, String language) {
    if (getRawLabel(code, application, language) != null) {
      I18nEntry i18nEntry = (I18nEntry) codeEntries.get(code);
      ApplicationEntry ae = (ApplicationEntry) i18nEntry.applicationEntries.get(application);
      ae.entries.remove(language);
      if (ae.entries.size() <= 0)
        i18nEntry.applicationEntries.remove(application);
      if (i18nEntry.applicationEntries.size() <= 0)
        codeEntries.remove(code);
    }
  }

  /**
   * @deprecated use persistI18nByApplicationAndLanguage
   */
  public void persistI18nByApplication(Application application, File file) throws IOException {

    final String applicationName = application != null ? application.getName() : "";
    SortedMap i18nEntriesForApplication = getEntriesForApplication(applicationName);

    file.getParentFile().mkdirs();
    FileOutputStream fos = new FileOutputStream(file);

    Charset charset = Charset.forName("UTF-8");

    OutputStreamWriter osw = new OutputStreamWriter(fos, charset);
    CSVPrinter cvsp = new CSVPrinter(osw);

    try {
      cvsp.changeDelimiter('\t');
    } catch (BadDelimiterException e) {
      throw new PlatformRuntimeException(e);
    }

    if (i18nEntriesForApplication != null && i18nEntriesForApplication.size() > 0) {
      for (Iterator iterator = i18nEntriesForApplication.keySet().iterator(); iterator.hasNext();) {
        String code = (String) iterator.next();
        I18nEntry ie = (I18nEntry) i18nEntriesForApplication.get(code);
        ApplicationEntry ae = (ApplicationEntry) ie.applicationEntries.get(applicationName);
        if (ae != null) {
          for (Iterator itl = ae.entries.keySet().iterator(); itl.hasNext();) {
            String language = (String) itl.next();
            String value = (String) ae.entries.get(language);
            if (value != null) {
              cvsp.print(applicationName);
              cvsp.print(ie.getCode());
              cvsp.print(language);
              cvsp.println(value);
            }
          }
        }
      }
    }

    osw.close();
    fos.close();
  }

  public void persistI18nByApplicationAndLanguage(Application application, File file) throws IOException {

    final String applicationName = application != null ? application.getName() : "";
    SortedMap i18nEntriesForApplication = getEntriesForApplication(applicationName);


    for (String language : ApplicationState.i18n.supportedLanguages) {

      File fileForLang = new File(FileUtilities.getNameWithoutExt(file.getPath()) + "." + language + ".i18n");

      fileForLang.getParentFile().mkdirs();
      FileOutputStream fos = new FileOutputStream(fileForLang);

      Charset charset = Charset.forName("UTF-8");

      OutputStreamWriter osw = new OutputStreamWriter(fos, charset);
      CSVPrinter cvsp = new CSVPrinter(osw);

      try {
        cvsp.changeDelimiter('\t');
      } catch (BadDelimiterException e) {
        throw new PlatformRuntimeException(e);
      }

      if (i18nEntriesForApplication != null && i18nEntriesForApplication.size() > 0) {
        for (Iterator iterator = i18nEntriesForApplication.keySet().iterator(); iterator.hasNext();) {
          String code = (String) iterator.next();
          I18nEntry ie = (I18nEntry) i18nEntriesForApplication.get(code);
          ApplicationEntry ae = (ApplicationEntry) ie.applicationEntries.get(applicationName);
          if (ae != null) {
            String value = (String) ae.entries.get(language);
            if (value!=null) {
            cvsp.print(applicationName);
            cvsp.print(ie.getCode());
            cvsp.print(language);
            cvsp.println(value);
            }
          }

        }
      }
      osw.close();
      fos.close();
    }

  }


  public void loadI18n(File file) throws IOException {
    FileInputStream fos = new FileInputStream(file);
    InputStreamReader isr = new InputStreamReader(fos, Charset.forName("UTF-8"));
    CSVParser cvsr = new CSVParser(isr);
    try {
      cvsr.changeDelimiter('\t');
      cvsr.setEscapes("nrf", "\n\r\f");
    } catch (BadDelimiterException e) {
      throw new PlatformRuntimeException(e);
    }

    String values[][] = cvsr.getAllValues();
    if (values != null) {
      for (int i = 0; i < values.length; i++) {
        String[] value = values[i];
        if (value == null || value.length != 4)
          Tracer.platformLogger.warn("open lab platform - invalid row " + i + " for file " + file.getName());
        else
          ApplicationState.i18n.addEntry(value[1], value[0], value[2].toUpperCase(), value[3]);
      }
    }
    isr.close();
    fos.close();
  }

  /**
   * @param code
   * @param application
   * @param language
   * @return the label related to the "code". Can be exact or the most similar in relation to the lenient level
   */
  public String getLabel(String code, String application, String language, PageState pageState) {
    // transform two part language as EN_GB to EN only
    if (language.indexOf("_") > 1)
      language = language.substring(0, language.indexOf("_"));

    I18nEntry i18nEntry = (I18nEntry) codeEntries.get(code);
    String label = null;

    if (i18nEntry != null) // record the seen
      i18nEntry.setSeen(true);

    if (i18nEntry != null && !i18nEntry.isSuspect()) // the suspected one are omitted
      label = i18nEntry.getLabel(application, language);

    // handling for "append" mode
    if (label == null && EDIT_STATUS_APPEND.equals(getEditStatus())) {
      I18nEntry i18ne = addEntry(code, application, language, "missing_" + code);
      i18ne.setSuspect(true);
    }

    if (label == null)
      label = code;

    // EDIT MODE HANDLING && EDITING OPERATOR ONLY
    try {
      if ((EDIT_STATUS_EDIT.equals(editStatus) || EDIT_STATUS_EDIT_POPUP.equals(editStatus)) &&
        pageState.sessionState.getLoggedOperator() != null &&
        pageState.sessionState.getLoggedOperator().getId().equals(getI18nEditingOperatorId())) {
        String id = "lb" + code.hashCode();
        Img img = new Img(pageState.getSkin().imgPath + "i18n.gif", "i18n");

        if (EDIT_STATUS_EDIT.equals(editStatus)) {
          img.script = "onClick=\"event.cancelBubble = true;editLabel('" + id + "','" + code + "','" + application + "');\"";
        } else {
          PageSeed edit = new PageSeed(ApplicationState.contextPath + "/commons/administration/i18nEdit.jsp");

          if (pageState.getEntry(application).stringValueNullIfEmpty() != null)
            edit.addClientEntry(Fields.APPLICATION_NAME, pageState.getEntry(Fields.APPLICATION_NAME).stringValueNullIfEmpty());
          edit.setCommand(Commands.EDIT);
          edit.addClientEntry(Fields.FORM_PREFIX + "code", code);
          img.script = "onClick=\"event.cancelBubble = true; width=window.open('" + edit.toLinkToHref() + "','i18nwindow','height=500,width=650,scrollbars=yes'); width.focus();\"";
        }

        img.toolTip = code + " " + pageState.getApplication().getName() + " " + pageState.sessionState.getCurrentLocale().getLanguage();
        label = "<span id=\"" + id + "\">" + label + "</span>" + img.toHtmlStringBuffer(null);
      }
    } catch (PersistenceException e) {
    }

    return label;
  }

  public String getLabelNoPageState(String code, String application, String language) {
    // transform two part language as EN_GB to EN only
    if (language.indexOf("_") > 1)
      language = language.substring(0, language.indexOf("_"));

    String label = null;
    I18nEntry i18nEntry = (I18nEntry) codeEntries.get(code);
    if (i18nEntry != null)
      label = i18nEntry.getLabel(application, language);
    if (label == null)
      label = code;
    return label;
  }

  /**
   * @param code
   * @param code
   * @param language
   * @return the exact matching label. null if not found
   */
  public String getRawLabel(String code, String application, String language) {
    I18nEntry i18nEntry = (I18nEntry) codeEntries.get(code);
    String label = null;
    if (i18nEntry != null) {
      ApplicationEntry applicationEntry = (ApplicationEntry) i18nEntry.applicationEntries.get(application);
      if (applicationEntry != null) {
        label = (String) applicationEntry.entries.get(language);
      }
    }
    return label;
  }


  public String getEditStatus() {
    return editStatus;
  }

  public void setEditStatus(String editStatus) {
    this.editStatus = editStatus;
  }

  public int getLenient() {
    return lenient;
  }

  public void setLenient(int lenient) {
    this.lenient = lenient;
  }

  public static void loadI18n(Application application, HttpServletRequest request) throws IOException {
    String i18nPath = application.getRootFolder() + File.separator + "settings" + File.separator + "i18n";
    String root = HttpUtilities.getFileSystemRootPathForRequest(request);
    final String path = root + i18nPath;
    File i18nFolder = new File(path);
    if (i18nFolder.exists()) {
      if (!i18nFolder.isDirectory())
        throw new PlatformRuntimeException(path + " must be directory");
      File[] files = i18nFolder.listFiles();
      for (int i = 0; i < files.length; i++) {
        File file = files[i];
        if (file.getName().toLowerCase().endsWith(".i18n"))
          ApplicationState.i18n.loadI18n(file);
      }
      Tracer.platformLogger.info("open lab platform - " + application.getName().toLowerCase() + " i18n settings loaded ok");
    }
  }

  public Serializable getI18nEditingOperatorId() {
    return i18nEditingOperatorId;
  }

  public void setI18nEditingOperatorId(Serializable i18nEditingOperatorId) {
    this.i18nEditingOperatorId = i18nEditingOperatorId;
  }


  /**
   *
   */
  public class I18nEntry {

    public String code;
    private boolean suspect = false;
    private boolean seen = false;

    public Map<String, ApplicationEntry> applicationEntries = new TreeMap<String, ApplicationEntry>();


    protected I18nEntry(String code, String application, String language, String value) {
      ApplicationEntry ae = applicationEntries.get(application);
      this.code = code;
      if (ae == null)
        ae = new ApplicationEntry(application);
      ae.addEntry(language, value);
      applicationEntries.put(application, ae);
    }


    protected ApplicationEntry getApplicationEntry(String application) {
      ApplicationEntry ret = null;
      String defaultApplication = Application.PLATFORM_APP_NAME;

      if (application == null) {
        application = defaultApplication;
      }
      ret = applicationEntries.get(application);
      if (ret == null) {
        if (lenient == LENIENT_APP || lenient == LENIENT_APP_LANG) {
          ret = ret = applicationEntries.get(defaultApplication);
          if (ret == null) {
            if (applicationEntries.size() > 0)
              ret = applicationEntries.get(applicationEntries.keySet().iterator().next());
          }
        } else {
          ret = new ApplicationEntry(application);
        }
      }
      return ret;
    }

    protected String getLabel(String application, String language) {
      return getApplicationEntry(application).getLabel(language);
    }

    public boolean matches(String textToFind) {
      boolean match = false;
      if (code.toLowerCase().indexOf(textToFind.toLowerCase()) >= 0) {
        match = true;
      } else {
        for (Iterator<ApplicationEntry> iterator = applicationEntries.values().iterator(); iterator.hasNext() && !match;) {
          ApplicationEntry ae = iterator.next();
          match = ae.matches(textToFind);
        }

      }
      return match;
    }

    public boolean isSuspect() {
      return suspect;
    }

    public void setSuspect(boolean suspect) {
      this.suspect = suspect;
    }

    public boolean isSeen() {
      return seen;
    }

    public void setSeen(boolean seen) {
      this.seen = seen;
    }

    public String getCode() {
      return code;
    }

  }

  /**
   *
   */
  public class ApplicationEntry {

    public String applicationName;

    public ApplicationEntry(String applicationName) {
      this.applicationName = applicationName;
    }

    public Map<String, String> entries = new TreeMap();


    protected void addEntry(String lang, String value) {
      entries.put(lang, value);
    }

    protected String getLabel(String language) {

      String ret = null;
      String defaultLanguage = ApplicationState.getApplicationSetting(OperatorConstants.FLD_SELECT_LANG);
      if (language == null) {
        language = defaultLanguage;
      }
      language = language.toUpperCase();
      ret = entries.get(language);
      if (ret == null && (lenient == LENIENT_LANG || lenient == LENIENT_APP_LANG)) {
        if (defaultLanguage!=null)
          ret = entries.get(defaultLanguage);
        if (ret == null)
          entries.get("EN");
        if (ret == null) {
          if (entries.size() > 0)
            ret = entries.get(entries.keySet().iterator().next());
        }

      }
      return ret;
    }

    public boolean matches(String textToFind) {
      boolean match = false;
      if (applicationName.toLowerCase().indexOf(textToFind.toLowerCase()) >= 0) {
        match = true;
      } else {
        for (Iterator<String> iterator = entries.values().iterator(); iterator.hasNext() && !match;) {
          String label = iterator.next();
          match = label.indexOf(textToFind) >= 0;
        }

      }
      return match;
    }


  }


}
