package org.jblooming.waf.settings;

import org.apache.log4j.Logger;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.operator.Operator;
import org.jblooming.persistence.hibernate.HibernateFactory;
import org.jblooming.tracer.Tracer;
import org.jblooming.transaction.Transaction;
import org.jblooming.transaction.hibernate.HibernateTransaction;
import org.jblooming.utilities.DateUtilities;
import org.jblooming.utilities.HttpUtilities;
import org.jblooming.utilities.file.FileUtilities;
import org.jblooming.waf.constants.Commands;
import org.jblooming.waf.constants.OperatorConstants;
import org.logicalcobwebs.proxool.ProxoolFacade;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.PageContext;
import java.io.File;
import java.util.*;

/**
 * Holds a list of applications, and with the setup method lauches their config in the correct order.
 *
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 */
public class PlatformConfiguration {

  /**
   * used for checking launch of admin features with no operator logged
   */
  public static String psw = "dom";
  public static String globalSettingsFileName = "global.properties";
  public static Class defaultOperatorSubclass = Operator.class;
  public Map<String, Application> applications = new TreeMap();
  public String driver_class;
  public String driver_url;
  public Class dialect;
  public String db_user_name;
  public String db_user_psw;
  public boolean useProxool = false;
  public boolean useC3P0 = true;
  public static String POOL_ALIAS = "pool";
  public static boolean schedulerRunsByDefault = false;



  /**
   * This should be specified as tablespace name in oracle when the user has sufficient rights to access different schemas
   */
  public static String schemaName;
      

  //public static boolean oracle10gRelease2caseInsensitiveSupport = false;


  /**
   * @see "http://www.hibernate.org/293.html"
   *      Sample config for case-sensitive dbs (Oracle):
   *      stringPrimitiveFieldsConversion = PlatformConfiguration.StringConversionType.UPPER;
   *      searchStringParametersConversion = PlatformConfiguration.StringConversionType.UPPER
   *      <p/>
   *      Sample config for case-insensitive dbs (SqlServer, MySQL):
   *      leave them all to none
   */
  public static enum StringConversionType {

    UPPER,LOWER,NONE}

  public StringConversionType stringPrimitiveFieldsConversion = PlatformConfiguration.StringConversionType.NONE;
  public StringConversionType searchStringParametersConversion = PlatformConfiguration.StringConversionType.NONE;


  /**
   * distinguish between development and production environment
   * example: allows to use a different login only in deployment
   */
  public boolean development = false;

  public String defaultIndex = "/index.jsp";

  public Application defaultApplication;

  public AdvancedSettings injectableSettings;

  public boolean catchPersistenceErrors = false;

  public boolean loadedBasicPersistentSettings = false;

  public PlatformConfiguration(HttpServletRequest request, LogConfiguration lc, AdvancedSettings injectableSettings) {
    injectableSettings.setRoot(request.getSession(true).getServletContext().getRealPath("/").replaceAll("\\\\", "/"));
    injectableSettings.setLogConfiguration(lc);
    this.injectableSettings = injectableSettings;
  }

  public void addApplication(Application application) {
    applications.put(application.getName(), application);
  }

  public synchronized void setup(PageContext pageContext) {

    //if (applications == null || applications.size() < 2)
    //  throw new RuntimeException("You must configure platform and at least one application");
    HttpServletRequest request = (HttpServletRequest) pageContext.getRequest();

    if (
      PlatformConfiguration.psw.equalsIgnoreCase(pageContext.getRequest().getParameter("psw")) &&
      "restart".equals(pageContext.getRequest().getParameter(Commands.COMMAND))
      )
      HibernateFactory.resetSessionFactory();

    injectableSettings.configLog4J();
    Tracer.hibernateLogger.setLevel(injectableSettings.getLogConfiguration().getHibernateLogLevel());
    Tracer.platformLogger.setLevel(injectableSettings.getLogConfiguration().getPlatformLogLevel());
    Tracer.i18nLogger.setLevel(injectableSettings.getLogConfiguration().getI18nLogLevel());
    Tracer.jobLogger.setLevel(injectableSettings.getLogConfiguration().getJobLogLevel());

    System.setProperty("file.encoding","UTF-8");
    //load global settings
    injectableSettings.configGlobalSettings(pageContext);

    applyGlobalSettingsFromMemory(request);

    //verify correct configuration of platform base app
    Application platform = applications.get(Application.PLATFORM_APP_NAME);
    if (platform == null)
      throw new PlatformRuntimeException("Fatal error in settings.jsp: platform base application not found in PlatformConfiguration.applications");

    //first launch base classes
    launch(platform, pageContext, null);

    //verify correct configuration of at least one web app
    if (applications.keySet().size() < 2)
      throw new PlatformRuntimeException("Fatal error in settings.jsp: no application found in PlatformConfiguration.applications other than platform base");

    //launch single web apps
    for (Iterator iterator = new HashSet(applications.keySet()).iterator(); iterator.hasNext();) {
      final String key = (String) iterator.next();
      if (!Application.PLATFORM_APP_NAME.equals(key)) {
        Application application = (Application) applications.get(key);
        launch(application, pageContext, key);
      }
    }

    configPersistence();

    //todo: setupAuditAndSubscribable(injectableSettings.getHibConfiguration());

    if (loadedBasicPersistentSettings) {
      injectableSettings.configAndBuildHibSessionFactory();
      injectableSettings.createDefaultOperators();
      Tracer.platformLogger.info("open lab platform - persistence settings loaded ok");
    }

    afterPersistenceSetup();

    withPageContextSetup(pageContext);

    if (development)
      Tracer.platformLogger.warn("open lab platform - LOADED SETTINGS FOR DEVELOPMENT ENVIRONMENT");

    ApplicationState.loaded = true;

    if (defaultApplication==null)
        Tracer.platformLogger.warn(
                "open lab platform\nCould not find find a default application: add\n" +
                "  ApplicationState.platformConfiguration.defaultApplication = ...\n" +
                "in your settings.jsp.");

  }

  public static void applyGlobalSettingsFromMemory(HttpServletRequest request) {

    ApplicationState.serverURL = HttpUtilities.serverURL(request); // this uses PUBLIC_SERVER_NAME if exists
    ApplicationState.webAppFileSystemRootPath = HttpUtilities.getFileSystemRootPathForRequest(request);
    ApplicationState.contextPath = request.getContextPath();

    applyGlobalSettingDefaults();
  }

  public static void applyGlobalSettingDefaults() {
    String selectLang = ApplicationState.getApplicationSetting(OperatorConstants.FLD_SELECT_LANG);
    if (selectLang==null || selectLang.trim().length()==0) {
      selectLang="EN";
      Tracer.platformLogger.error("SELECT_LANG not found in global properties");
    }
    try {
      DateUtilities.DEFAULT_LOCALE = new Locale(selectLang);
    } catch (Exception e) {
      DateUtilities.DEFAULT_LOCALE = Locale.UK;
      Tracer.platformLogger.error("SELECT_LANG has invalid value in global properties:"+selectLang);
    }
    //very important! so the JDK is aware of the application needs
    Locale.setDefault(DateUtilities.DEFAULT_LOCALE);
  }

  public void afterPersistenceSetup() {
    if (loadedBasicPersistentSettings) {
      for (Application application : applications.values()) { 

        try {
          application.configureNeedingPersistence(this);
          Tracer.platformLogger.info("open lab platform - " + application.getName().toLowerCase() + " settings after persistence loaded ok");

        } catch (Exception e) {
          Tracer.platformLogger.error("Application " + application.getName() + " removed as it failed post persistence setup");
          e.printStackTrace();
        }
      }
    }
  }

  public void withPageContextSetup(PageContext pageContext) {
    if (loadedBasicPersistentSettings) {

      for (Application application : applications.values()) {
        try {
          application.configureNeedingPageContext(pageContext);
          Tracer.platformLogger.info("open lab platform - " + application.getName().toLowerCase() + " settings with PageContext loaded ok");
        } catch (Exception e) {
          Tracer.platformLogger.error("Application " + application.getName() + " removed as it failed settings with PageContext setup");
          e.printStackTrace();
        }
      }
    }
  }

  public void configPersistence() {
    if (catchPersistenceErrors)
      try {
        configPoolAndHibernateProperties();
        loadedBasicPersistentSettings = true;
      } catch (Exception e) {
        Tracer.platformLogger.error("open lab platform - basic persistence is not configured");
        //e.printStackTrace();
      }
    else {
      configPoolAndHibernateProperties();
      loadedBasicPersistentSettings = true;
    }
  }

  /* private void setupAuditAndSubscribable(Configuration hibConfiguration) {
    hibConfiguration.getSessionEventListenerConfig().setSaveEventListener( new PlatformSaveListener() );
    Tracer.platformLogger.info("open lab platform - loaded audit and subscribe listeners");

  }*/

  private void launch(Application application, PageContext pageContext, final String key) {

    try {
      application.configureNotMonitored(this);
      application.configureFreeAccess(this);
      application.configurePersistence(this);
      application.configureViews(this);
      I18n.loadI18n(application, (HttpServletRequest) pageContext.getRequest());
      ApplicationState.getPermissions().putAll(application.getPermissions());
      loadApplicationGlobalSettings(pageContext, application);
      Tracer.platformLogger.info("open lab platform - " + application.getName().toLowerCase() + " settings before persistence loaded ok");

    } catch (Exception e) {
      if (key != null)
        applications.remove(key);

      Tracer.platformLogger.error("open lab platform - " + "application " + application.getName().toLowerCase() + " removed as it failed persistence setup");
      e.printStackTrace();
    }
  }

  private void loadApplicationGlobalSettings(PageContext pageContext, Application application) {
    String appGloblaPath =
      HttpUtilities.getFileSystemRootPathForRequest((HttpServletRequest) pageContext.getRequest()) +
        application.getRootFolder() + File.separator + application.getName() + ".properties";

    final Properties properties = FileUtilities.getProperties(appGloblaPath);
    if (properties != null)
      ApplicationState.applicationSettings.putAll(properties);
  }

  public void configPoolAndHibernateProperties() {

    // this is in order to enable relaunch on the fly
    //issue_PLA_8
    Transaction t = HibernateTransaction.getCurrentTransaction();
    if (t != null)
      t.setFailed();
    HibernateFactory.resetSessionFactory();
    ApplicationState.loaded = false;
    if (ApplicationState.poolLoaded) {
      if (useProxool)
      ProxoolFacade.shutdown(0);
      //ProxoolFacade.removeConnectionPool(url);
      ApplicationState.poolLoaded = false;
    }
   Properties hibProp = injectableSettings.getHibConfiguration().getProperties();
    if (useProxool || useC3P0) {
      injectableSettings.configPool(hibProp, POOL_ALIAS, db_user_name, db_user_psw, driver_class, driver_url);
    } else {
      injectableSettings.configNoPool(hibProp, db_user_name, db_user_psw, driver_class, driver_url);
    }

    ApplicationState.poolLoaded = true;
    // beautiful hack: proxool re-enables it !
    Logger.getRootLogger().removeAllAppenders();
    //String logFilesRoot = injectableSettings.getRoot() + "WEB-INF/log/";
    //injectableSettings.createLogger("%d{yyyy MMM dd HH:mm:ss} %5p %c{1}:%L - %m%n",logFilesRoot,Logger.getRootLogger(),"R","tomcat.log","tomFileApp");
    injectableSettings.configHibernateProperties(injectableSettings.getHibConfiguration(), POOL_ALIAS, dialect);
  }

  public Application getDefaultApplication() {
    return defaultApplication;
  }

  public Application getApplication(String appName) {

    //in case of Platform, fall to default
    Application app = null;
    if (appName != null && !"Platform".equalsIgnoreCase(appName))
      app = applications.get(appName);
    if (app != null)
      return app;
    else
      return getDefaultApplication();
  }

}
