package org.jblooming.waf.view;

import antlr.RecognitionException;
import antlr.TokenStreamException;
import org.jblooming.agenda.CompanyCalendar;
import org.jblooming.ontology.Hidrator;
import org.jblooming.ontology.SerializedList;
import org.jblooming.ontology.SerializedMap;
import org.jblooming.utilities.DateUtilities;
import org.jblooming.utilities.NumberUtilities;
import org.jblooming.utilities.StringUtilities;
import org.jblooming.utilities.math.MathParse;
import org.jblooming.waf.constants.FieldErrorConstants;
import org.jblooming.waf.constants.Fields;
import org.jblooming.waf.exceptions.ActionException;
import org.jblooming.waf.settings.ApplicationState;

import java.text.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;


public class ClientEntry {

  public String name;
  private String value;
  public String errorCode = null;

  public boolean required = false;

  private boolean fillErrorCodeOnError = true;

  public String suggestedValue;

  //mark the businessLogic so that PageState.saveEntriesInDefaults will persist it in options if called on same pageState
  public boolean persistInOptions = false;


  public ClientEntry(String name, int value) {
    this(name, value + "", null, null);
  }

  public ClientEntry(String name, String value) {
    this(name, value, null, null);
  }


  public ClientEntry(String name, String value, String suggestedValue, String errorDescription) {

    this.name = name;
    this.setValue(value);

    this.suggestedValue = suggestedValue;
    this.errorCode = errorDescription;
  }

  public ClientEntry getNewInstance() {
    return new ClientEntry(name, ((value != null) ? new String(value) : null));
  }

  public String stringValueNullIfEmpty() {
    String ret = null;
    if (!(value != null && value.trim().length() == 0))
      ret = value;
    return ret;
  }

  public String stringValue() throws ActionException {
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }
    return value;
  }

  public SerializedList listValue() throws ActionException {
    String value = stringValue();
    return SerializedList.deserialize(value);
  }

  public List listValue(Hidrator hidrator) throws ActionException {
    String value = stringValue();
    List returnValues = new ArrayList();
    List<String> values = SerializedList.deserialize(value);
    for (String s : values) {
      returnValues.add(hidrator.hidrate(s));
    }
    return returnValues;
  }


  public SerializedMap mapValue() throws ActionException {
    String value = stringValue();
    return SerializedMap.deserialize(value);
  }

  public int intValue() throws ActionException, ParseException {
    int result = 0;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }

    if (isFilled())
      try {
        result = Integer.parseInt(value);
      } catch (NumberFormatException e) {
        result = (int) processEquation(e);
        value = result + "";
      }
    else
      throw new ParseException("Null value not allowed", 0);
    return result;
  }

  public int intValueNoErrorCodeNoExc() {
    fillErrorCodeOnError = false;
    try {
      return intValue();
    } catch (ParseException e) {
    } catch (ActionException e) {
    }
    return 0;
  }


  public long longValueNoErrorNoCatchedExc() {
    long ret = 0;
    try {
      return longValue();
    } catch (ActionException e) {
    } catch (ParseException e) {
    }
    return ret;
  }

  public long longValue() throws ActionException, ParseException {
    long result = 0L;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }
    if (isFilled())
      try {
        result = Long.parseLong(value);
      } catch (NumberFormatException e) {
        result = (long) processEquation(e);
        value = result + "";
      }
    else
      throw new ParseException("Null value not allowed", 0);
    return result;
  }

  public double doubleValue() throws ActionException, ParseException {
    double result = 0.0;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }
    if (isFilled()) {
      ParsePosition pp = new ParsePosition(0);
      NumberFormat instance = NumberFormat.getInstance();
      //result=Double.parseDouble(value);
      Number num = instance.parse(value, pp);
      if (num == null) {
        result = processEquation(null);
      } else {
        result = num.doubleValue();
      }
      if (pp.getIndex() < value.length()) {
        result = processEquation(null);
        value = NumberUtilities.decimalNoGrouping(result, 2);
      }
    } else
      throw new ParseException("Null value not allowed", 0);
    return result;
  }

  public double doubleValueNoErrorNoCatchedExc() {
    try {
      return doubleValue();
    } catch (ActionException e) {
    } catch (ParseException e) {
    }
    return 0;
  }

  public double currencyValue() throws ActionException, ParseException {
    double result = 0.0;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }
    if (isFilled()) {
      try {
        String currFormat = ApplicationState.getApplicationSetting("CURRENCY_FORMAT");
        if (currFormat == null)
          currFormat = NumberUtilities.DEFAULT_CURRENCY_FORMAT;
        DecimalFormat format = new DecimalFormat(currFormat);
        result = format.parse(value).doubleValue();
      } catch (ParseException e) {
        result = processEquation(null);
        value = result + "";
      }
    } else
      throw new ParseException("Null value not allowed", 0);
    return result;
  }

  public double currencyValueNoErrorNoCatchedExc() {
    try {
      return currencyValue();
    } catch (ActionException e) {
    } catch (ParseException e) {
    }
    return 0;
  }


  private double processEquation(Exception e) throws ParseException {

    double result;

    String reworkedValue = StringUtilities.replaceAllNoRegex(value, new DecimalFormatSymbols().getGroupingSeparator() + "", "");
    reworkedValue = StringUtilities.replaceAllNoRegex(reworkedValue, new DecimalFormatSymbols().getDecimalSeparator() + "", ".");

    try {
      result = MathParse.parse(reworkedValue);
    } catch (TokenStreamException e1) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_MUST_BE_NUMERIC;
      if (e != null)
        throw new ParseException(value, 0);
      else
        throw new ParseException(value, 0);
    } catch (RecognitionException e1) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_MUST_BE_NUMERIC;
      if (e != null)
        throw new ParseException(value, 0);
      else
        throw new ParseException(value, 0);
    }
    return result;
  }


  public Date dateValue() throws ActionException, ParseException {
    Date result = null;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }
    if (isFilled())
      try {
        result = DateUtilities.dateFromString(value);
      } catch (ParseException e) {
        if (fillErrorCodeOnError)
          errorCode = FieldErrorConstants.ERR_FIELD_MUST_BE_DATELIKE;
        throw e;
      }
    return result;
  }


  public Date dateValueNoErrorNoCatchedExc() {
    try {
      fillErrorCodeOnError = false;
      return dateValue();
    } catch (ActionException e) {
    } catch (ParseException e) {
    }
    return null;
  }

  public long timeValueInMillis() throws ActionException, ParseException {
    long result = 0;
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
      throw new ActionException(FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY);
    }

    if (value != null && value.trim().length() > 0) {

      try {
        result = DateUtilities.millisFromHourMinuteSmart(value);

      } catch (NumberFormatException e) {
        if (fillErrorCodeOnError)
          errorCode = FieldErrorConstants.ERR_FIELD_MUST_BE_TIME;
        throw new ParseException(FieldErrorConstants.ERR_FIELD_MUST_BE_TIME, 0);
      }


    } else
      throw new ParseException("Null value not allowed", 0);

    return result;
  }


  public boolean checkFieldValue() {
    return Fields.TRUE.equals(value);
  }

  public String checkFieldHtmlValue() {
    if (Fields.TRUE.equals(value))
      return Fields.TRUE;
    else
      return Fields.FALSE;
  }

  public void setValue(String value) {
    this.value = value;
  }

  public boolean isFilled() {
    return value != null && value.trim().length() > 0;
  }

//  ---------------------------------------------------------------------------  deprecated methods ---------------------------------------------------------------

  /**
   * @deprecated use intValue()
   */
  public int intValueWithErrorCode() throws ActionException, ParseException {
    fillErrorCodeOnError = true;
    return intValue();
  }

  /**
   * @deprecated
   */
  public String stringValueRequiredNoExc() {
    required = true;
    try {
      return stringValue();
    } catch (ActionException e) {
      return null;
    }
  }

  /**
   * @deprecated
   */
  public String getString() {
    if (required && !isFilled()) {
      if (fillErrorCodeOnError)
        errorCode = FieldErrorConstants.ERR_FIELD_CANT_BE_EMPTY;
    }
    return value;
  }

  /**
   * @deprecated
   */
  public double doubleValueNoErrorCode() throws ActionException, ParseException {
    fillErrorCodeOnError = false;
    return doubleValue();
  }

  /**
   * @deprecated
   */
  public long longValueNoErrorCode() throws ActionException, ParseException {
    fillErrorCodeOnError = false;
    return longValue();
  }

  /**
   * @deprecated use CheckField component
   */
  public boolean checkBoxValue() throws ActionException {
    return !(value == null);
  }

  public long durationInMillis() throws ParseException, ActionException {
    return (long) (doubleValue() * CompanyCalendar.MILLIS_IN_DAY);
  }
}
