package com.opnlb.website.forum;

import com.opnlb.website.forum.security.ForumPermissions;
import org.hibernate.annotations.CollectionOfElements;
import org.hibernate.annotations.ForeignKey;
import org.hibernate.annotations.MapKeyManyToMany;
import org.hibernate.search.annotations.*;
import org.jblooming.ontology.Node;
import org.jblooming.ontology.SecuredNodeSupport;
import org.jblooming.operator.Operator;
import org.jblooming.operator.User;
import org.jblooming.security.Permission;
import org.jblooming.utilities.HashTable;
import org.jblooming.utilities.JSP;

import javax.persistence.*;
import java.util.Date;
import java.util.Map;
import java.util.Set;


/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 */

@Entity
@Table(name = "_ws_forum")
@Indexed(index = "ForumEntry")
public class ForumEntry extends SecuredNodeSupport {


  private Operator author;

  private String title;

  private String content = "ROOT";

  private boolean active = true;
  private int order;
  private Date postedOn;
  private Date lastPostOnBranch;
  private String lastPosterOnBranch;
  protected ForumEntry parent;

  private int hits;

  public static final String FORUM_ENTRY = "FRME";

  private boolean postForLogged;
  private boolean postToAll;
  private boolean privateForum;

  private Map<Operator, Integer> votedBy = new HashTable();

  @Id
  @GeneratedValue(strategy = GenerationType.AUTO)
  @DocumentId
  public Integer getId() {
    return (Integer) super.getId();
  }


  public void setParent(ForumEntry n) {
    parent = n;
  }

  @ManyToOne(targetEntity = ForumEntry.class)
  @ForeignKey(name = "fk_forum_parent")
  @JoinColumn(name = "parent")
  public ForumEntry getParent() {
    return parent;
  }

  @Transient
  public Node getParentNode() {
    return getParent();
  }

  public void setParentNode(Node node) {
    setParent((ForumEntry) node);
  }

  @Field(name = "Title", index = Index.TOKENIZED, store = Store.NO)
  @Boost(2.5f)
  public String getTitle() {
    return title;
  }

  public void setTitle(String title) {
    this.title = title;
  }

  @Lob
  @Column(name = "contentx")
  public String getContent() {
    return content;
  }

  @Transient
  @Field(name = "Content", index = Index.TOKENIZED, store = Store.NO)
  public String getContentClean() {
    return JSP.cleanHTML(content);
  }


  public void setContent(String content) {
    this.content = content;
  }

  @ManyToOne(targetEntity = Operator.class)
  @ForeignKey(name = "fk_forum_operator")
  public Operator getAuthor() {
    return author;
  }

  public void setAuthor(Operator author) {
    this.author = author;
  }

  public boolean isActive() {
    return active;
  }

  public void setActive(boolean active) {
    this.active = active;
  }

  @Column(name = "orderx")
  public int getOrder() {
    return order;
  }

  public void setOrder(int order) {
    this.order = order;
  }

  public Date getPostedOn() {
    return postedOn;
  }

  public void setPostedOn(Date postedOn) {
    this.postedOn = postedOn;
  }

  @Column(name = "lastpost")
  public Date getLastPostOnBranch() {
    return lastPostOnBranch;
  }

  public void setLastPostOnBranch(Date lastPostOnBranch) {
    this.lastPostOnBranch = lastPostOnBranch;
  }

  @Column(name = "hitsx")
  public int getHits() {
    return hits;
  }

  public void setHits(int hits) {
    this.hits = hits;
  }

  @Column(name = "writetologged")
  public boolean isPostForLogged() {
    return postForLogged;
  }

  public void setPostForLogged(boolean postForLogged) {
    this.postForLogged = postForLogged;
  }

  @Column(name = "writetoall")
  public boolean isPostToAll() {
    return postToAll;
  }

  public void setPostToAll(boolean postToAll) {
    this.postToAll = postToAll;
  }

  @Transient
  public String getName() {
    return getTitle();
  }

  @Column(name = "lastposter")
  public String getLastPosterOnBranch() {
    return lastPosterOnBranch;
  }

  public void setLastPosterOnBranch(String lastPosterOnBranch) {
    this.lastPosterOnBranch = lastPosterOnBranch;
  }

  public boolean isPrivateForum() {
    return privateForum;
  }

  @Transient
  @Field(name = "Private", index = Index.UN_TOKENIZED, store = Store.NO)
  public String isPrivateForumValue() {
    return privateForum+"";
  }

  public void setPrivateForum(boolean privateForum) {
    this.privateForum = privateForum;
  }

  @Transient
  public ForumEntry getRootPost() {
    if (getDepth() < 3)
      return this;
    else
      return getParent();
  }

  @OneToMany(cascade = CascadeType.REMOVE, targetEntity = ForumEntry.class)
  @JoinColumn(name = "parent")
  public Set getChildren() {
    return super.getChildren();
  }

  public boolean hasPermissionFor(User u, Permission p) {

    boolean result = false;
    if (u != null)
      result = super.hasPermissionFor(u, p);

    if (!result) {

      if (ForumPermissions.forum_canRead.equals(p)) {

        result = !isPrivateForum();

      } else if (ForumPermissions.forum_canPost.equals(p)) {

        result = isPostToAll() || (u != null && isPostForLogged());

      } else if (ForumPermissions.forum_canEdit.equals(p)) {

        result = u != null && u.equals(getAuthor());
      }
    }
    return result;
  }


  @CollectionOfElements
  @JoinTable(name = "forum_votes",
          joinColumns = @JoinColumn(name = "forum_id", nullable = false))
  @MapKeyManyToMany(targetEntity = Operator.class, joinColumns = @JoinColumn(name = "op_id", nullable = false))
  @Column(name = "votes")
  @ForeignKey(name = "fk_forum_op_votes")
  public Map<Operator, Integer> getVotedBy() {
    return votedBy;
  }

  public void setVotedBy(Map<Operator, Integer> votedBy) {
    this.votedBy = votedBy;
  }

  @Transient
  public int getAverageRating() {
    int average = 0;
    if (getVotedBy().size() > 0) {
      for (int i : getVotedBy().values()) {
        average = average + i;
      }
      average = average / getVotedBy().size();
    }
    return average;
  }

}
