package com.twproject.agenda;

import com.twproject.resource.Person;
import com.twproject.resource.Resource;
import org.hibernate.Query;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.agenda.EventType;
import org.jblooming.agenda.Period;
import org.jblooming.agenda.ScheduleSupport;
import org.jblooming.logging.Sniffable;
import org.jblooming.ontology.Identifiable;
import org.jblooming.ontology.LoggableIdentifiableSupport;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.FindByPrimaryKeyException;
import org.jblooming.persistence.exceptions.FindException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.persistence.exceptions.StoreException;
import org.jblooming.persistence.hibernate.PersistenceContext;

import java.io.Serializable;
import java.util.*;

/**
 * Models the notion of event in time.Events are persistent
 * and collectable by groups. They tipically concern a human operator,
 * but as notion they cover also automation ends.
 */
public class Event extends LoggableIdentifiableSupport implements Sniffable {

  private Set<Resource> targets = new HashSet<Resource>();
  private Person author;
  private ScheduleSupport schedule;
  private String location;
  private boolean personal;
  private boolean reminder;
  private boolean unavailability;

  /**
   *
   */
  private EventType type;
  private int status;

  private String summary;
  private String description;

  //class__id of the eventual exact OperatorAggregator generating targets
  String originatingObject;

  // refernces used for creating links on the event display
  private Set<String> references = new HashSet();

  private String icalId;


  public static final String EVENT = "EVENT";


  public Event() {
  }

  public Event(Serializable aeId) {
    setId(aeId);
  }

  public void setReminder(boolean b) {
    this.reminder = b;
  }

  public boolean isReminder() {
    return reminder;
  }

  public Set<Resource> getTargets() {
    return targets;
  }

  public void setTargets(Set<Resource> targets) {
    this.targets = targets;
  }

  private Set<String> getReferences() {
    return references;
  }

  private void setReferences(Set<String> references) {
    this.references = references;
  }

  public int getTargetSize() {
    int ret = 0;
    if (targets != null)
      ret = targets.size();
    return ret;
  }

  public ScheduleSupport getSchedule() {
    return schedule;
  }

  public EventType getType() {
    return type;
  }

  public void setType(EventType type) {
    this.type = type;
  }

  public int getStatus() {
    return status;
  }

  /**
   * Not in use yet.
   */
  public void setStatus(final int status) {
    this.status = status;
  }

  /**
   * Permits to calculate the time intersections of it.
   */
  public void setSchedule(ScheduleSupport schedule) {
    this.schedule = schedule;
  }

  public String getDescription() {
    return description;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public String getLocation() {
    return location;
  }

  /**
   * Where it will take place.
   */
  public void setLocation(String location) {
    this.location = location;
  }


  public boolean isPersonal() {
    return personal;
  }

  /**
   * Whether it is visible to non-owners.
   */
  public void setPersonal(boolean personal) {
    this.personal = personal;
  }

  public Person getAuthor() {
    return author;
  }

  public Date getLastModified() {
    return lastModified;
  }

  public void setLastModified(Date lastModified) {
    this.lastModified = lastModified;
  }

  public String getName() {
    return getSummary();
  }


  public void setAuthor(Person author) {
    this.author = author;
  }


  public void addReference(Identifiable reference) {
    references.add(generateName(reference));
  }

  private String generateName(Identifiable reference) {
    return reference.getId() + "__" + PersistenceHome.deProxy(reference.getClass().getName());
  }

  public void removeReference(Identifiable reference) {
    references.remove(generateName(reference));
  }

  public Iterator getReferencesIterator() {
    return references.iterator();
  }

  public Iterator getReferencesInstancesIterator() {
    Set<Identifiable> refs = new HashSet<Identifiable>();
    if (references != null && references.size() > 0) {
      for (Iterator<String> iterator = new HashSet<String>(references).iterator(); iterator.hasNext();) {
        String name = iterator.next();
        String[] spl = name.split("__");
        String id = spl[0];
        String clazz = spl[1];
        try {
          refs.add(PersistenceHome.findByPrimaryKey((Class<? extends Identifiable>) Class.forName(clazz), id));
        } catch (FindByPrimaryKeyException e) {
          references.remove(name);
        } catch (ClassNotFoundException e) {
          throw new PlatformRuntimeException(e);
        }

      }
    }
    return refs.iterator();
  }


  public static Event newInstanceWithScheduleNull(Event e) {

    Event newEvent = new Event();
    newEvent.targets = e.targets;
    newEvent.author = e.author;
    newEvent.schedule = null;
    newEvent.location = e.location;
    newEvent.personal = e.personal;
    newEvent.type = e.type;
    newEvent.status = e.status;
    newEvent.description = e.description;
    newEvent.lastModified = e.lastModified;
    newEvent.references = e.references;
    return newEvent;
  }

  // --------------------------------------------------------- UTILITY METHODS ---------------------------------------------------


  /**
   * @param resource
   * @param period   if null get today
   * @param refine   If true loop events in order to remove the event doesen't match
   * @return a list of candidate events. Do not checks recursive events, but the edge only.
   *         The list is order by schedule.start: this do not means it is order correctly ;-)
   * @throws FindException
   */
  public static List<Event> getEventsInPeriodFor(Resource resource, Period period, boolean showAuthored, boolean refine) throws FindException {

    if (period == null) {
      period = Period.getDayPeriodInstance(new Date());
    }
    String hql = "select event from " + Event.class.getName() + " as event left join event.targets as target where " +
            "(target=:logged " + (showAuthored ? "or author=:logged" : "") + ") " +
            "and not (event.schedule.end < :start or event.schedule.start > :end) order by event.schedule.start";
    //    "and not (event.schedule.end <= :start or event.schedule.start >= :end) order by event.schedule.start";
    OqlQuery oql = new OqlQuery(hql);
    Query query = oql.getQuery();
    query.setParameter("logged", resource);
    query.setParameter("start", period.getValidityStartDate());
    query.setParameter("end", period.getValidityEndDate());
    List<Event> result = (List<Event>) oql.list();


    if (refine) { // remove non-exactly-matching events
      List<Event> refined = new ArrayList<Event>();
      for (Event event : result) {
        if (!refined.contains(event)) {
          ScheduleSupport schedule = event.getSchedule();
          //long nextFire = schedule.getNextFireTimeAfter(period.getValidityStartTime());
          //if (nextFire < period.getValidityEndTime()) {
          //  refined.add(event);
          //}
          if (schedule.overlap(period)) {
            refined.add(event);
          }
        }
      }
      return refined;
    } else {
      return result;
    }
  }


  /**
   * @param resource
   * @param period   if null get today
   * @return a Map<Period,Event> list of atomic event as periods for the matching period and person. It checks for recursive events.
   *         The list is order by schedule.start: this do not means it is order correctly ;-)
   * @throws FindException
   */
  public static TreeMap<Period, Event> getPeriodsInPeriodFor(Resource resource, Period period, boolean showAuthored) throws FindException {
    if (period == null) {
      period = Period.getDayPeriodInstance(new Date());
    }
    List<Event> result = getEventsInPeriodFor(resource, period, showAuthored, true);

    TreeMap<Period, Event> disgraned = new TreeMap<Period, Event>();

    for (Event event : result) {
      ScheduleSupport schedule = event.getSchedule();
      List<Period> sgran = schedule.getPeriods(period, false);
      for (Period p : sgran) {
        disgraned.put(p, event);
      }
    }
    return disgraned;
  }

  public static List<Event> getFilteredEventsInPeriodWithCollisionFor(List<Resource> resources, Period period, int typeId, boolean showWork, boolean isOlnyMine, boolean showAuthored, boolean showTargets, boolean isPersonal) throws PersistenceException {
    if (period == null) {
      period = Period.getDayPeriodInstance(new Date());
    }
    String hql = "select distinct event from " + Event.class.getName() + " as event join event.targets as target where ";
    hql += " (not (event.schedule.end < :start or event.schedule.start > :end)) "; //order by event.schedule.start

    if (showTargets)
      hql += " and ( target in (:myList) )";
    if (showWork)
      hql += " and ( target in (:myList) and personal=:isPersonal)";
    if (showAuthored)
      hql += " and ( author in (:myList) )";
    //if(showTargets && showAuthored)
    //  hql+=")";
    if (isPersonal)
      hql += " and ( author in (:myList) and personal=:isPersonal )";
    else if (isOlnyMine)
      hql += " and ( target in (:myList) and size(event.targets) = 1) ";
    if (typeId > 0)
      hql += " and ( event.type.id=:typeId )";

    //avoid remainders
    if (resources.size() > 1) {
      hql += " and ( event.reminder=:remFalsity )";
    }

    OqlQuery oql = new OqlQuery(hql);
    Query query = oql.getQuery();
    if (showAuthored || showTargets || isOlnyMine || isPersonal || showWork)
      query.setParameterList("myList", resources);
    if (isPersonal || showWork)
      query.setBoolean("isPersonal", isPersonal);
    if (typeId > 0)
      query.setParameter("typeId", typeId);
    query.setTimestamp("start", period.getValidityStartDate());
    query.setTimestamp("end", period.getValidityEndDate());
    if (resources.size() > 1) {
      query.setBoolean("remFalsity", Boolean.FALSE);
    }

    List<Event> result = (List<Event>) oql.list();
    Collections.sort(result, Event.eventPeriod);
    return result;
  }

  public static Comparator eventPeriod = new Comparator() {
    public int compare(Object o1, Object o2) {
      Event p1 = (Event) o1;
      Event p2 = (Event) o2;
      int retVal = 0;
      if (p1.getSchedule().getValidityStartTime() > p2.getSchedule().getValidityStartTime())
        retVal = 1;
      if (p2.getSchedule().getValidityStartTime() > p1.getSchedule().getValidityStartTime())
        retVal = -1;
      return retVal;
    }
  };

  public String getSummary() {
    return summary;
  }

  public void setSummary(String summary) {
    this.summary = summary;
  }

  public String getIcalId() {
    return icalId;
  }

  public void setIcalId(String icalId) {
    this.icalId = icalId;
  }

  public void store(PersistenceContext pc) throws StoreException {
    boolean isNew = isNew();
    super.store(pc);
    if (getIcalId()==null)
      setIcalId("TW_" + getId());
  }

  public boolean isUnavailability() {
    return unavailability;
  }

  public void setUnavailability(boolean unavailability) {
    this.unavailability = unavailability;
  }
}



