package com.twproject.resource;

import com.twproject.classification.ResourceClassification;
import com.twproject.classification.ResourceClassificationAssociation;
import com.twproject.document.TeamworkDocument;
import com.twproject.operator.TeamworkOperator;
import com.twproject.task.*;
import com.twproject.task.financial.CostAggregator;
import com.twproject.worklog.Worklog;
import org.jblooming.ApplicationRuntimeException;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.agenda.CompanyCalendar;
import org.jblooming.agenda.Period;
import org.jblooming.anagraphicalData.AnagraphicalData;
import org.jblooming.company.Location;
import org.jblooming.ontology.Documentable;
import org.jblooming.ontology.Node;
import org.jblooming.ontology.PersistentFile;
import org.jblooming.ontology.SecuredNodeWithAreaSupport;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.FindException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.persistence.exceptions.RemoveException;
import org.jblooming.persistence.exceptions.StoreException;

import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import java.io.UnsupportedEncodingException;
import java.util.*;

/**
 * @author Pietro Polsinelli ppolsinelli@open-lab.com
 * @author Roberto Bicchierai rbicchierai@open-lab.com
 */
public abstract class Resource extends SecuredNodeWithAreaSupport implements PeopleAggregator, Documentable {

  public static final String RESOURCE = "RESOURCE";

  private String name;
  protected Resource parent;
  private String code;
  private Location location;
  private Set<AnagraphicalData> anagraphicalDatas = new HashSet<AnagraphicalData>();
  private Set<TeamworkDocument> documents = new HashSet<TeamworkDocument>();
  private String notes;
  private Set<Assignment> assignments = new HashSet<Assignment>();

  /**
   * Introduced in order to introduce virtual or multiuser resources, such as "a board", "a counsel" "stearing committee"
   */

  private boolean staff = false;

  private Person myManager;
  //for performance
  private String myManagerIds;

  private CostAggregator myCostAggregator;

  // read only colls: is the collection of resources where "this" is the manager
  private Set<Resource> myStaff;

  private Set<Issue> issuesAssignedToMe;

  private Set<ResourceClassificationAssociation> resourceClassificationAssociations = new HashSet<ResourceClassificationAssociation>();
  protected String jobDescription;

  private PersistentFile myPhoto;

  private int hourlyCost;


  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public Set<Assignment> getAssignments() {
    return assignments;
  }

  public int getAssignmentsSize() {
    return getAssignments() != null ? getAssignments().size() : 0;
  }

  public Iterator<Assignment> getAssignmentsIterator() {
    return assignments.iterator();
  }

  public void setAssignments(Set<Assignment> assignments) {
    this.assignments = assignments;
  }

  public void setDocuments(Set<TeamworkDocument> documents) {
    this.documents = documents;
  }

  public Set getDocuments() {
    return documents;
  }

  public int getDocumentsSize() {
    return getDocuments() != null ? getDocuments().size() : 0;
  }

  public void addDocument(TeamworkDocument cost) {
    documents.add(cost);
  }

  public void removeCost(TeamworkDocument cost) {
    documents.remove(cost);
  }

  public Iterator<TeamworkDocument> getDocumentsIterator() {
    return documents.iterator();
  }


  private void setParent(Resource n) {
    parent = n;
  }

  public Resource getParent() {
    return parent;
  }

  public Node getParentNode() {
    return getParent();
  }

  public void setParentNode(Node node) {
    setParent((Resource) node);
  }

  public String getCode() {
    return code;
  }

  public void setCode(String code) {
    this.code = code;
  }

  public Location getLocation() {
    return location;
  }

  public void setLocation(Location location) {
    this.location = location;
  }

  public Set<AnagraphicalData> getAnagraphicalDatas() {
    return anagraphicalDatas;
  }

  public void setAnagraphicalDatas(Set<AnagraphicalData> anagraphicalDatas) {
    this.anagraphicalDatas = anagraphicalDatas;
  }

  public String getNotes() {
    return notes;
  }

  public void setNotes(String notes) {
    this.notes = notes;
  }

  public boolean isStaff() {
    return staff;
  }

  public void setStaff(boolean staff) {
    this.staff = staff;
  }

  public Person getMyManager() {
    return myManager;
  }

  private void setMyManager(Person myManager) {
    this.myManager = myManager;
  }

  public void setMyManagerAndPerfIds(Person n) {

    if (n != null && n.getMyManagerIds() != null && n.getMyManagerIds().indexOf(n.getId() + "") != -1)
      throw new ApplicationRuntimeException(
              "You can't set this node as manager of its own: manager ancestor ids = '" + n.getAncestorIds() + "', manager id = '" + n.getId() + "'");

    setMyManager(n);
    if (n != null)
      n.getMyStaff().add(this);

    if (n != null)
      setMyManagerIds((n.getMyManagerIds() != null ? n.getMyManagerIds() : "") + n.getId() + SEPARATOR);
    else
      setMyManagerIds(null);
  }


  public CostAggregator getMyCostAggregator() {
    return myCostAggregator;
  }

  public void setMyCostAggregator(CostAggregator myCostAggregator) {
    this.myCostAggregator = myCostAggregator;
  }

  public Set<Resource> getMyStaff() {
    return myStaff;
  }

  public void setMyStaff(Set<Resource> myStaff) {
    this.myStaff = myStaff;
  }


  public Set<ResourceClassificationAssociation> getResourceClassificationAssociations() {
    return resourceClassificationAssociations;
  }

  private void setResourceClassificationAssociations(Set<ResourceClassificationAssociation> resourceClassificationAssociations) {
    this.resourceClassificationAssociations = resourceClassificationAssociations;
  }


  public boolean addClassificationAndPersist(ResourceClassification classification) throws StoreException, RemoveException {

    boolean result = removeClassificationAndDepersist(classification);

    ResourceClassificationAssociation tca = new ResourceClassificationAssociation();
    tca.setClassification(classification);
    tca.setResource(this);
    this.resourceClassificationAssociations.add(tca);
    tca.store();

    return result;
  }

  public boolean removeClassificationAndDepersist(ResourceClassification classification) throws RemoveException {

    boolean result = false;

    //find the tca
    for (ResourceClassificationAssociation tca : new HashSet<ResourceClassificationAssociation>(resourceClassificationAssociations)) {
      if (tca.getClassification().equals(classification)) {
        this.resourceClassificationAssociations.remove(tca);
        PersistenceHome.remove(tca);
        result = true;
        break;
      }
    }
    return result;
  }

  public Iterator<ResourceClassification> getClassificationIterator() {

    Set<ResourceClassification> tc = new HashSet<ResourceClassification>();
    for (ResourceClassificationAssociation tca : resourceClassificationAssociations) {
      tc.add(tca.getClassification());
    }
    return tc.iterator();
  }


  /**
   * Warning: this is implemented without caring for performance
   */
/*public Iterator getEventsIterator() {
  Iterator i = null;
  if (workerEvents != null) {
    Set events = new HashSet();
    for (Iterator iterator = workerEvents.iterator(); iterator.hasNext();) {
      events.add(((WorkerEvent) iterator.next()).getEvent());
    }
    i = events.iterator();
  }
  return i;
}*/
  public abstract String getDisplayName();


  public String getJobDescription() {
    return jobDescription;
  }

  public void setJobDescription(String jobDescription) {
    this.jobDescription = jobDescription;
  }


  public String getMyManagerIds() {
    return myManagerIds;
  }

  public void setMyManagerIds(String myManagerIds) {
    this.myManagerIds = myManagerIds;
  }

  public List<Assignment> getAllAssignmentsOfActiveTask() throws PersistenceException {

    List<Assignment> assig = new ArrayList();
    for (Assignment a : getAssignments()) {
      if (a.getTask().isActive()) {
        assig.add(a);
      }
    }
    return assig;
  }


  public List<Assignment> getActiveAssignments() throws PersistenceException {

    List<Assignment> assig = new ArrayList();
    for (Assignment a : getAssignments()) {
      if (a.isEnabled() && a.getTask().isActive()) {
        assig.add(a);
      }
    }
    return assig;
  }

  public List<Assignment> getActiveAssignments(Period period, boolean showActiveOnly) throws PersistenceException {
    return getActiveAssignments(period, showActiveOnly, true);
  }


  public List<Assignment> getActiveAssignments(Period period, boolean showActiveOnly, boolean showActiveTasksOnly) throws PersistenceException {
    String hql = "select assignment from " + Assignment.class.getName() + " as assignment where assignment.resource=:resource";

    if (showActiveOnly) {
      hql = hql + " and assignment.enabled=:truth";
    }

    if (showActiveTasksOnly) {
      hql = hql + " and assignment.task.status=:STATUS_ACTIVE";
    }

//    if (period != null)
//      hql = hql + " and not (assignment.task.schedule.start>:periodEnd or assignment.task.schedule.end<:periodStart)";
    if (period != null) {
      hql = hql +
              " and  ( " + "" +
              "(assignment.task.schedule.start<=:periodEnd or assignment.task.schedule.start is null) and " +
              "(assignment.task.schedule.end>:periodStart or assignment.task.schedule.end is null) " +
              " )";
    }
    OqlQuery query = new OqlQuery(hql);
    query.setParameter("resource", this);

    if (showActiveOnly) {
      query.setParameter("truth", Boolean.TRUE);
    }

    if (showActiveTasksOnly) {
      query.setParameter("STATUS_ACTIVE", TaskStatus.STATUS_ACTIVE);
    }

    if (period != null) {
      query.setParameter("periodEnd", period.getValidityEndDate());
      query.setParameter("periodStart", period.getValidityStartDate());
    }
    return query.list();

  }


  private Set<Issue> getIssuesAssignedToMe() {
    return issuesAssignedToMe;
  }

  private void setIssuesAssignedToMe(Set<Issue> issuesAssignedToMe) {
    this.issuesAssignedToMe = issuesAssignedToMe;
  }


  public List<Assignment> getMyAssignmentsByType() {

    List<Assignment> assigs = new ArrayList<Assignment>();
    if (getAssignmentsSize() > 0) {
      Iterator<Assignment> itass = getAssignmentsIterator();

      while (itass.hasNext()) {
        Assignment ass = itass.next();
        assigs.add(ass);
      }

      Collections.sort(assigs, new Comparator<Assignment>() {
        public int compare(Assignment a1, Assignment a2) {

          int result = 0;

          Task t1 = a1.getTask();
          String displayName1 = t1.getDisplayName().toLowerCase();
          TaskType type1 = t1.getType();
          String typeName1 = (type1 != null && type1.getDescription() != null ? type1.getDescription() : "");

          Task t2 = a2.getTask();
          String displayName2 = t2.getDisplayName().toLowerCase();
          TaskType type2 = t2.getType();
          String typeName2 = (type2 != null && type2.getDescription() != null ? type2.getDescription() : "");

          if (type1 == null && type2 != null)
            result = -1;
          else if (type1 != null && type2 == null)
            result = 1;
          else if ((type1 == null && type2 == null) || (type1.equals(type2)))
            result = displayName1.compareTo(displayName2);
          else
            result = typeName1.compareTo(typeName2);
          return result;
        }
      });

    }
    return assigs;
  }


  public List<Assignment> getMyAssignmentsByPriority() {

    List<Assignment> assigs = null;
    try {
      assigs = getActiveAssignments(new Period(0, CompanyCalendar.MILLIS_IN_YEAR*1000),true,true);
    } catch (PersistenceException e) {
      throw new PlatformRuntimeException(e);
    }

    /*List<Assignment> assigs = new ArrayList<Assignment>();
    if (getAssignmentsSize() > 0) {
      Iterator<Assignment> itass = getAssignmentsIterator();

      while (itass.hasNext()) {
        Assignment ass = itass.next();
        assigs.add(ass);
      }   */

      Collections.sort(assigs, new Comparator<Assignment>() {
        public int compare(Assignment a1, Assignment a2) {

          int result = a2.getPriorityAtTime(System.currentTimeMillis()) - a1.getPriorityAtTime(System.currentTimeMillis());
          if (result == 0) {
            Task t1 = a1.getTask();
            String displayName1 = t1.getDisplayName().toLowerCase();

            Task t2 = a2.getTask();
            String displayName2 = t2.getDisplayName().toLowerCase();

            result = displayName1.compareTo(displayName2);
          }
          return result;
        }
      });

    return assigs;
  }

  public long getTotalWorklogDone() {
    long wd = 0;
    for (Assignment a : assignments) {
      wd = wd + a.getWorklogDone();
    }
    return wd;
  }

  public long getWorklogDone(Date from) {
    return getWorklogDone(from, null);
  }

  public long getWorklogDone(Date from, Date to) {

    String hql = "select sum(wklg.duration) from " + Worklog.class.getName() + " as wklg where wklg.assig.resource = :assignee and wklg.inserted>=:fromx";
    if (to != null)
      hql = hql + " and wklg.inserted<=:tox";

    OqlQuery oql = new OqlQuery(hql);
    oql.getQuery().setEntity("assignee", this);
    oql.getQuery().setTimestamp("fromx", from);
    if (to != null)
      oql.getQuery().setTimestamp("tox", to);
    try {
      return (Long) oql.uniqueResult();
    } catch (FindException e) {
      return 0;
    }
  }


  public List<AnagraphicalData> getAnagraphicalDataOrdered() {

    ArrayList<AnagraphicalData> orderAnagraphicalData = new ArrayList<AnagraphicalData>(getAnagraphicalDatas());

    Collections.sort(orderAnagraphicalData, new Comparator<AnagraphicalData>() {
      public int compare(AnagraphicalData d1, AnagraphicalData d2) {
        return d1.getOrderFactor() - d2.getOrderFactor();
      }
    });

    return orderAnagraphicalData;
  }

  public List<AnagraphicalData> getAnagraphicalDataOrderedNoHidden() {
    List<AnagraphicalData> ads = getAnagraphicalDataOrdered();
    List<AnagraphicalData> adsNoHid = new ArrayList<AnagraphicalData>();
    for (AnagraphicalData ad : ads) {
      if (!ad.isHideAnagraphicalData())
        adsNoHid.add(ad);
    }
    return adsNoHid;
  }

  public String getDefaultEmail() {
    List<AnagraphicalData> ads = getAnagraphicalDataOrderedNoHidden();
    String email = null;
    for (AnagraphicalData ad : ads) {
      if (ad.getEmail() != null) {
        email = ad.getEmail();
        break;
      }
    }
    return email;
  }


  public abstract long getWorkingHourTotal();


  public PersistentFile getMyPhoto() {
    return myPhoto;
  }

  public void setMyPhoto(PersistentFile myPhoto) {
    this.myPhoto = myPhoto;
  }

  public int getHourlyCost() {
    return hourlyCost;
  }

  public void setHourlyCost(int hourlyCost) {
    this.hourlyCost = hourlyCost;
  }

  abstract public TeamworkOperator getMyself();

  public InternetAddress generateEmail() {
    InternetAddress ia = null;
    if (getDefaultEmail() != null && getDefaultEmail().trim().length() > 0) {
      try {
        ia = new InternetAddress(getDefaultEmail());
        ia.setPersonal(getDisplayName(), "UTF-8");
      } catch (AddressException e) {
        throw new PlatformRuntimeException(e);
      } catch (UnsupportedEncodingException e) {
        throw new PlatformRuntimeException(e);
      }
    }
    return ia;
  }

   public List<Resource> getChildrenOrdered() {

     List ch = new ArrayList(getChildren());
     Collections.sort(ch,new Comparator<Resource>() {

       public int compare(Resource o1, Resource o2) {
         if (o1.getClass().equals(o2.getClass())) {
           return o1.getDisplayName().compareTo(o2.getDisplayName());
         } else {
           if (o1 instanceof Company)
            return -1;
           else
            return 1;
         }
       }
     });

     return ch;
   }

}
