package com.twproject.task;

import com.twproject.classification.TaskClassification;
import com.twproject.classification.TaskClassificationAssociation;
import com.twproject.operator.TeamworkOperator;
import com.twproject.resource.PeopleAggregator;
import com.twproject.resource.Person;
import com.twproject.resource.Resource;
import com.twproject.security.RoleTeamwork;
import com.twproject.security.TeamworkPermissions;
import org.jblooming.PlatformRuntimeException;
import org.jblooming.agenda.CompanyCalendar;
import org.jblooming.agenda.Period;
import org.jblooming.logging.Auditable;
import org.jblooming.messaging.SomethingHappened;
import org.jblooming.ontology.Node;
import org.jblooming.ontology.PerformantNodeSupport;
import org.jblooming.operator.Operator;
import org.jblooming.operator.User;
import org.jblooming.oql.OqlQuery;
import org.jblooming.persistence.PersistenceHome;
import org.jblooming.persistence.exceptions.FindException;
import org.jblooming.persistence.exceptions.PersistenceException;
import org.jblooming.persistence.exceptions.RemoveException;
import org.jblooming.persistence.exceptions.StoreException;
import org.jblooming.persistence.hibernate.HibernateFactory;
import org.jblooming.persistence.hibernate.PersistenceContext;
import org.jblooming.security.Area;
import org.jblooming.security.Permission;
import org.jblooming.security.Role;
import org.jblooming.utilities.DateUtilities;
import org.jblooming.utilities.JSP;
import org.jblooming.waf.constants.Commands;
import org.jblooming.waf.html.button.ButtonLink;
import org.jblooming.waf.settings.ApplicationState;
import org.jblooming.waf.view.PageSeed;
import org.jblooming.waf.view.PageState;

import java.lang.reflect.Field;
import java.util.*;

public class Task extends TaskPersistent implements PeopleAggregator, Auditable {

  public boolean resynchedDependencies = false;
  public List<SomethingHappened> happenings = new ArrayList();


  /**
   * Task constructor
   */
  public Task() {
    super();
  }


  private void setParent(Task n) {
    parent = n;
  }

  public Task getParent() {
    return (Task) parent;
  }


  public Node getParentNode() {
    return getParent();
  }

  public void setParentNode(Node node) {
    setParent((Task) node);
  }


  /**
   * We overload SecuredNode's implementation as give the project manager
   * full control on task excluding changing project manager even if she (;-)) has no global permission on area.
   * If task is not persisted, we check on the parent if exists.
   *
   * @param u user for which we ask permission
   * @param p permission required
   */
  public boolean hasPermissionFor(User u, Permission p) {

    boolean result = false;

    result = super.hasPermissionFor(u, p);

    if (!result)
      result = hasPermissionForBasedOnAssignment((TeamworkOperator) u, p);

    return result;
  }

  private boolean hasPermissionForBasedOnAssignment(TeamworkOperator operator, Permission p) {

    Resource myself = operator.getMyPerson();
    boolean result = false;
    for (Assignment assignment : assignments) {
      if (assignment.getResource().equals(myself)) {
        Role role = assignment.getRole();
        if (role != null && role.hasPermissionFor(p)) {
          result = true;
          break;
        }
      }
    }

    return result;
  }

  public Set<Area> getAreasForPermission(TeamworkOperator operator, Permission p) {

    Resource myself = operator.getMyPerson();
    Set<Area> areas = new HashSet();
    for (Assignment assignment : assignments) {
      if (assignment.getResource().equals(myself)) {
        RoleTeamwork role = assignment.getRole();
        if (role != null && role.getArea() != null && role.hasPermissionFor(p)) {
          areas.add(role.getArea());
        }
      }
    }

    if (getParent() != null && (isInherit() || getParent().isPropagate())) {
      Set<Area> parAreas = getParent().getAreasForPermission(operator, p);
      if (parAreas.size() > 0)
        areas.addAll(parAreas);
    }

    return areas;
  }


  public boolean addClassificationAndPersist(TaskClassification classification) throws StoreException, RemoveException {

    boolean result = removeClassificationAndDepersist(classification);

    TaskClassificationAssociation tca = new TaskClassificationAssociation();
    tca.setClassification(classification);
    tca.setTask(this);
    this.getTaskClassificationAssociations().add(tca);
    tca.store();

    return result;
  }

  public boolean removeClassificationAndDepersist(TaskClassification classification) throws RemoveException {

    boolean result = false;

    //find the tca
    for (TaskClassificationAssociation tca : new HashSet<TaskClassificationAssociation>(getTaskClassificationAssociations()))
    {
      if (tca.getClassification().equals(classification)) {
        this.getTaskClassificationAssociations().remove(tca);
        PersistenceHome.remove(tca);
        result = true;
        break;
      }
    }
    return result;
  }


  public static List<String> getStatusesNames() {

    List<String> statuses = new ArrayList<String>();
    Field[] flds = TaskStatus.class.getFields();

    for (Field fld : flds) {
      statuses.add(fld.getName());
    }
    return statuses;
  }


  public String getStatusColor() {
    String color = "#666666";
    if (TaskStatus.STATUS_ACTIVE.equals(getStatus()))
      color = "#66FF99";
    else if (TaskStatus.STATUS_SUSPENDED.equals(getStatus()))
      color = "#FFFF99";
    else if (TaskStatus.STATUS_DONE.equals(getStatus()))
      color = "#0099FF";
    else if (TaskStatus.STATUS_FAILED.equals(getStatus()))
      color = "#660066";
    else if (TaskStatus.STATUS_UNDEFINED.equals(getStatus()))
      color = "#ffffff"; //#666666

    /*else if (TaskStatus.STATUS_TO_BE_SENT_TO_REVIEW.equals(status))
      color = "#888888";
    else if (TaskStatus.STATUS_TO_BE_APPROVED.equals(status))
      color = "#888888";*/
    return color;

  }


  public String getDisplayName() {
    return (getCode() != null ? getCode() + " - " : "") + JSP.w(getName());
  }

  public boolean changeStatusPersistAndPropagate(String oldStatus, String newStatus, String log, boolean generateEvent) throws PersistenceException {
    return changeStatusPersistAndPropagate(oldStatus, newStatus, log, generateEvent, null);
  }


  public boolean changeStatusPersistAndPropagate(String oldStatus, String newStatus, String log, boolean generateEvent, PageState pageState) throws PersistenceException {

    TaskStatusHistory tsth = null;

    boolean doSetStatus = true;

    //takes care of new tasks
    if (oldStatus != null && !(("" + oldStatus).equals("" + newStatus))) {

      //all descendants and all nexts must become suspended
      if (TaskStatus.STATUS_ACTIVE.equals(oldStatus) && TaskStatus.STATUS_SUSPENDED.equals(newStatus)) {

        setStatus(newStatus);
        for (TaskDependency nextTD : getNexts()) {
          Task next = nextTD.getTask();
          if (!TaskStatus.STATUS_DONE.equals(next.getStatus()))
            next.changeStatusPersistAndPropagate(next.getStatus(), newStatus, "", generateEvent, pageState);
        }

        for (PerformantNodeSupport pns : getChildren()) {
          Task task = (Task) pns;
          if (!TaskStatus.STATUS_DONE.equals(task.getStatus()))
            task.changeStatusPersistAndPropagate(task.getStatus(), newStatus, "", generateEvent, pageState);
        }

      } else if (TaskStatus.STATUS_ACTIVE.equals(oldStatus) && TaskStatus.STATUS_FAILED.equals(newStatus)) {

        setStatus(newStatus);

        for (TaskDependency nextTD : getNexts()) {
          Task next = nextTD.getTask();
          if (!TaskStatus.STATUS_DONE.equals(next.getStatus()))
            next.changeStatusPersistAndPropagate(next.getStatus(), newStatus, "", generateEvent, pageState);
        }

        for (PerformantNodeSupport pns : getChildren()) {
          Task task = (Task) pns;
          if (!TaskStatus.STATUS_DONE.equals(task.getStatus()))
            task.changeStatusPersistAndPropagate(task.getStatus(), newStatus, "", generateEvent, pageState);
        }

      } else if (TaskStatus.STATUS_ACTIVE.equals(oldStatus) && TaskStatus.STATUS_UNDEFINED.equals(newStatus)) {

        setStatus(newStatus);

        for (PerformantNodeSupport pns : getChildren()) {
          Task task = (Task) pns;
          if (!TaskStatus.STATUS_DONE.equals(task.getStatus()))
            task.changeStatusPersistAndPropagate(task.getStatus(), newStatus, "", generateEvent, pageState);
        }


      } else if (
              (TaskStatus.STATUS_SUSPENDED.equals(oldStatus) && TaskStatus.STATUS_ACTIVE.equals(newStatus)) ||
                      (TaskStatus.STATUS_UNDEFINED.equals(oldStatus) && TaskStatus.STATUS_ACTIVE.equals(newStatus))
              ) {
        //he he he
        if (getParent() == null || TaskStatus.STATUS_ACTIVE.equals(getParent().getStatus())) {
          setStatus(newStatus);
          for (PerformantNodeSupport pns : getChildren()) {
            Task task = (Task) pns;
            boolean blockedByDependencies = false;
            blockedByDependencies = task.isLocallyBlockedByDependencies();
            if (!blockedByDependencies && !TaskStatus.STATUS_DONE.equals(task.getStatus()))
              task.changeStatusPersistAndPropagate(task.getStatus(), newStatus, "", generateEvent, pageState);
          }
        } else
          doSetStatus = false;


      } else if (TaskStatus.STATUS_DONE.equals(oldStatus) && TaskStatus.STATUS_ACTIVE.equals(newStatus)) {

        setStatus(newStatus);
        for (TaskDependency nextTD : getNexts()) {
          Task next = nextTD.getTask();
          if (!TaskStatus.STATUS_DONE.equals(next.getStatus()))
            next.changeStatusPersistAndPropagate(next.getStatus(), TaskStatus.STATUS_SUSPENDED, "", generateEvent, pageState);
        }

      } else if (
              (TaskStatus.STATUS_SUSPENDED.equals(oldStatus) && TaskStatus.STATUS_DONE.equals(newStatus)) ||
                      (TaskStatus.STATUS_ACTIVE.equals(oldStatus) && TaskStatus.STATUS_DONE.equals(newStatus))
              ) {

        setStatus(newStatus);
        for (PerformantNodeSupport pns : getChildren()) {
          Task task = (Task) pns;
          if (!TaskStatus.STATUS_DONE.equals(task.getStatus()))
            task.changeStatusPersistAndPropagate(task.getStatus(), newStatus, "", generateEvent, pageState);
        }

        for (TaskDependency nextTD : getNexts()) {
          Task next = nextTD.getTask();
          if (TaskStatus.STATUS_SUSPENDED.equals(next.getStatus()) || TaskStatus.STATUS_UNDEFINED.equals(next.getStatus()))
            next.changeStatusPersistAndPropagate(next.getStatus(), TaskStatus.STATUS_ACTIVE, "", generateEvent, pageState);
        }

      } else if (TaskStatus.STATUS_FAILED.equals(oldStatus) && TaskStatus.STATUS_ACTIVE.equals(newStatus)) {

        setStatus(newStatus);

      } else {
        doSetStatus = false;
      }

      if (doSetStatus) {
        tsth = new TaskStatusHistory();
        tsth.setIdAsNewSer();
        tsth.setTask(this);
        tsth.setChangeLog(log);
        tsth.setFromStatus(oldStatus);
        tsth.setToStatus(newStatus);
        tsth.store();


        if (generateEvent) {
          //generate the event
          SomethingHappened change = new SomethingHappened();
          change.setIdAsNew();
          change.setEventType("TASK_STATUS_CHANGE");
          change.setMessageTemplate("TASK_STATUS_CHANGE_MESSAGE_TEMPLATE");

          change.getMessageParams().put("task", getDisplayName());
          change.getMessageParams().put("fromStatus", oldStatus);
          change.getMessageParams().put("toStatus", newStatus);
          change.getMessageParams().put("reason", log);
          change.setWhoCausedTheEvent(pageState.sessionState.getLoggedOperator());

          PageSeed ps = new PageSeed(ApplicationState.serverURL + "/applications/teamwork/task/taskEditor.jsp");
          ps.setCommand(Commands.EDIT);
          ps.setMainObjectId(getId());
          ButtonLink edit = new ButtonLink(ps);
          edit.label = getDisplayName();
          change.setLink(edit.toPlainLink());

          change.setIdentifiable(this);
          happenings.add(change);

        }
      }
    } else {
      setStatus(newStatus);
    }

    return doSetStatus;
  }

  public boolean isLocallyBlockedByDependencies() {
    boolean blockedByDependencies =false;
    for (TaskDependency td : getPreviouses()) {
      if (!TaskStatus.STATUS_DONE.equals(td.getDepends().getStatus())) {
        blockedByDependencies = true;
        break;
      }
    }
    return blockedByDependencies;
  }

  private boolean isFreeFromDependencies() {
    boolean result = true;
    Set<TaskDependency>depts = getPreviouses();
    for (TaskDependency td : depts) {
      if (!TaskStatus.STATUS_DONE.equals(td.getDepends().getStatus())) {
        result = false;
        break;
      }
    }
    return result;
  }


  public TaskScheduleHistory changeSchedule(Date ceStart, int duration, Date ceEnd, String log, Operator causedTheEvent) throws PersistenceException {

    //should we create/update/reset period ?
    Period oldSchedule = getSchedule();

    CompanyCalendar cc = new CompanyCalendar();
    cc.setTime(ceStart != null ? ceStart : new Date());
    Date start = cc.setAndGetTimeToDayStart();

    cc.setTime(ceEnd != null ? ceEnd : (ceStart != null ? ceStart : new Date()));
    Date end = cc.setAndGetTimeToDayEnd();
    Period p = new Period(start, end);

    //boolean oldIsChanged = !(oldSchedule == null ? "" : oldSchedule.toString()).equals(new Period(start, end).toString());
    boolean oldIsChanged =
            (oldSchedule == null && (ceStart != null || ceEnd != null)) ||
            (oldSchedule != null && (ceStart == null && ceEnd == null)) ||
            (oldSchedule == null || !oldSchedule.toString().equals(new Period(start, end).toString()));

    TaskScheduleHistory tsh = null;
    if (oldIsChanged) {
      if (ceStart == null && ceEnd == null) {
        setSchedule(null);
      } else {
        setSchedule(p);
        setDuration(duration);
      }

      tsh = new TaskScheduleHistory();
      tsh.setIdAsNew();
      tsh.setTask(this);
      tsh.setChangeLog(log);
      tsh.setSchedule(oldSchedule);

      //generate the event
      if (getSchedule() != null) {
        SomethingHappened change = new SomethingHappened();
        change.setIdAsNew();
        change.setEventType("TASK_DATE_CHANGE");
        change.setMessageTemplate("TASK_DATE_CHANGE_MESSAGE_TEMPLATE");

        change.getMessageParams().put("task", getDisplayName());

        change.getMessageParams().put("oldstart", oldSchedule != null ? JSP.w(oldSchedule.getStartDate()) : "-");
        change.getMessageParams().put("oldend", oldSchedule != null ? JSP.w(oldSchedule.getEndDate()) : "-");
        change.getMessageParams().put("start", JSP.w(start));
        change.getMessageParams().put("end", JSP.w(end));
        change.getMessageParams().put("reason", JSP.w(log));
        change.setWhoCausedTheEvent(causedTheEvent);

        PageSeed ps = new PageSeed(ApplicationState.serverURL + "/applications/teamwork/task/taskEditor.jsp");
        ps.setCommand(Commands.EDIT);
        ps.setMainObjectId(getId());
        ButtonLink edit = new ButtonLink(ps);
        edit.label = getDisplayName();
        change.setLink(edit.toPlainLink());

        change.setIdentifiable(this);
        happenings.add(change);
      }
    }
    return tsh;
  }


  public void shift(int workingDays) {

    if (getSchedule() != null && getSchedule().getStartDate() != null) {

      CompanyCalendar cc = new CompanyCalendar();
      cc.setTime(getSchedule().getStartDate());
      cc.addWorkingDays(workingDays);
      getSchedule().setStartDate(cc.getTime());

      cc.setTime(getSchedule().getEndDate());
      cc.addWorkingDays(workingDays);
      getSchedule().setEndDate(cc.getTime());

    }

  }

  public boolean previousesContain(Task dependency) {
    boolean result = false;
    for (TaskDependency taskDependency : getPreviouses()) {
      if (taskDependency.getDepends().equals(dependency))
        result = true;
    }
    return result;
  }

  public boolean iDependOn(Task aTask) {
    boolean result = false;
    for (TaskDependency taskDependency : getPreviouses()) {
      if (taskDependency.getDepends().equals(aTask)) {
        result = true;
        break;
      } else
        result = taskDependency.getDepends().iDependOn(aTask);

    }
    return result;
  }

  public void addDependencyInMemory(TaskDependency td) {
    getPreviouses().add(td);
  }

  public void removeDependencyInMemory(TaskDependency taskDependency) {
    getPreviouses().remove(taskDependency);
  }

  public String getNextRelevantEvent(PageState pageState) {

    String relevantEvent = null;
    List<TaskMilestone>mls = getMilestones();
    if (getSchedule() != null) {
      if (isStartIsMilestone()) {
        TaskMilestone tm = new TaskMilestone(getSchedule().getStartDate(), true, this);
        mls.add(tm);
      } else if (isEndIsMilestone()) {
        TaskMilestone tm = new TaskMilestone(getSchedule().getEndDate(), false, this);
        mls.add(tm);
      }

    }

    for (TaskMilestone taskMilestone : mls) {
      //get the first one after now
      if (taskMilestone.when.getTime() > DateUtilities.getAbsoluteNow().getTime()) {

        long missing = getSchedule().getValidityEndTime() - DateUtilities.getAbsoluteNow().getTime();
        relevantEvent = "<b>" + DateUtilities.dateToString(DateUtilities.getAbsoluteInterval(missing), "DD") + "</b>";
        break;
      }
    }

    if (relevantEvent == null) {
      //days missing
      if (getSchedule() != null) {
        if (getSchedule().getValidityEndTime() > DateUtilities.getAbsoluteNow().getTime()) {
          long missing = getSchedule().getValidityEndTime() - DateUtilities.getAbsoluteNow().getTime();
          relevantEvent = DateUtilities.dateToString(DateUtilities.getAbsoluteInterval(missing), "DD");
        }
      }
    }

    if (relevantEvent == null)
      relevantEvent = "?";

    return relevantEvent;
  }

  public List<Task> getChildrenSortedByDateAndCodeName() {

    String hql = "from " + Task.class.getName() + " as task where task.parent = :myself order by task.schedule.start, task.code, task.name";
    OqlQuery oql = new OqlQuery(hql);
    oql.getQuery().setEntity("myself", this);
    try {
      return oql.list();
    } catch (FindException e) {
      throw new PlatformRuntimeException(e);
    }
  }

  public Set<Assignment> getHierarchyAssignments() {
    Set<Assignment> assigs = new HashSet(getAssignments());

    if (getParent() != null) {
      if (isInherit() || getParent().isPropagate())
        assigs.addAll(getParent().getHierarchyAssignments());
    }
    return assigs;
  }

  public int getOpenIssuesSize() {
    String hql = "select count (issue.id) from " + Issue.class.getName() + " as issue where issue.task = :myself and issue.status=:status";
    OqlQuery oql = new OqlQuery(hql);
    oql.getQuery().setString("status", Issue.STATUS_OPEN);
    oql.getQuery().setEntity("myself", this);
    Long i = (Long) oql.uniqueResultNullIfEmpty();
    if (i == null)
      return 0;
    else
      return i.intValue();
  }


  private static int step = 1;


  public static boolean analyzeScheduleChangesRun(Date newStart,
                                                  int newDuration,
                                                  Date newEnd,
                                                  Task task,
                                                  StringBuffer log,
                                                  TeamworkOperator logged,
                                                  Map<Task, TaskScheduleCandidate>taskCandidates
  ) throws FindException {

    boolean result = analyzeScheduleChanges(newStart, newDuration, newEnd, task, log, logged, true, true, taskCandidates);
    //if (result)
    //  result = shift(log, taskCandidates);
    return result;

  }

  /**
   * BEWARE, YOU WHO ENTER: LEAVE ALL HOPES OF IMPROVING THIS METHOD!!!!
   *
   * @param levelEntry means that is the first node modified on the tree level, from which to recur on the parent
   * @return true if dates are consistent and can be set on tasks
   */
  private static boolean analyzeScheduleChanges(Date newStart,
                                                int newDuration,
                                                Date newEnd,
                                                Task task,
                                                StringBuffer log,
                                                TeamworkOperator logged,
                                                boolean changedByHand,
                                                boolean levelEntry,
                                                Map<Task, TaskScheduleCandidate>taskCandidates
  ) throws FindException {

    boolean result = true;
    //System.out.println((step++) + " examining " + task.getName() + ": " + JSP.w(newStart) + "-" + JSP.w(newEnd));

    CompanyCalendar cc = new CompanyCalendar();
    cc.setTime(newStart);
    newStart = cc.setAndGetTimeToDayStart();
    cc.setTime(newEnd);
    newEnd = cc.setAndGetTimeToDayStart();

    if (!task.hasPermissionFor(logged, TeamworkPermissions.task_canWrite)) {

      log.append(task.getDisplayName() + ": permission denied" + "<br>");
      result = false;

    } else {

      Date oldStart = null;
      Date oldEnd = null;
      TaskScheduleCandidate taskScheduleCandidateForTask = createCandidate(task, taskCandidates);
      Period schedule = task.getSchedule();
      if (schedule == null)
        schedule = new Period(newStart, newEnd);
      cc.setTime(schedule.getStartDate());
      oldStart = cc.setAndGetTimeToDayStart();
      cc.setTime(schedule.getEndDate());
      oldEnd = cc.setAndGetTimeToDayStart();

      taskScheduleCandidateForTask.visited = true;

      if (taskScheduleCandidateForTask.start != null) {
        oldStart = taskScheduleCandidateForTask.start;
        oldEnd = taskScheduleCandidateForTask.end;
      }
      taskScheduleCandidateForTask.start = newStart;
      taskScheduleCandidateForTask.end = newEnd;
      taskScheduleCandidateForTask.duration = newDuration;

      if (!changedByHand && task.isStartIsMilestone() && !oldStart.equals(newStart)) {
        log.append(task.getDisplayName() + ": start is milestone - cannot move it from " + JSP.timeStamp(oldStart) + " to " + JSP.timeStamp(newStart) + "<br>");
        result = false;
      } else if (!changedByHand && task.isEndIsMilestone() && !oldEnd.equals(newEnd)) {
        log.append(task.getDisplayName() + ": end is milestone - cannot move it from " + JSP.timeStamp(oldEnd) + " to " + JSP.timeStamp(newEnd) + "<br>");
        result = false;

      }

      if (result)
        findStartFromLefts(task, taskCandidates, taskScheduleCandidateForTask);

      if (result)
        result = shiftDescendants(task, taskCandidates, log, cc);

      if (result)
        enlargeIfNeeded(task, taskScheduleCandidateForTask);

      if (result)
        stabilizeAllNexts(task, taskCandidates);

      if (result)
        result = recurOnNexts(task, taskCandidates, log, logged);

      if (result)
        result = moveUpToParent(task, levelEntry, taskCandidates, log, logged);

    }
    return result;
  }

  private static boolean moveUpToParent(Task task, boolean levelEntry, Map<Task, TaskScheduleCandidate> taskCandidates, StringBuffer log, TeamworkOperator logged) throws FindException {

    TaskScheduleCandidate mainTaskScheduleCandidate = taskCandidates.get(task);
    Date newStart = mainTaskScheduleCandidate.start;
    Date newEnd = mainTaskScheduleCandidate.end;

    boolean result = true;
    //move up to parent if exist and have no left dependencies, i.e. have examined ALL brothers
    Task parent = task.getParent();
    Period max = new Period(newStart, newEnd);
    if (parent != null && levelEntry) {
      for (Object o : parent.getChildren()) {
        Task child = (Task) o;
        TaskScheduleCandidate taskScheduleCandidate = taskCandidates.get(child);
        if (taskScheduleCandidate != null)
          max = max.union(new Period(taskScheduleCandidate.start, taskScheduleCandidate.end));
        else if (child.getSchedule() != null)
          max = max.union(child.getSchedule());
      }
      //System.out.println("moving up from " + task.getName());

      //if parent is wider, leave it wider
      Date maxStartDate = max.getStartDate();
      Date parentStartDate = parent.getSchedule() == null ? new Date(Long.MAX_VALUE) : parent.getSchedule().getStartDate();
      Date maxEndDate = max.getEndDate();
      Date parentEndDate = parent.getSchedule() == null ? new Date(Long.MIN_VALUE) : parent.getSchedule().getEndDate();

      if (parentEndDate != null) {
        if (maxStartDate.getTime() < parentStartDate.getTime() || (maxEndDate.getTime() > parentEndDate.getTime())) {

          Date newMinStart = new Date(Math.min(maxStartDate.getTime(), parentStartDate.getTime()));
          Date newMaxEnd = new Date(Math.max(maxEndDate.getTime(), parentEndDate.getTime()));
          int duration = CompanyCalendar.getWorkingDaysCountInPeriod(new Period(newMinStart, newMaxEnd));

          boolean problemFromParent = !analyzeScheduleChanges(
                  newMinStart,
                  duration,
                  newMaxEnd,
                  parent, log, logged, false, true, taskCandidates);
          if (problemFromParent) {
            result = false;
          }
        }
      }
    }
    return result;
  }

  private static boolean recurOnNexts(Task task, Map<Task, TaskScheduleCandidate> taskCandidates, StringBuffer log, TeamworkOperator logged) throws FindException {
    boolean result = true;
    //recur on nexts
    for (TaskDependency td : task.getNexts()) {

      Task next = td.getTask();
      //System.out.println("nexts: " + task.getName() + " -> " + next.getName());
      TaskScheduleCandidate taskScheduleCandidateForNext = taskCandidates.get(next);
      if (!taskScheduleCandidateForNext.visited && next.getSchedule() != null) {
        if (!analyzeScheduleChanges(taskScheduleCandidateForNext.start, taskScheduleCandidateForNext.duration, taskScheduleCandidateForNext.end, next, log, logged, false, false, taskCandidates))
        {
          result = false;
          break;
        }
      }
    }
    return result;
  }

  private static void stabilizeAllNexts(Task task, Map<Task, TaskScheduleCandidate> taskCandidates) {

    Date newEnd = taskCandidates.get(task).end;

    CompanyCalendar cc = new CompanyCalendar();
    //stabilize all nexts
    for (TaskDependency td : task.getNexts()) {

      Task next = td.getTask();
      //System.out.println("nexts: " + task.getName() + " -> " + next.getName());

      TaskScheduleCandidate taskScheduleCandidateForNext = createCandidate(next, taskCandidates);
      cc.setTime(newEnd);
      cc.addWorkingDays(td.getLag() + 1);
      Date newStartOfDep = new Date(cc.getTime().getTime());
      int duration = taskScheduleCandidateForNext.duration;//getMaximumDurationOfPreviouses(next,taskCandidates);
      cc.addWorkingDays(duration - 1);
      Date newEndOfDep = new Date(cc.getTime().getTime());
      taskScheduleCandidateForNext.start = newStartOfDep;
      taskScheduleCandidateForNext.end = newEndOfDep;
    }
  }

  private static void enlargeIfNeeded(Task task, TaskScheduleCandidate taskScheduleCandidateForTask) {
    //get fatter if needed
    long maxEndFromChildren = 0;
    for (PerformantNodeSupport pns : task.getChildren()) {
      Task t = (Task) pns;
      if (t.getSchedule() != null && t.getSchedule().getEndDate() != null)
        maxEndFromChildren = Math.max(maxEndFromChildren, t.getSchedule().getEndDate().getTime());
    }
    if (taskScheduleCandidateForTask.end.getTime() < maxEndFromChildren) {
      taskScheduleCandidateForTask.end = new Date(maxEndFromChildren);
      taskScheduleCandidateForTask.duration = CompanyCalendar.getWorkingDaysCountInPeriod(new Period(taskScheduleCandidateForTask.start, taskScheduleCandidateForTask.end));
      //CompanyCalendar.getDistanceInWorkingDays(taskScheduleCandidateForTask.start, taskScheduleCandidateForTask.end);
      //System.out.println("widening task: " + task.getName() + " -> " + taskScheduleCandidateForTask.duration);
    }
  }

  private static boolean shiftDescendants(Task task, Map<Task, TaskScheduleCandidate> taskCandidates, StringBuffer log, CompanyCalendar cc) {

    Date newStart = taskCandidates.get(task).start;

    boolean result = true;
    //did start change? if yes, just shift everything if you can
    if (task.getSchedule() != null && !(newStart.getTime() == task.getSchedule().getStartDate().getTime()) && task.getChildrenSize() > 0)
    {
      //System.out.println("moving down from " + task.getName());
      //is there a miles on descendants ?
      Set<Task> descs = task.getDescendants();
      Date startOfDescendants = task.getMinimumStartOfDescendants(taskCandidates);
      int workingDays = CompanyCalendar.getDistanceInWorkingDays(startOfDescendants, newStart);
      int shiftInWorkingDays = workingDays - (int) Math.signum(workingDays);

      for (Task desc : descs) {
        //shift if not already ciancicated
        if (!taskCandidates.containsKey(desc) && desc.getSchedule() != null && desc.getSchedule().getStartDate() != null && desc.getSchedule().getEndDate()!=null)
        {

          if (desc.isStartIsMilestone() || desc.isEndIsMilestone()) {
            result = false;
            log.append(task.getDisplayName() + ": there is a milestone - cannot shift it<br>");
          }

          cc.setTime(desc.getSchedule().getStartDate());
          cc.addWorkingDays(shiftInWorkingDays);
          Date newStartOfDesc = cc.getTime();

          cc.setTime(desc.getSchedule().getEndDate());
          cc.addWorkingDays(shiftInWorkingDays);
          Date newEndOfDesc = cc.getTime();
          //System.out.println((step++) + " shifting " + desc.getName() + ": " + JSP.w(newStartOfDesc) + "-" + JSP.w(newEndOfDesc));

          TaskScheduleCandidate tsc = createCandidate(desc, taskCandidates);
          tsc.start = newStartOfDesc;
          tsc.end = newEndOfDesc;
          tsc.duration = desc.getDuration();
        }
      }
    }
    return result;
  }

  private static void findStartFromLefts(Task task, Map<Task, TaskScheduleCandidate> taskCandidates, TaskScheduleCandidate taskScheduleCandidateForTask) {
    CompanyCalendar cc = new CompanyCalendar();
    long maxEndOfPreviouses = 0;

    //find start from max of ends of lefts
    for (TaskDependency td : task.getPreviouses()) {
      Task previous = td.getDepends();
      //System.out.println("prev: " + task.getName() + " -> " + previous.getName());
      if (previous.getSchedule() != null && previous.getSchedule().getEndDate() != null) {
        long time = 0;
        TaskScheduleCandidate taskScheduleCandidateForPrev = taskCandidates.get(previous);
        if (taskScheduleCandidateForPrev != null)
          time = taskScheduleCandidateForPrev.end.getTime();
        else
          time = previous.getSchedule().getEndDate().getTime();
        cc.setTimeInMillis(time);
        cc.addWorkingDays(td.getLag() + 1);
        maxEndOfPreviouses = Math.max(cc.getTime().getTime(), maxEndOfPreviouses);
      }
    }

    if (maxEndOfPreviouses > 0) {
      //System.out.println("maxEndOfPreviouses for : " + task.getName() + " -> " + new Date(maxEndOfPreviouses));
      taskScheduleCandidateForTask.start = new Date(maxEndOfPreviouses);
      cc.setTimeInMillis(maxEndOfPreviouses);
      taskScheduleCandidateForTask.start = cc.setAndGetTimeToDayStart();
      cc.addWorkingDays(taskScheduleCandidateForTask.duration - 1);
      taskScheduleCandidateForTask.end = cc.getTime();
    }
  }

  private static TaskScheduleCandidate createCandidate(Task t, Map<Task, TaskScheduleCandidate> taskCandidates) {
    TaskScheduleCandidate tsc = taskCandidates.get(t);
    if (tsc == null) {
      tsc = new TaskScheduleCandidate(t);
      tsc.duration = t.getDuration();
      taskCandidates.put(t, tsc);
      //System.out.println("created new for " + t.getName());
    }
    return tsc;
  }

  public Set<Assignment> getAssignmentsForResource(Resource res) {
    Set<Assignment> assigs = new HashSet();
    for (Assignment a : getAssignments()) {
      if (a.getResource().equals(res))
        assigs.add(a);
    }
    return assigs;
  }

  public static class TaskMilestone implements Comparable {

    public Date when;
    public boolean isStart;
    public Task task;

    public TaskMilestone(Date when, boolean isStart, Task task) {
      this.when = when;
      this.isStart = isStart;
      this.task = task;
    }

    public int compareTo(Object miles) {

      TaskMilestone rd2 = (TaskMilestone) miles;
      if (when != null)
        return when.compareTo(rd2.when);
      else
        return 0;
    }

  }

  public List<TaskMilestone> getMilestones() {
    return getMilestones(false);
  }

  public List<TaskMilestone> getMilestones(boolean includeMyself) {

    List<TaskMilestone> ml = new ArrayList<TaskMilestone>();
    if (getSchedule() != null) {
      if (isStartIsMilestone()) {
        ml.add(new TaskMilestone(getSchedule().getStartDate(), true, this));
      }
      if (isEndIsMilestone()) {
        ml.add(new TaskMilestone(getSchedule().getEndDate(), false, this));
      }
    }

    Iterator chI = getChildrenIterator();

    while (chI.hasNext()) {
      Task task = (Task) chI.next();
      if (task.getSchedule() != null) {
        if (task.isStartIsMilestone()) {
          ml.add(new TaskMilestone(task.getSchedule().getStartDate(), true, task));
        }
        if (task.isEndIsMilestone()) {
          ml.add(new TaskMilestone(task.getSchedule().getEndDate(), false, task));
        }
      }
    }

    if (includeMyself && getSchedule() != null) {
      if (isStartIsMilestone() && getSchedule().getStartDate() != null) {
        ml.add(new TaskMilestone(getSchedule().getStartDate(), true, this));
      }
      if (isEndIsMilestone() && getSchedule().getEndDate() != null) {
        ml.add(new TaskMilestone(getSchedule().getEndDate(), false, this));
      }

    }
    Collections.sort(ml);
    return ml;
  }


  public boolean isActive() {
    return TaskStatus.STATUS_ACTIVE.equals(getStatus()); //|| TaskStatus.STATUS_UNDEFINED.equals(getStatus());
    //return ! (TaskStatus.STATUS_DONE.equals(getStatus()) || TaskStatus.STATUS_SUSPENDED.equals(getStatus()) || TaskStatus.STATUS_FAILED.equals(getStatus()));
  }

  public boolean isClosed() {
    return TaskStatus.STATUS_DONE.equals(getStatus()) || TaskStatus.STATUS_FAILED.equals(getStatus());
  }

  public void addAssignmentInMemory(Assignment a) {
    getAssignments().add(a);
  }

  public Iterator<TaskDependency> getPreviousesIteratorByStart() {
    List<TaskDependency> sorted = new ArrayList(getPreviouses());
    Collections.sort(sorted, new Comparator() {
      public int compare(Object o1, Object o2) {
        TaskDependency s1 = (TaskDependency) o1;
        TaskDependency s2 = (TaskDependency) o2;
        Period schedule1 = s1.getDepends().getSchedule();
        Period schedule2 = s2.getDepends().getSchedule();
        if (schedule1 != null && schedule2 != null)
          return schedule1.getValidityStartDate().compareTo(schedule2.getValidityStartDate());
        else
          return 0;
      }
    });
    return sorted.iterator();
  }

  public Iterator<TaskDependency> getNextsIteratorByStart() {
    List<TaskDependency> sorted = new ArrayList(getNexts());
    Collections.sort(sorted, new Comparator() {
      public int compare(Object o1, Object o2) {
        TaskDependency s1 = (TaskDependency) o1;
        TaskDependency s2 = (TaskDependency) o2;
        Period schedule1 = s1.getDepends().getSchedule();
        Period schedule2 = s2.getDepends().getSchedule();
        if (schedule1 != null && schedule2 != null)
          return schedule1.getValidityStartDate().compareTo(schedule2.getValidityStartDate());
        else
          return 0;
      }
    });
    return sorted.iterator();
  }


  public List<Resource> getWorkGroup() {

    List<Resource> result = new ArrayList<Resource>();
    addWorkersToList(this, result);
    return result;
  }

  private static void addWorkersToList(Task task, List<Resource> result) {

    //get from assigs
    Iterator it = task.getAssignementsIterator();
    while (it.hasNext()) {
      Assignment assignment = (Assignment) it.next();
      result.add(assignment.getResource());
    }

    //if propag., recurse
    if (task.isPropagate() && task.getChildrenNode().size() > 0) {
      for (Object child : task.getChildrenNode()) {
        addWorkersToList((Task) child, result);
      }
    }
  }

  public LinkedList<Issue> getOpenIssuesById(PageState pageState) throws PersistenceException {

    TeamworkOperator logged = (TeamworkOperator) pageState.sessionState.getLoggedOperator();

    LinkedList<Issue> ll = new LinkedList<Issue>();

    Iterator i = getIssuesIterator();
    while (i.hasNext()) {
      Issue issue = (Issue) i.next();
      if (Issue.STATUS_OPEN.equals(issue.getStatus())) {
        if (issue.hasPermissionFor(logged, TeamworkPermissions.issue_canRead))
          ll.add(issue);
      }
    }
    Collections.sort(ll);
    return ll;
  }

  public List<Task> getDescendantsOrdered(boolean showDescendants, String orderBy) throws FindException {

    List descs = getDescendantIds(Task.class);
    String hql = "from " + Task.class.getName() + " as task where task.parent = :myself";
    if (showDescendants && descs.size() > 0)
      hql = hql + " or task.id in (:descs) ";
    if (orderBy != null)
      hql = hql + " order by " + orderBy;
    OqlQuery oql = new OqlQuery(hql);
    if (showDescendants && descs.size() > 0)
      oql.getQuery().setParameterList("descs", descs);
    oql.getQuery().setEntity("myself", this);
    //TreeSet<Task> ts = new TreeSet<Task>(oql.list());
    return oql.list();
  }

  /**
   * does not include this
   * surfes objects instead of doing an optimized query
   */
  public Set<Task> getDescendants() {
    return addDescendants(this, new HashSet());
  }

  private Set<Task> addDescendants(Task task, Set<Task> descs) {
    for (PerformantNodeSupport performantNodeSupport : task.getChildren()) {
      Task desc = (Task) performantNodeSupport;
      descs.add(desc);
      addDescendants(desc, descs);
    }
    return descs;
  }

  public TreeSet<Task> getDescendantsActiveOrdered(boolean showDescendants, String orderBy) throws FindException {

    List descs = getDescendantIds(Task.class);
    showDescendants = showDescendants && descs.size() > 0;
    String hql = "from " + Task.class.getName() + " as task where task.parent = :myself";
    if (showDescendants)
      hql = hql + " or task.id in (:descs) ";
    hql = hql + " and (task.status = :stActive or task.status = :stUndef)";
    if (orderBy != null)
      hql = hql + " order by " + orderBy;
    OqlQuery oql = new OqlQuery(hql);
    oql.getQuery().setString("stActive", TaskStatus.STATUS_ACTIVE);
    oql.getQuery().setString("stUndef", TaskStatus.STATUS_UNDEFINED);

    if (showDescendants)
      oql.getQuery().setParameterList("descs", descs);

    oql.getQuery().setEntity("myself", this);
    TreeSet<Task> ts = new TreeSet<Task>(oql.list());
    return ts;
  }

  public long getTotalWorklogDone(boolean includeChildren) {
    return getTotalWorklogDone(includeChildren, HibernateFactory.getDefaultPersistenceContext());
  }

  public long getTotalWorklogDone(boolean includeChildren, PersistenceContext pc) {
    long twd = 0;
    for (Assignment a : getAssignments()) {
      twd += a.getWorklogDone(pc);
    }
    if (includeChildren) {
      Iterator i = getChildrenIterator();
      while (i.hasNext()) {
        Task task = (Task) i.next();
        twd += task.getTotalWorklogDone(includeChildren, pc);
      }
    }
    return twd;
  }

  public long getTotalWorklogEstimated(boolean includeChildren) {

    long twd = 0;

    for (Assignment a : getAssignments()) {
      twd += a.getEstimatedWorklog();
    }
    if (includeChildren) {
      Iterator i = getChildrenIterator();
      while (i.hasNext()) {
        Task task = (Task) i.next();
        twd += task.getTotalWorklogEstimated(includeChildren);
      }
    }

    return twd;
  }

  public boolean isPersonIn(Person o) {
    return getPersons().contains(o);
  }

  public Set<Person> getPersons() {

    return getPersons(new HashSet<PeopleAggregator>(), new HashSet<Person>());
  }


  public Set<Person> getPersons(Set<PeopleAggregator> visitedNodes, Set<Person> workers) {

    Resource res;

    if (!visitedNodes.contains(this)) {

      visitedNodes.add(this);

      for (Object assignement : assignments) {
        Assignment assignment = (Assignment) assignement;
        res = assignment.getResource();
        if (res != null && !res.isHidden())
          workers.addAll(res.getPersons());
      }

      // visiting parent node

      if (this.isInherit() && getParent() != null) {
        workers.addAll((getParent()).getPersons(visitedNodes, workers));
      }

      // visiting children

      if (this.isPropagate()) {
        Iterator iter = this.getChildrenIterator();
        Task t;
        while (iter.hasNext()) {
          t = (Task) iter.next();
          workers.addAll(t.getPersons(visitedNodes, workers));
        }
      }
    }
    return workers;
  }

  public void updateProgress(PersistenceContext pc) throws StoreException {
    if (isProgressByWorklog()) {
      long totalWorklogEstimated = getTotalWorklogEstimated(true);
      updateProgressFromEstimated(totalWorklogEstimated, pc);
    }
  }

  private void updateProgressFromEstimated(long totalWorklogEstimated, PersistenceContext pc) throws StoreException {
    if (totalWorklogEstimated > 0) {
      long deltaWklgPerc = getTotalWorklogDone(true, pc) * 100 / totalWorklogEstimated;
      setProgress(deltaWklgPerc);
      this.store(pc);
      if (getParent() != null)
        getParent().updateProgress(pc);
    }
  }

  public Date getMinimumStartOfDescendants(Map<Task, TaskScheduleCandidate> taskCandidates) {

    Set<Task> descs = getDescendants();
    long min = Long.MAX_VALUE;
    for (Task task : descs) {
      TaskScheduleCandidate taskScheduleCandidate = taskCandidates.get(task);
      if (taskScheduleCandidate !=null)
        min = Math.min(min, taskScheduleCandidate.start.getTime());      
      else if (task.getSchedule() != null && task.getSchedule().getStartDate() != null)
        min = Math.min(min, task.getSchedule().getStartDate().getTime());
    }
    return new Date(min);
  }


  private static long getMaximumDurationOfPreviouses(Task task, Map<Task, TaskScheduleCandidate>taskCandidates) {

    long result = 0;
    for (TaskDependency td : task.getPreviouses()) {
      Task previous = td.getDepends();

      TaskScheduleCandidate taskScheduleCandidate = taskCandidates.get(previous);
      if (taskScheduleCandidate != null)
        result = Math.max(result, taskScheduleCandidate.duration);
      else
        result = Math.max(result, previous.getDuration());
    }
    return result;
  }

  private static long getMaximumDurationOfNexts(Task task, Map<Task, TaskScheduleCandidate>taskCandidates) {

    long result = 0;
    for (TaskDependency td : task.getNexts()) {
      Task next = td.getTask();

      TaskScheduleCandidate taskScheduleCandidate = taskCandidates.get(next);
      if (taskScheduleCandidate != null)
        result = Math.max(result, taskScheduleCandidate.duration);
      else
        result = Math.max(result, next.getDuration());
    }
    return result;
  }

}
