package com.twproject.two.agenda;


import org.jblooming.ApplicationException;
import org.jblooming.agenda.CompanyCalendar;

import java.util.*;

public class ScheduleWeeklyOld extends ScheduleSupportOld {

  private int freq = 1;                 // how many weeks the event is repeated
  private int repeat = 0;                // number of times the event is repeated
  protected boolean onlyWorkingDays = false;    // whether the event is repeated only on working days
  private int dayArray[];                         //specifies the days the event is repeated {Calendar.Monday,... and so on....

  public static final int FIRST = 1;
  public static final int SECOND = 2;
  public static final int THIRD = 3;
  public static final int FOURTH = 4;
  public static final int LAST = 5;

  public String id;

  public ScheduleWeeklyOld() {
  }

  public ScheduleWeeklyOld(String i, PeriodOld p) {
    this(i,
        CompanyCalendar.ALLDAYS,
        new Date(p.getValidityStart()),
        new Date(p.getValidityEnd()),
        p.getStartMillis(),
        p.getEndMillis());
  }

  public ScheduleWeeklyOld(String i, int days[]) {
    this(i, days, CompanyCalendar.MIN_DATE, CompanyCalendar.MAX_DATE, CompanyCalendar.MIN_TIME, CompanyCalendar.MAX_TIME);
  }

  public ScheduleWeeklyOld(String i, int days[], Date s) {
    this(i, days, s, CompanyCalendar.MAX_DATE, CompanyCalendar.MIN_TIME, CompanyCalendar.MAX_TIME);
  }

  public ScheduleWeeklyOld(String i, int days[], Date s, Date e) {
    this(i, days, s, e, CompanyCalendar.MIN_TIME, CompanyCalendar.MAX_TIME);
  }

  public ScheduleWeeklyOld(String i, int days[], Date s, Date e, int st, int et) {
    this(i, days, s, e, st, et, 1, 1, false);
  }

  /**
   * @param pId             specifies schedule id
   * @param days            specifies the days the event is repeated Calendar.MONDAY etc.
   * @param start           specifies the initial validity date (no meaning for time)
   * @param end             specifies the final   validity date (no meaning for time)
   * @param startTime       specifies the start time in milliseconds from midnight
   * @param endTime         specifies the end   time in milliseconds from midnight
   * @param freq            specifies the frequency (how every weeks) the event is repeated
   * @param repetition      specifies how many times the event is repeated
   * @param onlyWorkingDays specifies if repeates only on the workin days
   */
  public ScheduleWeeklyOld(String pId, int days[], Date start, Date end, int startTime, int endTime, int freq, int repetition, boolean onlyWorkingDays) {
    this.setStart(start);
    this.setEnd(end);
    this.setId(pId);

    setFrequency(freq);
    setStartMillis(startTime);
    setEndMillis(endTime);
    setRepetitions(repetition);
    setDays(days);
    setOnlyWorkingDays(onlyWorkingDays);
  }

  /**
   * This method return the validity period.
   * The period is trimmed to the starting time and the ending time.
   * There is no test for working days validity (eg. for example if  the validity period starts on sat at 8.00
   * and you choos working days only, the validity period start on sat at 8.00 NOT on mon 8.00)
   */
  public PeriodOld getValidityPeriod() {
    return new PeriodOld(getStart(), getEnd());
  }


  /**
   * Sets the event frequency in weeks;
   */
  public void setFrequency(int f) {
    freq = f > 0 ? f : 1;
    setRepetitions(repeat); // in this way the endpoints are synchronized
  }

  /**
   * Return the event frequency in days;
   */
  public int getFrequency() {
    return freq;
  }


  /**
   * Sets if only the working days are considered.
   * It is based on CompanyCalendar
   *
   * @see CompanyCalendar
   */
  public void setOnlyWorkingDays(boolean f) {
    onlyWorkingDays = f;
  }

  public boolean getOnlyWorkingDays() {
    return onlyWorkingDays;
  }

  /**
   * Set the week_days when the event occurs.
   * The validity period is recalculated according the days specified.
   * The start_date is set to the lowest day (of the same week), the end_date is set to the great one.
   */
  public void setDays(int d[]) {
    dayArray = d;
    Calendar cal = new CompanyCalendar();
    cal.setTime(start);
    for (int i = 0; i < d.length; i++) {
      cal.set(Calendar.DAY_OF_WEEK, d[i]);
      if (cal.getTime().before(start))
        start = cal.getTime();
    }

    cal.setTime(end);
    for (int i = 0; i < d.length; i++) {
      cal.set(Calendar.DAY_OF_WEEK, d[i]);
      if (cal.getTime().after(end))
        end = cal.getTime();
    }
  }

  public int[] getDays() {
    return dayArray;
  }

  /**
   * specifies the number of cycles before expires
   * If repetitions is >=1 the end date is recalculated using start+frequency*(repetition-1)*7
   * repetitions=0 means forever (the end date is moved to MAXDATE)
   */
  public void setRepetitions(int repetitions) {
    Calendar cal;
    cal = new CompanyCalendar();
    int originalEndTime = getEndMillis();

    if (repetitions > 1) {
      repeat = repetitions;
      cal.setTime(start);
      cal.add(cal.WEEK_OF_YEAR, (repeat - 1) * freq);
    } else {
      repeat = 0;
      cal.setTime(CompanyCalendar.MAX_DATE);
      cal.set(Calendar.HOUR_OF_DAY, 0);
      cal.clear(cal.MINUTE);
      cal.clear(cal.SECOND);
      cal.clear(cal.MILLISECOND);
      cal.add(cal.MILLISECOND, getEndMillis());
    }
    setEnd(cal.getTime());
    setEndMillis(originalEndTime);
  }

  public int getRepetitions() {
    return repeat;
  }


  /**
   * Return true if the specified date is in ANY of schedule; false oterwhise
   */
  public boolean contains(Date d) {
    long distInMillisec;
    int distInWeek;
    int distInDays;

    CompanyCalendar cal = new CompanyCalendar();
    cal.setTime(d);

    // tests if the date is in the validity period
    if (!getValidityPeriod().contains(d))
      return false;

    // if the week_day is not present
    boolean ok = false;
    for (int i = 0; i < dayArray.length && !ok; i++)
      ok = (dayArray[i] == cal.DAY_OF_WEEK);

    // test for working days //PENDING potrebbe essere utile testare per le festivit� mobili
    if (onlyWorkingDays && (cal.isWorkingDay()))
      return false;

    //normalize dist
    distInMillisec = d.getTime() - start.getTime();
    distInDays = (int) (distInMillisec + CompanyCalendar.MAX_TIME - 1) / CompanyCalendar.MAX_TIME;
    distInWeek = (distInDays / 7);

    // test if the week is candidate using module computation
    if (distInWeek % (freq) != 0)
      return false;

    Date s, e;
    cal.setTime(getStart());
    cal.roll(cal.DATE, distInDays);
    s = cal.getTime();

    cal.roll(cal.MILLISECOND, getEndMillis() - getStartMillis());
    e = cal.getTime();

    return (s.getTime() <= d.getTime()) && (d.getTime() <= e.getTime());
  }


  /**
   * This method return true when the period is
   * overlapped (partially o complete) on the validity period, even if actually, it is not, looking on days.
   */
  public boolean isIn(PeriodOld p) {
    return !((getEnd().getTime() < p.getValidityEnd()) || (getStart().getTime() > p.getValidityEnd()));
  }


  /**
   * Calculates the duration of the period as difference (in millis) from start and end date
   */
  public long getDuration() {
    if (getStart().getTime() == CompanyCalendar.MIN_DATE.getTime() || getEnd().getTime() == CompanyCalendar.MAX_DATE.getTime())
      return Long.MAX_VALUE;
    return getEnd().getTime() - getStart().getTime();
  }


  /**
   * Return the collection of periods, compliant with the ScheduleWeekly, in period specified
   * WARNING this method should be used for short periods
   */
  public Collection getPeriods(PeriodOld p) throws ApplicationException {

    if (!((p.getValidityEnd() < this.start.getTime()) || (this.end.getTime() < p.getValidityStart()))) {
      int increment;
      increment = freq > 0 ? freq : 1;

      List v = new ArrayList();
      if (p.getValidityEnd() < CompanyCalendar.MAX_DATE.getTime() && p.getValidityStart() > CompanyCalendar.MIN_DATE.getTime()) {
        CompanyCalendar st = new CompanyCalendar();
        CompanyCalendar en = new CompanyCalendar();
        int j;

        st.setTime(new Date(this.getStart().getTime()));

        en.setTime(st.getTime());
        en.setMillisFromMidnight(getEndMillis());

        // loop inside the period
        for (int i = 0; !en.getTime().after(new Date(p.getValidityEnd() + CompanyCalendar.MAX_TIME * 7)) && (i < repeat || repeat == 0); i++) {
          for (j = 0; j < dayArray.length; j++) {

            st.set(CompanyCalendar.DAY_OF_WEEK, dayArray[j]);
            en.set(CompanyCalendar.DAY_OF_WEEK, dayArray[j]);
            PeriodOld candidate = new PeriodOld(st.getTime(), en.getTime());

            // test if the item is in the desired period
            if (p.overlap(candidate)) {
              // test for working days
              if (onlyWorkingDays) {
                if (st.isWorkingDay() && en.isWorkingDay())
                  v.add(candidate);
              } else {
                v.add(candidate);
              }
            }
          }

          st.add(CompanyCalendar.WEEK_OF_YEAR, increment);
          en.add(CompanyCalendar.WEEK_OF_YEAR, increment);
        }
      } else {
        throw new ApplicationException("Invalid period specification");
      }
      return v;
    } else {
      return Collections.EMPTY_LIST;
    }
  }


  protected int getFreq() {
    return freq;
  }

  protected void setFreq(int freq) {
    this.freq = freq;
  }

  protected int getRepeat() {
    return repeat;
  }

  protected void setRepeat(int repeat) {
    this.repeat = repeat;
  }

  protected int[] getDayArray() {
    return dayArray;
  }

  protected void setDayArray(int[] dayArray) {
    this.dayArray = dayArray;
  }
}

